@@ -9,10 +9,10 @@
 echo "Autoconf in root..."
 autoconf
 
-for dir in vpip vvp tgt-vvp tgt-fpga
+for dir in vpip vpi vvp tgt-vvp tgt-fpga libveriuser cadpli
 do
     echo "Autoconf in $dir..."
-    ( cd $dir ; autoconf )
+    ( cd ./$dir ; autoconf --include=.. )
 done
 
 echo "Precompiling lexor_keyword.gperf"
