@@ -16,8 +16,8 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT) && !defined(macintosh)
-#ident "$Id: design_dump.cc,v 1.121 2001/12/31 00:03:05 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: design_dump.cc,v 1.149 2004/10/04 01:10:52 steve Exp $"
 #endif
 
 # include "config.h"
@@ -93,7 +93,7 @@ void NetNet::dump_net(ostream&o, unsigned ind) const
 	    o << " inout";
 	    break;
       }
-      o << " (eref=" << get_eref() << ")";
+      o << " (eref=" << peek_eref() << ", lref=" << peek_lref() << ")";
       if (scope())
 	    o << " scope=" << scope()->name();
       o << " #(" << rise_time() << "," << fall_time() << "," <<
@@ -117,6 +117,13 @@ void NetNet::dump_net(ostream&o, unsigned ind) const
       dump_obj_attr(o, ind+4);
 }
 
+void NetSubnet::dump_net(ostream&o, unsigned ind) const
+{
+      o << setw(ind) << "" << "** " << name() << " is a NetSubnet **"
+	<< endl;
+      NetNet::dump_net(o, ind);
+}
+
 void NetMemory::dump(ostream&o, unsigned ind) const
 {
       o << setw(ind) << ""  << name_ << "[" << width_ << "] " <<
@@ -174,9 +181,9 @@ void NetObj::dump_node_pins(ostream&o, unsigned ind) const
 void NetObj::dump_obj_attr(ostream&o, unsigned ind) const
 {
       unsigned idx;
-      for (idx = 0 ;  idx < attributes_.size() ;  idx += 1) {
-	    o << setw(ind) << "" << attributes_.key(idx) << " = \"" <<
-		  attributes_.value(idx) << "\"" << endl;
+      for (idx = 0 ;  idx < attr_cnt() ;  idx += 1) {
+	    o << setw(ind) << "" << attr_key(idx) << " = \"" <<
+		  attr_value(idx) << "\"" << endl;
       }
 }
 
@@ -264,7 +271,8 @@ void NetConst::dump_node(ostream&o, unsigned ind) const
 void NetFF::dump_node(ostream&o, unsigned ind) const
 {
       o << setw(ind) << "" << "LPM_FF: " << name()
-	<< " scope=" << (scope()? scope()->name() : "") << endl;
+	<< " scope=" << (scope()? scope()->name() : "")
+	<< " aset_value=" << aset_value_ << endl;
 
       dump_node_pins(o, ind+4);
       dump_obj_attr(o, ind+4);
@@ -353,6 +361,14 @@ void NetRamDq::dump_node(ostream&o, unsigned ind) const
       dump_obj_attr(o, ind+4);
 }
 
+void NetUserFunc::dump_node(ostream&o, unsigned ind) const
+{
+      o << setw(ind) << "" << def_->name() << "(";
+      o << ")" << endl;
+      dump_node_pins(o, ind+4);
+      dump_obj_attr(o, ind+4);
+}
+
 void NetTaskDef::dump(ostream&o, unsigned ind) const
 {
       o << setw(ind) << "" << "task " << name_ << ";" << endl;
@@ -405,6 +421,11 @@ void NetProcTop::dump(ostream&o, unsigned ind) const
 	    break;
       }
 
+      for (unsigned idx = 0 ;  idx < attr_cnt() ;  idx += 1) {
+	    o << setw(ind+2) << "" << "(* " << attr_key(idx) << " = "
+	      << attr_value(idx) << " *)" << endl;
+      }
+
       statement_->dump(o, ind+2);
 }
 
@@ -418,6 +439,16 @@ void NetAssign_::dump_lval(ostream&o) const
 	    } else {
 		  o << "[" << (loff_+lwid_-1) << ":" << loff_ << "]";
 	    }
+      } else if (mem_) {
+	    // Is there an obvious way to flag memories in the dump
+	    // as different from the _real_ bit mux case?
+	    // o << "**memory**";
+	    o << mem_->name() << "[";
+	    if (bmux_) o << *bmux_;
+	      else     o << "**oops**";
+	    o << "]";
+      } else if (var_) {
+	    o << "<real " << var_->basename() << ">";
       } else {
 	    o << "";
       }
@@ -464,30 +495,22 @@ void NetAssignNB::dump(ostream&o, unsigned ind) const
 
 }
 
-void NetAssignMem::dump(ostream&o, unsigned ind) const
+void NetAssignBase::dump(ostream&o, unsigned ind) const
 {
-      o << setw(ind) << "";
-      o << "/* " << get_line() << " */" << endl;
-      o << setw(ind) << "";
-      o << memory()->name() << "[" << *index() << "] = ";
-      rval()->dump(o);
-      o << ";" << endl;
-}
-
-void NetAssignMemNB::dump(ostream&o, unsigned ind) const
-{
-      o << setw(ind) << "";
-      o << "/* " << get_line() << " */" << endl;
-      o << setw(ind) << "";
-      o << memory()->name() << "[" << *index() << "] <= ";
-      rval()->dump(o);
-      o << ";" << endl;
+      if (const NetAssignNB *n1 = dynamic_cast<const NetAssignNB*>(this)) {
+	    n1->dump(o,ind);
+      } else if (const NetAssign *n2 = dynamic_cast<const NetAssign*>(this)) {
+	    n2->dump(o,ind);
+      }
 }
 
 /* Dump a block statement */
 void NetBlock::dump(ostream&o, unsigned ind) const
 {
-      o << setw(ind) << "" << type_ << endl;
+      o << setw(ind) << "" << type_;
+      if (subscope_)
+	    o << " : " << subscope_->name();
+      o << endl;
 
       if (last_) {
 	    const NetProc*cur = last_;
@@ -622,6 +645,13 @@ void NetForever::dump(ostream&o, unsigned ind) const
 void NetFuncDef::dump(ostream&o, unsigned ind) const
 {
       o << setw(ind) << "" << "function " << scope_->name() << endl;
+      if (result_sig_)
+	    o << setw(ind+2) << "" << "Return signal: "
+	      << result_sig_->name() << endl;
+      if (result_var_)
+	    o << setw(ind+2) << "" << "Return variable: "
+	      << result_var_->basename() << endl;
+
       if (statement_)
 	    statement_->dump(o, ind+2);
       else
@@ -647,8 +677,12 @@ void NetPDelay::dump(ostream&o, unsigned ind) const
 
 void NetRelease::dump(ostream&o, unsigned ind) const
 {
-      o << setw(ind) << "" << "release " << lval_->name() << "; "
-	<< "/* " << get_line() << " */" << endl;
+      if (lval_)
+	    o << setw(ind) << "" << "release " << lval_->name() << "; "
+	      << "/* " << get_line() << " */" << endl;
+      else
+	    o << setw(ind) << "" << "release (null); "
+	      << "/* " << get_line() << " */" << endl;
 }
 
 void NetRepeat::dump(ostream&o, unsigned ind) const
@@ -659,7 +693,9 @@ void NetRepeat::dump(ostream&o, unsigned ind) const
 
 void NetScope::dump(ostream&o) const
 {
+	/* This is a constructed hierarchical name. */
       o << name();
+
       switch (type_) {
 	  case BEGIN_END:
 	    o << " sequential block";
@@ -671,7 +707,7 @@ void NetScope::dump(ostream&o) const
 	    o << " function";
 	    break;
 	  case MODULE:
-	    o << " module <" << (module_name_? module_name_ : "") << ">";
+	    o << " module <" << (module_name_? module_name_.str() : "") << ">";
 	    break;
 	  case TASK:
 	    o << " task";
@@ -679,22 +715,35 @@ void NetScope::dump(ostream&o) const
       }
       o << endl;
 
+      for (unsigned idx = 0 ;  idx < attr_cnt() ;  idx += 1)
+	    o << "    (* " << attr_key(idx) << " = "
+	      << attr_value(idx) << " *)" << endl;
+
       o << "    timescale = 10e" << time_unit() << " / 10e"
 	<< time_precision() << endl;
 
 	/* Dump the parameters for this scope. */
       {
-	    map<string,NetExpr*>::const_iterator pp;
-	    for (pp = parameters_.begin()
-		       ; pp != parameters_.end() ;  pp ++) {
-		  o << "    parameter " << (*pp).first << " = "  <<
-			*(*pp).second << ";" << endl;
+	    map<perm_string,param_expr_t>::const_iterator pp;
+	    for (pp = parameters.begin()
+		       ; pp != parameters.end() ;  pp ++) {
+		  o << "    parameter ";
+
+		  if ((*pp).second.signed_flag)
+			o << "signed ";
+
+		  if ((*pp).second.msb)
+			o << "[" << *(*pp).second.msb
+			  << ":" << *(*pp).second.lsb << "] ";
+
+		  o << (*pp).first << " = "  <<
+			*(*pp).second.expr << ";" << endl;
 	    }
 
-	    for (pp = localparams_.begin()
-		       ; pp != localparams_.end() ;  pp ++) {
+	    for (pp = localparams.begin()
+		       ; pp != localparams.end() ;  pp ++) {
 		  o << "    localparam " << (*pp).first << " = "  <<
-			*(*pp).second << ";" << endl;
+			*(*pp).second.expr << ";" << endl;
 	    }
       }
 
@@ -708,6 +757,11 @@ void NetScope::dump(ostream&o) const
 	    }
       }
 
+      for (NetVariable*cur = vars_ ;  cur ;  cur = cur->snext_) {
+	    o << "    real " << cur->basename() << " // "
+	      << cur->get_line() << endl;
+      }
+
 	/* Dump the events in this scope. */
       for (NetEvent*cur = events_ ;  cur ;  cur = cur->snext_) {
 	    o << "    event " << cur->name() << "; nprobe="
@@ -734,11 +788,16 @@ void NetScope::dump(ostream&o) const
 
       switch (type_) {
 	  case FUNC:
-	    func_def()->dump(o, 4);
+	    if (func_def())
+		  func_def()->dump(o, 4);
+	    else
+		  o << "    MISSING FUNCTION DEFINITION" << endl;
 	    break;
 	  case TASK:
 	    task_def()->dump(o, 4);
 	    break;
+	  default:
+	    break;
       }
 
 	/* Dump any sub-scopes. */
@@ -783,7 +842,7 @@ void NetProc::dump(ostream&o, unsigned ind) const
       o << setw(ind) << "" << "// " << typeid(*this).name() << endl;
 }
 
-/* Dump an expression that noone wrote a dump method for. */
+/* Dump an expression that no one wrote a dump method for. */
 void NetExpr::dump(ostream&o) const
 {
       o << "(?" << typeid(*this).name() << "?)";
@@ -831,6 +890,9 @@ void NetEBinary::dump(ostream&o) const
 	  case 'r':
 	    o << ">>";
 	    break;
+	  case 'R':
+	    o << ">>>";
+	    break;
 	  case 'X':
 	    o << "~^";
 	    break;
@@ -842,8 +904,12 @@ void NetEBinary::dump(ostream&o) const
 
 void NetEConcat::dump(ostream&o) const
 {
-      if (repeat_ != 1)
-	    o << repeat_;
+      if (repeat_calculated_) {
+	    if (repeat_value_ != 1)
+		  o << repeat_value_;
+      } else if (repeat_) {
+	    o << "<" << *repeat_ << ">";
+      }
 
       if (parms_[0])
 	    o << "{" << *parms_[0];
@@ -867,11 +933,54 @@ void NetEConst::dump(ostream&o) const
 	    o << value_;
 }
 
+void NetEConstParam::dump(ostream&o) const
+{
+      o << "<" << name_ << "=";
+      NetEConst::dump(o);
+      o << ">";
+}
+
+void NetECReal::dump(ostream&o) const
+{
+      o << value_;
+}
+
+void NetECRealParam::dump(ostream&o) const
+{
+      o << "<" << name_ << "=";
+      NetECReal::dump(o);
+      o << ">";
+}
+
+void NetEEvent::dump(ostream&o) const
+{
+      o << "<event=" << event_->name() << ">";
+}
+
 void NetEScope::dump(ostream&o) const
 {
       o << "<scope=" << scope_->name() << ">";
 }
 
+void NetESelect::dump(ostream&o) const
+{
+      o << "<select";
+      if (has_sign())
+	    o << "+=";
+      else
+	    o << "=";
+
+      expr_->dump(o);
+      o << "[";
+
+      if (base_)
+	    base_->dump(o);
+      else
+	    o << "(0)";
+
+      o << "+:" << expr_width() << "]>";
+}
+
 void NetESFunc::dump(ostream&o) const
 {
       o << name_ << "(";
@@ -906,7 +1015,10 @@ void NetEMemory::dump(ostream&o) const
 
 void NetEParam::dump(ostream&o) const
 {
-      o << "<" << scope_->name() << "." << name_ << ">";
+      if (scope_ != 0)
+	    o << "<" << scope_->name() << "." << name_ << ">";
+      else
+	    o << "<" << name_ << ">";
 }
 
 void NetETernary::dump(ostream&o) const
@@ -943,11 +1055,16 @@ void NetEUnary::dump(ostream&o) const
       o << ")";
 }
 
+void NetEVariable::dump(ostream&o) const
+{
+      o << var_->basename();
+}
+
 void Design::dump(ostream&o) const
 {
       o << "DESIGN TIME PRECISION: 10e" << get_precision() << endl;
       o << "SCOPES:" << endl;
-      for (list<NetScope*>::const_iterator scope = root_scopes_.begin(); 
+      for (list<NetScope*>::const_iterator scope = root_scopes_.begin();
 	   scope != root_scopes_.end(); scope++)
 	    (*scope)->dump(o);
 
@@ -972,6 +1089,113 @@ void Design::dump(ostream&o) const
 
 /*
  * $Log: design_dump.cc,v $
+ * Revision 1.149  2004/10/04 01:10:52  steve
+ *  Clean up spurious trailing white space.
+ *
+ * Revision 1.148  2004/05/31 23:34:36  steve
+ *  Rewire/generalize parsing an elaboration of
+ *  function return values to allow for better
+ *  speed and more type support.
+ *
+ * Revision 1.147  2004/02/20 06:22:56  steve
+ *  parameter keys are per_strings.
+ *
+ * Revision 1.146  2004/02/18 17:11:54  steve
+ *  Use perm_strings for named langiage items.
+ *
+ * Revision 1.145  2003/12/17 16:52:39  steve
+ *  Debug dumps for synth2.
+ *
+ * Revision 1.144  2003/07/26 03:34:42  steve
+ *  Start handling pad of expressions in code generators.
+ *
+ * Revision 1.143  2003/07/05 20:42:08  steve
+ *  Fix some enumeration warnings.
+ *
+ * Revision 1.142  2003/06/20 00:53:19  steve
+ *  Module attributes from the parser
+ *  through to elaborated form.
+ *
+ * Revision 1.141  2003/06/18 03:55:18  steve
+ *  Add arithmetic shift operators.
+ *
+ * Revision 1.140  2003/05/30 02:55:32  steve
+ *  Support parameters in real expressions and
+ *  as real expressions, and fix multiply and
+ *  divide with real results.
+ *
+ * Revision 1.139  2003/04/22 04:48:29  steve
+ *  Support event names as expressions elements.
+ *
+ * Revision 1.138  2003/03/10 23:40:53  steve
+ *  Keep parameter constants for the ivl_target API.
+ *
+ * Revision 1.137  2003/01/27 05:09:17  steve
+ *  Spelling fixes.
+ *
+ * Revision 1.136  2003/01/26 21:15:58  steve
+ *  Rework expression parsing and elaboration to
+ *  accommodate real/realtime values and expressions.
+ *
+ * Revision 1.135  2002/10/23 01:47:17  steve
+ *  Fix synth2 handling of aset/aclr signals where
+ *  flip-flops are split by begin-end blocks.
+ *
+ * Revision 1.134  2002/10/19 22:59:49  steve
+ *  Redo the parameter vector support to allow
+ *  parameter names in range expressions.
+ *
+ * Revision 1.133  2002/08/19 00:06:11  steve
+ *  Allow release to handle removal of target net.
+ *
+ * Revision 1.132  2002/08/13 05:35:00  steve
+ *  Do not elide named blocks.
+ *
+ * Revision 1.131  2002/08/12 01:34:58  steve
+ *  conditional ident string using autoconfig.
+ *
+ * Revision 1.130  2002/08/04 18:28:14  steve
+ *  Do not use hierarchical names of memories to
+ *  generate vvp labels. -tdll target does not
+ *  used hierarchical name string to look up the
+ *  memory objects in the design.
+ *
+ * Revision 1.129  2002/06/19 04:20:03  steve
+ *  Remove NetTmp and add NetSubnet class.
+ *
+ * Revision 1.128  2002/06/14 21:38:41  steve
+ *  Fix expression width for repeat concatenations.
+ *
+ * Revision 1.127  2002/06/08 23:42:46  steve
+ *  Add NetRamDq synthsesis from memory l-values.
+ *
+ * Revision 1.126  2002/06/05 03:44:25  steve
+ *  Add support for memory words in l-value of
+ *  non-blocking assignments, and remove the special
+ *  NetAssignMem_ and NetAssignMemNB classes.
+ *
+ * Revision 1.125  2002/06/04 05:38:44  steve
+ *  Add support for memory words in l-value of
+ *  blocking assignments, and remove the special
+ *  NetAssignMem class.
+ *
+ * Revision 1.124  2002/05/26 01:39:02  steve
+ *  Carry Verilog 2001 attributes with processes,
+ *  all the way through to the ivl_target API.
+ *
+ *  Divide signal reference counts between rval
+ *  and lval references.
+ *
+ * Revision 1.123  2002/05/05 21:11:49  steve
+ *  Put off evaluation of concatenation repeat expresions
+ *  until after parameters are defined. This allows parms
+ *  to be used in repeat expresions.
+ *
+ *  Add the builtin $signed system function.
+ *
+ * Revision 1.122  2002/03/09 02:10:22  steve
+ *  Add the NetUserFunc netlist node.
+ *
  * Revision 1.121  2001/12/31 00:03:05  steve
  *  Include s indicator in dump of signed numbers.
  *
