@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999-2000 Stephen Williams (steve@icarus.com)
+ * Copyright (c) 1999-2003 Stephen Williams (steve@icarus.com)
  *
  *    This source code is free software; you can redistribute it
  *    and/or modify it in source code form under the terms of the GNU
@@ -16,11 +16,12 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT) && !defined(macintosh)
-#ident "$Id: eval_tree.cc,v 1.33 2002/02/01 05:09:14 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: eval_tree.cc,v 1.62 2004/10/04 01:10:53 steve Exp $"
 #endif
 
 # include "config.h"
+# include "compiler.h"
 
 # include  <iostream>
 
@@ -81,8 +82,22 @@ NetEConst* NetEBBits::eval_tree()
       eval_sub_tree_();
 
       NetEConst*lc = dynamic_cast<NetEConst*>(left_);
-      if (lc == 0) return 0;
       NetEConst*rc = dynamic_cast<NetEConst*>(right_);
+
+	/* Notice the special case where one of the operands is 0 and
+	   this is a bitwise &. If this happens, then the result is
+	   known to be 0. */
+      if ((op() == '&') && lc && (lc->value() == verinum(0))) {
+	    verinum res (verinum::V0, expr_width());
+	    return new NetEConst(res);
+      }
+
+      if ((op() == '&') && rc && (rc->value() == verinum(0))) {
+	    verinum res (verinum::V0, expr_width());
+	    return new NetEConst(res);
+      }
+
+      if (lc == 0) return 0;
       if (rc == 0) return 0;
 
       verinum lval = lc->value();
@@ -135,6 +150,24 @@ NetEConst* NetEBBits::eval_tree()
 		break;
 	  }
 
+	  case '^': {
+		unsigned cnt = lwid;
+		if (cnt > wid)  cnt = wid;
+		if (cnt > rwid) cnt = rwid;
+		for (unsigned idx = 0 ;  idx < cnt ;  idx += 1)
+		      res.set(idx, lval.get(idx) ^ rval.get(idx));
+
+		if (lwid < rwid)
+		      for (unsigned idx = lwid ;  idx < rwid ;  idx += 1)
+			    res.set(idx, rval.get(idx));
+
+		if (rwid < lwid)
+		      for (unsigned idx = rwid ;  idx < lwid ;  idx += 1)
+			    res.set(idx, lval.get(idx));
+
+		break;
+	  }
+
 	  default:
 	    return 0;
       }
@@ -210,22 +243,85 @@ NetEConst* NetEBComp::eval_less_()
 	    return new NetEConst(result);
       }
 
-      if (lv.has_sign() && rv.has_sign() && (lv.as_long() < rv.as_long())) {
-	    verinum result(verinum::V1, 1);
-	    return new NetEConst(result);
-      }
-
-      if (lv.as_ulong() < rv.as_ulong()) {
-	    verinum result(verinum::V1, 1);
-	    return new NetEConst(result);
+      if (lv.has_sign() && rv.has_sign()) {
+	    if (lv.as_long() < rv.as_long()) {
+		  verinum result(verinum::V1, 1);
+		  return new NetEConst(result);
+	    }
+      } else {
+	    if (lv.as_ulong() < rv.as_ulong()) {
+		  verinum result(verinum::V1, 1);
+		  return new NetEConst(result);
+	    }
       }
 
       verinum result(verinum::V0, 1);
       return new NetEConst(result);
 }
 
+NetEConst* NetEBComp::eval_leeq_real_()
+{
+      NetEConst*vtmp;
+      NetECReal*rtmp;
+      double lv, rv;
+
+      switch (left_->expr_type()) {
+	  case ET_REAL:
+	    rtmp = dynamic_cast<NetECReal*> (left_);
+	    if (rtmp == 0)
+		  return 0;
+
+	    lv = rtmp->value().as_double();
+	    break;
+
+	  case ET_VECTOR:
+	    vtmp = dynamic_cast<NetEConst*> (left_);
+	    if (vtmp == 0)
+		  return 0;
+
+	    lv = vtmp->value().as_long();
+	    break;
+
+	  default:
+	    assert(0);
+      }
+
+
+      switch (right_->expr_type()) {
+	  case ET_REAL:
+	    rtmp = dynamic_cast<NetECReal*> (right_);
+	    if (rtmp == 0)
+		  return 0;
+
+	    rv = rtmp->value().as_double();
+	    break;
+
+	  case ET_VECTOR:
+	    vtmp = dynamic_cast<NetEConst*> (right_);
+	    if (vtmp == 0)
+		  return 0;
+
+	    rv = vtmp->value().as_long();
+	    break;
+
+	  default:
+	    assert(0);
+      }
+
+      verinum result((lv <= rv)? verinum::V1 : verinum::V0, 1);
+      vtmp = new NetEConst(result);
+      vtmp->set_line(*this);
+
+      return vtmp;
+}
+
 NetEConst* NetEBComp::eval_leeq_()
 {
+      if (right_->expr_type() == ET_REAL)
+	    return eval_leeq_real_();
+      if (left_->expr_type() == ET_REAL)
+	    return eval_leeq_real_();
+
       NetEConst*r = dynamic_cast<NetEConst*>(right_);
       if (r == 0) return 0;
 
@@ -235,6 +331,12 @@ NetEConst* NetEBComp::eval_leeq_()
 	    return new NetEConst(result);
       }
 
+      if (left_->expr_width() == 0) {
+	    cerr << get_line() << ": internal error: Something wrong "
+		 << "with the left side width of <= ?" << endl;
+	    cerr << get_line() << ":               : " << *this << endl;
+      }
+
 	/* Detect the case where the right side is greater that or
 	   equal to the largest value the left side can possibly
 	   have. */
@@ -254,14 +356,16 @@ NetEConst* NetEBComp::eval_leeq_()
 	    return new NetEConst(result);
       }
 
-      if (lv.has_sign() && rv.has_sign() && (lv.as_long() <= rv.as_long())) {
-	    verinum result(verinum::V1, 1);
-	    return new NetEConst(result);
-      }
-
-      if (lv.as_ulong() <= rv.as_ulong()) {
-	    verinum result(verinum::V1, 1);
-	    return new NetEConst(result);
+      if (lv.has_sign() && rv.has_sign()) {
+	    if (lv.as_long() <= rv.as_long()) {
+		  verinum result(verinum::V1, 1);
+		  return new NetEConst(result);
+	    }
+      } else {
+	    if (lv.as_ulong() <= rv.as_ulong()) {
+		  verinum result(verinum::V1, 1);
+		  return new NetEConst(result);
+	    }
       }
 
       verinum result(verinum::V0, 1);
@@ -270,6 +374,24 @@ NetEConst* NetEBComp::eval_leeq_()
 
 NetEConst* NetEBComp::eval_gt_()
 {
+      if ((left_->expr_type() == NetExpr::ET_REAL)
+	  && (right_->expr_type() == NetExpr::ET_REAL)) {
+
+	    NetECReal*tmpl = dynamic_cast<NetECReal*>(left_);
+	    if (tmpl == 0)
+		  return 0;
+
+	    NetECReal*tmpr = dynamic_cast<NetECReal*>(right_);
+	    if (tmpr == 0)
+		  return 0;
+
+	    double ll = tmpl->value().as_double();
+	    double rr = tmpr->value().as_double();
+
+	    verinum result ((ll > rr)? verinum::V1 : verinum::V0, 1, true);
+	    return new NetEConst(result);
+      }
+
       NetEConst*l = dynamic_cast<NetEConst*>(left_);
       if (l == 0) return 0;
 
@@ -279,19 +401,34 @@ NetEConst* NetEBComp::eval_gt_()
 	    return new NetEConst(result);
       }
 
-	/* Detect the case where the left side is greater than the
-	   largest value the right side can possibly have. */
-      assert(right_->expr_width() > 0);
-      verinum rv (verinum::V1, right_->expr_width());
-      if (lv > rv) {
-	    verinum result(verinum::V1, 1);
-	    return new NetEConst(result);
+	/* Check for the special case where we know, simply by the
+	   limited width of the right expression, that it cannot
+	   possibly be false. */
+      if (right_->expr_width() > 0) {
+	    verinum rv (verinum::V1, right_->expr_width());
+	    if (lv > rv) {
+		  verinum result(verinum::V1, 1);
+		  return new NetEConst(result);
+	    }
+      }
+
+	/* Compare with a real value. Do it as double precision. */
+      if (right_->expr_type() == NetExpr::ET_REAL) {
+	    NetECReal*tmp = dynamic_cast<NetECReal*>(right_);
+	    if (tmp == 0)
+		  return 0;
+
+	     double rr = tmp->value().as_double();
+	     double ll = lv.has_sign()? lv.as_long() : lv.as_ulong();
+
+	     verinum result ((ll > rr)? verinum::V1 : verinum::V0, 1, true);
+	     return new NetEConst(result);
       }
 
 	/* Now go on to the normal test of the values. */
       NetEConst*r = dynamic_cast<NetEConst*>(right_);
       if (r == 0) return 0;
-      rv = r->value();
+      verinum rv = r->value();
       if (! rv.is_defined()) {
 	    verinum result(verinum::Vx, 1);
 	    return new NetEConst(result);
@@ -313,6 +450,24 @@ NetEConst* NetEBComp::eval_gt_()
 
 NetEConst* NetEBComp::eval_gteq_()
 {
+      if ((left_->expr_type() == NetExpr::ET_REAL)
+	  && (right_->expr_type() == NetExpr::ET_REAL)) {
+
+	    NetECReal*tmpl = dynamic_cast<NetECReal*>(left_);
+	    if (tmpl == 0)
+		  return 0;
+
+	    NetECReal*tmpr = dynamic_cast<NetECReal*>(right_);
+	    if (tmpr == 0)
+		  return 0;
+
+	    double ll = tmpl->value().as_double();
+	    double rr = tmpr->value().as_double();
+
+	    verinum result ((ll >= rr)? verinum::V1 : verinum::V0, 1, true);
+	    return new NetEConst(result);
+      }
+
       NetEConst*l = dynamic_cast<NetEConst*>(left_);
       if (l == 0) return 0;
 
@@ -324,17 +479,32 @@ NetEConst* NetEBComp::eval_gteq_()
 
 	/* Detect the case where the left side is greater than the
 	   largest value the right side can possibly have. */
-      assert(right_->expr_width() > 0);
-      verinum rv (verinum::V1, right_->expr_width());
-      if (lv >= rv) {
-	    verinum result(verinum::V1, 1);
-	    return new NetEConst(result);
+      if (right_->expr_type() == NetExpr::ET_VECTOR) {
+	    assert(right_->expr_width() > 0);
+	    verinum rv (verinum::V1, right_->expr_width());
+	    if (lv >= rv) {
+		  verinum result(verinum::V1, 1);
+		  return new NetEConst(result);
+	    }
+      }
+
+	/* Compare with a real value. Do it as double precision. */
+      if (right_->expr_type() == NetExpr::ET_REAL) {
+	    NetECReal*tmp = dynamic_cast<NetECReal*>(right_);
+	    if (tmp == 0)
+		  return 0;
+
+	     double rr = tmp->value().as_double();
+	     double ll = lv.has_sign()? lv.as_long() : lv.as_ulong();
+
+	     verinum result ((ll >= rr)? verinum::V1 : verinum::V0, 1, true);
+	     return new NetEConst(result);
       }
 
 	/* Now go on to the normal test of the values. */
       NetEConst*r = dynamic_cast<NetEConst*>(right_);
       if (r == 0) return 0;
-      rv = r->value();
+      verinum rv = r->value();
       if (! rv.is_defined()) {
 	    verinum result(verinum::Vx, 1);
 	    return new NetEConst(result);
@@ -504,7 +674,7 @@ NetEConst* NetEBComp::eval_tree()
 	  case 'L': // <=
 	    return eval_leeq_();
 
-	  case 'N': // Cse inequality (!==)
+	  case 'N': // Case inequality (!==)
 	    return eval_neeqeq_();
 
 	  case 'n': // not-equal (!=)
@@ -522,27 +692,74 @@ NetEConst* NetEBComp::eval_tree()
 }
 
 /*
- * The NetEBDiv operator includes the / and % opeators. First evaluate
+ * The NetEBDiv operator includes the / and % operators. First evaluate
  * the sub-expressions, then perform the required operation.
  */
-NetEConst* NetEBDiv::eval_tree()
+NetExpr* NetEBDiv::eval_tree()
 {
       eval_sub_tree_();
 
-      NetEConst*lc = dynamic_cast<NetEConst*>(left_);
-      if (lc == 0) return 0;
-      NetEConst*rc = dynamic_cast<NetEConst*>(right_);
-      if (rc == 0) return 0;
+      if (expr_type() == NetExpr::ET_REAL) {
+	    NetECReal*lc = dynamic_cast<NetECReal*>(left_);
+	    if (lc == 0) return 0;
 
-      verinum lval = lc->value();
-      verinum rval = rc->value();
+	    verireal lval = lc->value();
 
-      switch (op_) {
-	  case '/':
-	    return new NetEConst(lval / rval);
+	    if (NetECReal*rc = dynamic_cast<NetECReal*>(right_)) {
+		  NetECReal*tmp = 0;
+		  verireal rval = rc->value();
 
-	  case '%':
-	    return new NetEConst(lval % rval);
+		  switch (op_) {
+		      case '/':
+			tmp = new NetECReal(lval / rval);
+			break;
+
+		      case '%':
+			tmp = new NetECReal(lval % rval);
+		  }
+
+		  assert(tmp);
+		  tmp->set_line(*this);
+		  return tmp;
+
+	    } else if (NetEConst*rc = dynamic_cast<NetEConst*>(right_)) {
+
+		  NetECReal*tmp = 0;
+		  verinum rval = rc->value();
+
+		  switch (op_) {
+		      case '/':
+			tmp = new NetECReal(lval / rval);
+			break;
+
+		      case '%':
+			tmp = new NetECReal(lval % rval);
+		  }
+
+		  assert(tmp);
+		  tmp->set_line(*this);
+		  return tmp;
+
+	    }
+
+
+      } else {
+	    assert(expr_type() == NetExpr::ET_VECTOR);
+	    NetEConst*lc = dynamic_cast<NetEConst*>(left_);
+	    if (lc == 0) return 0;
+	    NetEConst*rc = dynamic_cast<NetEConst*>(right_);
+	    if (rc == 0) return 0;
+
+	    verinum lval = lc->value();
+	    verinum rval = rc->value();
+
+	    switch (op_) {
+		case '/':
+		  return new NetEConst(lval / rval);
+
+		case '%':
+		  return new NetEConst(lval % rval);
+	    }
       }
 
       return 0;
@@ -574,7 +791,7 @@ NetEConst* NetEBLogic::eval_tree()
 	    if (v.get(idx) == verinum::V1)
 		  rv = verinum::V1;
 
-      if (lv == verinum::V0)
+      if (rv == verinum::V0)
 	    for (unsigned idx = 0 ;  idx < v.len() ;  idx += 1)
 		  if (v.get(idx) != verinum::V0)
 			rv = verinum::Vx;
@@ -614,10 +831,66 @@ NetEConst* NetEBLogic::eval_tree()
       return new NetEConst(verinum(res, 1));
 }
 
-NetEConst* NetEBMult::eval_tree()
+NetExpr* NetEBMult::eval_tree_real_()
+{
+      verireal lval;
+      verireal rval;
+
+      switch (left_->expr_type()) {
+	  case ET_REAL: {
+		NetECReal*lc = dynamic_cast<NetECReal*> (left_);
+		if (lc == 0) return 0;
+		lval = lc->value();
+		break;
+	  }
+
+	  case ET_VECTOR: {
+		NetEConst*lc = dynamic_cast<NetEConst*>(left_);
+		if (lc == 0) return 0;
+		verinum tmp = lc->value();
+		lval = verireal(tmp.as_long());
+		break;
+	  }
+
+	  default:
+	    assert(0);
+      }
+
+      switch (right_->expr_type()) {
+	  case ET_REAL: {
+		NetECReal*rc = dynamic_cast<NetECReal*> (right_);
+		if (rc == 0) return 0;
+		rval = rc->value();
+		break;
+	  }
+
+	  case ET_VECTOR: {
+		NetEConst*rc = dynamic_cast<NetEConst*>(right_);
+		if (rc == 0) return 0;
+		verinum tmp = rc->value();
+		rval = verireal(tmp.as_long());
+		break;
+	  }
+
+	  default:
+	    assert(0);
+      }
+
+
+      NetECReal*res = new NetECReal(lval * rval);
+      res->set_line(*this);
+      return res;
+}
+
+NetExpr* NetEBMult::eval_tree()
 {
       eval_sub_tree_();
 
+      if (expr_type() == ET_REAL)
+	    return eval_tree_real_();
+
+      assert(expr_type() == ET_VECTOR);
+
       NetEConst*lc = dynamic_cast<NetEConst*>(left_);
       if (lc == 0) return 0;
       NetEConst*rc = dynamic_cast<NetEConst*>(right_);
@@ -649,19 +922,26 @@ NetEConst* NetEBShift::eval_tree()
       verinum rv = re->value();
       verinum lv = le->value();
 
-	/* Calculate the width of the result. If it is not fixed, then
-	   get it from the left operand. */
+	/* Make an early estimate of the expression width. */
       unsigned wid = expr_width();
-      if (wid == 0)
-	    wid = left_->expr_width();
 
       if (rv.is_defined()) {
 
 	    unsigned shift = rv.as_ulong();
 
+	    if ((wid == 0) || ! lv.has_len()) {
+		    /* If the caller doesn't care what the width is,
+		       then calcuate a width from the trimmed left
+		       expression, plus the shift. This avoids
+		       data loss. */
+		  lv = trim_vnum(lv);
+		  wid = lv.len();
+		  if (op() == 'l')
+			wid = lv.len() + shift;
+	    }
 
 	    assert(wid);
-	    verinum nv (verinum::V0, wid);
+	    verinum nv (verinum::V0, wid, lv.has_len());
 
 	    if (op() == 'r') {
 		  unsigned cnt = wid;
@@ -690,7 +970,9 @@ NetEConst* NetEBShift::eval_tree()
 	    res = new NetEConst(nv);
 
       } else {
-	    assert(wid);
+	    if (wid == 0)
+		  wid = left_->expr_width();
+
 	    verinum nv (verinum::Vx, wid);
 	    res = new NetEConst(nv);
       }
@@ -700,6 +982,10 @@ NetEConst* NetEBShift::eval_tree()
 
 NetEConst* NetEConcat::eval_tree()
 {
+      unsigned repeat_val = repeat();
+      unsigned local_errors = 0;
+
+      unsigned gap = 0;
       for (unsigned idx = 0 ;  idx < parms_.count() ;  idx += 1) {
 
 	      // Parameter not here? This is an error, but presumably
@@ -707,10 +993,13 @@ NetEConst* NetEConcat::eval_tree()
 	    if (parms_[idx] == 0)
 		  continue;
 
+
 	      // If this parameter is already a constant, all is well
 	      // so go on.
-	    if (dynamic_cast<NetEConst*>(parms_[idx]))
+	    if (dynamic_cast<NetEConst*>(parms_[idx])) {
+		  gap += parms_[idx]->expr_width();
 		  continue;
+	    }
 
 	      // Finally, try to evaluate the parameter expression
 	      // that is here. If I succeed, reset the parameter to
@@ -720,37 +1009,64 @@ NetEConst* NetEConcat::eval_tree()
 	    if (expr) {
 		  delete parms_[idx];
 		  parms_[idx] = expr;
+
+		  if (! expr->has_width()) {
+			cerr << get_line() << ": error: concatenation "
+			     << "operand has indefinite width: "
+			     << *parms_[idx] << endl;
+			local_errors += 1;
+		  } else if (expr->expr_width() == 0) {
+			cerr << expr->get_line() << ": internal error: "
+			     << "Operand of concatenation has no width: "
+			     << *expr << endl;
+			local_errors += 1;
+		  }
+
+		  gap += expr->expr_width();
 	    }
+
       }
 
+      if (local_errors > 0)
+	    return 0;
+
 	// Handle the special case that the repeat expression is
 	// zero. In this case, just return a 0 value with the expected
 	// width.
-      if (repeat_ == 0) {
+      if (repeat_val == 0) {
 	    verinum val (verinum::V0, expr_width());
 	    NetEConst*res = new NetEConst(val);
 	    res->set_width(val.len());
 	    return res;
       }
 
-	// Figure out the width of the repeated expression, and make a
-	// verinum to hold the result.
-      unsigned gap = expr_width() / repeat_;
-      verinum val (verinum::Vx, repeat_ * gap);
-
+	// At this point, the "gap" is the width of a single repeat of
+	// the concatenation. The total width of the result is the gap
+	// times the repeat count.
+      verinum val (verinum::Vx, repeat_val * gap);
 
 	// build up the result from least significant to most.
 
       unsigned cur = 0;
+      bool is_string_flag = true;
       for (unsigned idx = parms_.count() ;  idx > 0 ;  idx -= 1) {
 	    NetEConst*expr = dynamic_cast<NetEConst*>(parms_[idx-1]);
 	    if (expr == 0)
 		  return 0;
 
 	    verinum tmp = expr->value();
-	    for (unsigned bit = 0 ;  bit < tmp.len() ;  bit += 1, cur += 1)
-		  for (unsigned rep = 0 ;  rep < repeat_ ;  rep += 1)
+	    for (unsigned bit = 0;  bit < tmp.len(); bit += 1, cur += 1)
+		  for (unsigned rep = 0 ;  rep < repeat_val ;  rep += 1)
 			val.set(rep*gap+cur, tmp[bit]);
+
+	    is_string_flag = is_string_flag && tmp.is_string();
+      }
+
+	/* If all the values were strings, then re-stringify this
+	   constant. This might be useful information in the code
+	   generator or other optimizer steps. */
+      if (is_string_flag) {
+	    val = verinum(val.as_string());
       }
 
       NetEConst*res = new NetEConst(val);
@@ -808,8 +1124,7 @@ NetExpr* NetEParam::eval_tree()
 	    return 0;
 
       assert(scope_);
-      assert(name_.peek_name(1) == 0);
-      const NetExpr*expr = scope_->get_parameter(name_.peek_name(0));
+      const NetExpr*expr = scope_->get_parameter(name_);
       if (expr == 0) {
 	    cerr << get_line() << ": internal error: Unable to match "
 		 << "parameter " << name_ << " in scope "
@@ -824,23 +1139,129 @@ NetExpr* NetEParam::eval_tree()
 
 	// If the parameter that I refer to is already evaluated, then
 	// return the constant value.
-      if (dynamic_cast<NetEConst*>(nexpr))
-	    return nexpr;
+      if (NetEConst*tmp = dynamic_cast<NetEConst*>(nexpr)) {
+	    verinum val = tmp->value();
+	    NetEConstParam*ptmp = new NetEConstParam(scope_, name_, val);
+	    ptmp->set_line(*this);
+	    delete nexpr;
+	    return ptmp;
+      }
+
+      if (NetECReal*tmp = dynamic_cast<NetECReal*>(nexpr)) {
+	    verireal val = tmp->value();
+	    NetECRealParam*ptmp = new NetECRealParam(scope_, name_, val);
+	    ptmp->set_line(*this);
+	    delete nexpr;
+	    return ptmp;
+      }
 
 	// Try to evaluate the expression. If I cannot, then the
 	// expression is not a constant expression and I fail here.
       NetExpr*res = nexpr->eval_tree();
       if (res == 0) {
+	    cerr << get_line() << ": internal error: Unable to evaluate "
+		 << "parameter " << name_ << " expression: "
+		 << *nexpr << endl;
 	    delete nexpr;
 	    return 0;
       }
 
 	// The result can be saved as the value of the parameter for
 	// future reference, and return a copy to the caller.
-      scope_->set_parameter(name_.peek_name(0), res);
-      return res->dup_expr();
+      scope_->replace_parameter(name_, res);
+
+	/* Return as a result a NetEConstParam or NetECRealParam
+	   object, depending on the type of the expression. */
+
+      switch (res->expr_type()) {
+
+	  case NetExpr::ET_VECTOR:
+	    { NetEConst*tmp = dynamic_cast<NetEConst*>(res);
+	      if (tmp == 0) {
+		    cerr << get_line() << ": internal error: parameter "
+			 << name_ << " evaluates to incomprehensible "
+			 << *res << "." << endl;
+		    return 0;
+	      }
+
+	      assert(tmp);
+
+	      verinum val = tmp->value();
+	      NetEConstParam*ptmp = new NetEConstParam(scope_, name_, val);
+
+	      return ptmp;
+	    }
+
+	  case NetExpr::ET_REAL:
+	    { NetECReal*tmp = dynamic_cast<NetECReal*>(res);
+	      if (tmp == 0) {
+		    cerr << get_line() << ": internal error: parameter "
+			 << name_ << " evaluates to incomprehensible "
+			 << *res << "." << endl;
+		    return 0;
+	      }
+
+	      assert(tmp);
+
+	      verireal val = tmp->value();
+	      NetECRealParam*ptmp = new NetECRealParam(scope_, name_, val);
+
+	      return ptmp;
+	    }
+
+	  default:
+	    assert(0);
+	    return 0;
+      }
 }
 
+NetEConst* NetESelect::eval_tree()
+{
+      NetEConst*expr = dynamic_cast<NetEConst*>(expr_);
+      if (expr == 0) {
+	    NetExpr*tmp = expr_->eval_tree();
+	    if (tmp != 0) {
+		  delete expr_;
+		  expr_ = tmp;
+	    }
+
+	    expr = dynamic_cast<NetEConst*>(expr_);
+      }
+
+      NetEConst*base = dynamic_cast<NetEConst*>(base_);
+      if (base == 0) {
+	    NetExpr*tmp = base_->eval_tree();
+	    if (tmp != 0) {
+		  delete base_;
+		  base_ = tmp;
+	    }
+
+	    base = dynamic_cast<NetEConst*>(base_);
+      }
+
+      if (expr == 0)
+	    return 0;
+      if (base == 0)
+	    return 0;
+
+      verinum eval = expr->value();
+      verinum oval (verinum::V0, expr_width(), true);
+      long bval = base->value().as_long();
+
+      for (unsigned long idx = 0 ;  idx < expr_width() ;  idx += 1) {
+	    if ((bval >= 0) && ((unsigned long) bval < eval.len()))
+		  oval.set(idx, eval.get(bval));
+	    else
+		  oval.set(idx, verinum::Vx);
+
+	    bval += 1;
+      }
+
+      NetEConst*res = new NetEConst(oval);
+      return res;
+}
+
+
 /*
  * A ternary expression evaluation is controlled by the condition
  * expression. If the condition evaluates to true or false, then
@@ -884,18 +1305,50 @@ NetExpr* NetETernary::eval_tree()
       if (c == 0)
 	    return 0;
 
+	/* Check the boolean value of the constant condition
+	   expression. Note that the X case is handled explicitly, so
+	   we must differentiate. */
+
+      verinum cond_value = c->value();
+      bool true_flag = false;
+      bool x_flag = false;
+
+      for (unsigned idx = 0 ;  idx < cond_value.len() ;  idx += 1) {
+	    switch (cond_value.get(idx)) {
+		case verinum::V1:
+		  true_flag = true;
+		  break;
+		case verinum::V0:
+		  break;
+		default:
+		  x_flag = true;
+	    }
+      }
+
 
 	/* If the condition is 1 or 0, return the true or false
 	   expression. Try to evaluate the expression down as far as
 	   we can. */
 
-      if (c->value().get(0) == verinum::V1)
+      if (true_flag) {
+	    if (debug_eval_tree) {
+		  cerr << get_line() << ": debug: Evaluate ternary with "
+		       << "constant condition value: " << c->value() << endl;
+		  cerr << get_line() << ":      : Selecting true case: "
+		       << *true_val_ << endl;
+	    }
 	    return true_val_->dup_expr();
+      }
 
-
-      if (c->value().get(0) == verinum::V0)
+      if (! x_flag) {
+	    if (debug_eval_tree) {
+		  cerr << get_line() << ": debug: Evaluate ternary with "
+		       << "constant condition value: " << c->value() << endl;
+		  cerr << get_line() << ":      : Selecting false case: "
+		       << *true_val_ << endl;
+	    }
 	    return false_val_->dup_expr();
-
+      }
 
 	/* Here we have a more complex case. We need to evaluate both
 	   expressions down to constants then compare the values to
@@ -911,13 +1364,15 @@ NetExpr* NetETernary::eval_tree()
 	    return 0;
 
 
-      unsigned size = t->expr_width();
-      assert(size == f->expr_width());
+      unsigned tsize = t->expr_width();
+      unsigned fsize = f->expr_width();
+	/* Size of the result is the size of the widest operand. */
+      unsigned rsize = tsize > fsize? tsize : fsize;
 
-      verinum val (verinum::V0, size);
-      for (unsigned idx = 0 ;  idx < size ;  idx += 1) {
-	    verinum::V tv = t->value().get(idx);
-	    verinum::V fv = f->value().get(idx);
+      verinum val (verinum::V0, rsize);
+      for (unsigned idx = 0 ;  idx < rsize ;  idx += 1) {
+	    verinum::V tv = idx < tsize? t->value().get(idx) : verinum::V0;
+	    verinum::V fv = idx < rsize? f->value().get(idx) : verinum::V0;
 
 	    if (tv == fv)
 		  val.set(idx, tv);
@@ -925,6 +1380,13 @@ NetExpr* NetETernary::eval_tree()
 		  val.set(idx, verinum::Vx);
       }
 
+      if (debug_eval_tree) {
+	    cerr << get_line() << ": debug: Evaluate ternary with "
+		 << "constant condition value: " << c->value() << endl;
+	    cerr << get_line() << ":      : Blending cases to get "
+		 << val << endl;
+      }
+
       NetEConst*rc = new NetEConst(val);
       rc->set_line(*this);
       return rc;
@@ -961,7 +1423,7 @@ NetEConst* NetEUnary::eval_tree()
 
 	  case '-': {
 		if (val.is_defined()) {
-		      
+
 		      verinum tmp (verinum::V0, val.len());
 		      tmp.has_sign(val.has_sign());
 		      val = tmp - val;
@@ -993,6 +1455,8 @@ NetEConst* NetEUnary::eval_tree()
 		return new NetEConst(val);
 	  }
 
+	  case '!':
+	    assert(0);
 	  default:
 	    return 0;
       }
@@ -1038,6 +1502,7 @@ NetEConst* NetEUReduce::eval_tree()
 		}
 
 		res = v1? verinum::V0 : (vx? verinum::Vx : verinum::V1);
+		break;
 	  }
 
 	  case '&': {
@@ -1054,6 +1519,28 @@ NetEConst* NetEUReduce::eval_tree()
 		break;
 	  }
 
+	  case '^': {
+		  /* Reduction XOR. */
+		unsigned ones = 0, unknown = 0;
+		for (unsigned idx = 0 ;  idx < val.len() ;  idx += 1)
+		      switch (val.get(idx)) {
+			  case verinum::V0:
+			    break;
+			  case verinum::V1:
+			    ones += 1;
+			    break;
+			  default:
+			    unknown += 1;
+			    break;
+		      }
+
+		if (unknown)
+		      return new NetEConst(verinum(verinum::Vx,1,true));
+		if (ones%2)
+		      return new NetEConst(verinum(verinum::V1,1,true));
+		return new NetEConst(verinum(verinum::V0,1,true));
+	  }
+
 	  default:
 	    return 0;
       }
@@ -1064,119 +1551,46 @@ NetEConst* NetEUReduce::eval_tree()
 
 /*
  * $Log: eval_tree.cc,v $
- * Revision 1.33  2002/02/01 05:09:14  steve
- *  Propagate sign in unary minus.
+ * Revision 1.62  2004/10/04 01:10:53  steve
+ *  Clean up spurious trailing white space.
  *
- * Revision 1.32  2002/01/22 01:40:04  steve
- *  Precalculate constant results of memory index expressions.
+ * Revision 1.61  2004/09/10 23:51:42  steve
+ *  Fix the evaluation of constant ternary expressions.
  *
- * Revision 1.31  2001/12/30 00:39:25  steve
- *  Evaluate constant unary minus.
+ * Revision 1.60  2004/02/20 06:22:56  steve
+ *  parameter keys are per_strings.
  *
- * Revision 1.30  2001/12/03 04:47:15  steve
- *  Parser and pform use hierarchical names as hname_t
- *  objects instead of encoded strings.
+ * Revision 1.59  2003/10/31 02:47:11  steve
+ *  NetEUReduce has its own dup_expr method.
  *
- * Revision 1.29  2001/11/19 01:54:14  steve
- *  Port close cropping behavior from mcrgb
- *  Move window array reset to libmc.
+ * Revision 1.58  2003/10/26 04:54:56  steve
+ *  Support constant evaluation of binary ^ operator.
  *
- * Revision 1.28  2001/10/22 15:31:21  steve
- *  fix constant overrun in | operands.
+ * Revision 1.57  2003/09/04 01:52:50  steve
+ *  Evaluate real parameter expressions that contain real parameters.
  *
- * Revision 1.27  2001/10/20 00:11:24  steve
- *  Evaluate constant == when operands differ in width.
+ * Revision 1.56  2003/08/01 02:12:30  steve
+ *  Fix || with true case on the right.
  *
- * Revision 1.26  2001/07/25 03:10:49  steve
- *  Create a config.h.in file to hold all the config
- *  junk, and support gcc 3.0. (Stephan Boettcher)
+ * Revision 1.55  2003/06/24 01:38:02  steve
+ *  Various warnings fixed.
  *
- * Revision 1.25  2001/07/16 20:41:41  steve
- *  Handle 0 counts in repeat concatenation.
+ * Revision 1.54  2003/06/05 04:28:24  steve
+ *  Evaluate <= with real operands.
  *
- * Revision 1.24  2001/02/10 20:29:39  steve
- *  In the context of range declarations, use elab_and_eval instead
- *  of the less robust eval_const methods.
+ * Revision 1.53  2003/06/04 01:26:17  steve
+ *  internal error for <= expression errors.
  *
- * Revision 1.23  2001/02/09 05:44:23  steve
- *  support evaluation of constant < in expressions.
+ * Revision 1.52  2003/05/30 02:55:32  steve
+ *  Support parameters in real expressions and
+ *  as real expressions, and fix multiply and
+ *  divide with real results.
  *
- * Revision 1.22  2001/02/07 02:46:31  steve
- *  Support constant evaluation of / and % (PR#124)
- *
- * Revision 1.21  2001/01/14 23:04:56  steve
- *  Generalize the evaluation of floating point delays, and
- *  get it working with delay assignment statements.
- *
- *  Allow parameters to be referenced by hierarchical name.
- *
- * Revision 1.20  2001/01/04 16:49:50  steve
- *  Evaluate constant === and !== expressions.
- *
- * Revision 1.19  2001/01/04 04:28:42  steve
- *  Evaluate constant logical && and ||.
- *
- * Revision 1.18  2001/01/02 04:21:14  steve
- *  Support a bunch of unary operators in parameter expressions.
- *
- * Revision 1.17  2001/01/02 03:23:40  steve
- *  Evaluate constant &, | and unary ~.
- *
- * Revision 1.16  2001/01/01 21:49:33  steve
- *  Fix shift and ternary operators in parameter expressions (PR#86)
- *
- * Revision 1.15  2000/12/16 20:00:17  steve
- *  Handle non-constant l-values.
- *
- * Revision 1.14  2000/12/16 19:03:30  steve
- *  Evaluate <= and ?: in parameter expressions (PR#81)
- *
- * Revision 1.13  2000/09/29 04:42:56  steve
- *  Cnstant evaluation of NE.
- *
- * Revision 1.12  2000/09/27 18:28:37  steve
- *  multiply in parameter expressions.
- *
- * Revision 1.11  2000/07/07 04:53:54  steve
- *  Add support for non-constant delays in delay statements,
- *  Support evaluating ! in constant expressions, and
- *  move some code from netlist.cc to net_proc.cc.
- *
- * Revision 1.10  2000/04/28 18:43:23  steve
- *  integer division in expressions properly get width.
- *
- * Revision 1.9  2000/03/08 04:36:53  steve
- *  Redesign the implementation of scopes and parameters.
- *  I now generate the scopes and notice the parameters
- *  in a separate pass over the pform. Once the scopes
- *  are generated, I can process overrides and evalutate
- *  paremeters before elaboration begins.
- *
- * Revision 1.8  2000/02/23 02:56:54  steve
- *  Macintosh compilers do not support ident.
- *
- * Revision 1.7  2000/01/13 03:35:35  steve
- *  Multiplication all the way to simulation.
- *
- * Revision 1.6  1999/10/22 23:57:53  steve
- *  do the <= in bits, not numbers.
- *
- * Revision 1.5  1999/10/10 23:29:37  steve
- *  Support evaluating + operator at compile time.
- *
- * Revision 1.4  1999/09/23 03:56:57  steve
- *  Support shift operators.
- *
- * Revision 1.3  1999/09/23 00:21:54  steve
- *  Move set_width methods into a single file,
- *  Add the NetEBLogic class for logic expressions,
- *  Fix error setting with of && in if statements.
- *
- * Revision 1.2  1999/09/21 00:13:40  steve
- *  Support parameters that reference other paramters.
- *
- * Revision 1.1  1999/09/20 02:21:10  steve
- *  Elaborate parameters in phases.
+ * Revision 1.51  2003/04/15 05:06:56  steve
+ *  Handle real constants evaluation > and >=.
  *
+ * Revision 1.50  2003/04/14 03:40:21  steve
+ *  Make some effort to preserve bits while
+ *  operating on constant values.
  */
 
