@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999-2000 Stephen Williams (steve@icarus.com)
+ * Copyright (c) 1999-2004 Stephen Williams (steve@icarus.com)
  *
  *    This source code is free software; you can redistribute it
  *    and/or modify it in source code form under the terms of the GNU
@@ -16,8 +16,8 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT) && !defined(macintosh)
-#ident "$Id: expr_synth.cc,v 1.31 2001/12/30 17:06:52 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: expr_synth.cc,v 1.59 2004/06/30 02:16:26 steve Exp $"
 #endif
 
 # include "config.h"
@@ -44,14 +44,21 @@ NetNet* NetEBAdd::synthesize(Design*des)
 
       NetNet*lsig = left_->synthesize(des);
       NetNet*rsig = right_->synthesize(des);
-      
+
+      assert(expr_width() >= lsig->pin_count());
+      assert(expr_width() >= rsig->pin_count());
+
+      lsig = pad_to_width(des, lsig, expr_width());
+      rsig = pad_to_width(des, rsig, expr_width());
+
       assert(lsig->pin_count() == rsig->pin_count());
       unsigned width=lsig->pin_count();
 
-      string path = lsig->scope()->name()+"."+lsig->scope()->local_symbol();
+      perm_string path = lsig->scope()->local_symbol();
       NetNet*osig = new NetNet(lsig->scope(), path, NetNet::IMPLICIT, width);
+      osig->local_flag(true);
 
-      string oname = des->local_symbol(path);
+      perm_string oname = osig->scope()->local_symbol();
       NetAddSub *adder = new NetAddSub(lsig->scope(), oname, width);
       for (unsigned idx = 0 ;  idx < width;  idx += 1) {
 	    connect(lsig->pin(idx), adder->pin_DataA(idx));
@@ -62,10 +69,10 @@ NetNet* NetEBAdd::synthesize(Design*des)
 
       switch (op()) {
 	  case '+':
-	    adder->attribute("LPM_Direction", "ADD");
+	    adder->attribute(perm_string::literal("LPM_Direction"), verinum("ADD"));
 	    break;
 	  case '-':
-	    adder->attribute("LPM_Direction", "SUB");
+	    adder->attribute(perm_string::literal("LPM_Direction"), verinum("SUB"));
 	    break;
       }
 
@@ -99,14 +106,34 @@ NetNet* NetEBBits::synthesize(Design*des)
       }
 
       assert(lsig->pin_count() == rsig->pin_count());
-      NetNet*osig = new NetNet(scope, path, NetNet::IMPLICIT,
-			       lsig->pin_count());
+      NetNet*osig = new NetNet(scope, scope->local_symbol(),
+			       NetNet::IMPLICIT, lsig->pin_count());
       osig->local_flag(true);
 
       for (unsigned idx = 0 ;  idx < osig->pin_count() ;  idx += 1) {
-	    string oname = des->local_symbol(path);
+	    perm_string oname = scope->local_symbol();
 	    NetLogic*gate;
 
+	      /* If the rsig bit is constant, then look for special
+		 cases that I can use to reduce the generated
+		 logic. If I find one, then handle it immediately and
+		 skip the rest of the processing of this bit. */
+	    if (rsig->pin(idx).nexus()->drivers_constant()) {
+		  verinum::V bval = rsig->pin(idx).nexus()->driven_value();
+
+		    /* (A & 0) is (0) */
+		  if ((op() == '&') && bval == verinum::V0) {
+			connect(osig->pin(idx), rsig->pin(idx));
+			continue;
+		  }
+
+		    /* (A & 1) is A */
+		  if ((op() == '&') && bval == verinum::V1) {
+			connect(osig->pin(idx), lsig->pin(idx));
+			continue;
+		  }
+	    }
+
 	    switch (op()) {
 		case '&':
 		  gate = new NetLogic(scope, oname, 3, NetLogic::AND);
@@ -154,23 +181,50 @@ NetNet* NetEBComp::synthesize(Design*des)
 		  : right_->synthesize(des);
 	    NetScope*scope = lsig->scope();
 	    assert(scope);
-	    string path = des->local_symbol(scope->name());
 
-	    NetNet*osig = new NetNet(scope, path, NetNet::IMPLICIT, 1);
+	    NetNet*osig = new NetNet(scope, scope->local_symbol(),
+				     NetNet::IMPLICIT, 1);
 	    osig->local_flag(true);
 
 	    NetLogic*gate;
 	    switch (op_) {
 		case 'e':
 		case 'E':
-		  gate = new NetLogic(scope, des->local_symbol(path),
+		  gate = new NetLogic(scope, scope->local_symbol(),
 				      lsig->pin_count()+1, NetLogic::NOR);
 		  break;
 		case 'n':
 		case 'N':
-		  gate = new NetLogic(scope, des->local_symbol(path),
+		  gate = new NetLogic(scope, scope->local_symbol(),
 				      lsig->pin_count()+1, NetLogic::OR);
 		  break;
+
+		case '>':
+		    /* sig > 0 is true if any bit in sig is set. This
+		       is very much like sig != 0. (0 > sig) shouldn't
+		       happen. */
+		  if (rcon) {
+			gate = new NetLogic(scope, scope->local_symbol(),
+					    lsig->pin_count()+1, NetLogic::OR);
+		  } else {
+			assert(0);
+			gate = new NetLogic(scope, scope->local_symbol(),
+				      lsig->pin_count()+1, NetLogic::NOR);
+		  }
+		  break;
+
+		case '<':
+		    /* 0 < sig is handled like sig > 0. */
+		  if (! rcon) {
+			gate = new NetLogic(scope, scope->local_symbol(),
+					    lsig->pin_count()+1, NetLogic::OR);
+		  } else {
+			assert(0);
+			gate = new NetLogic(scope, scope->local_symbol(),
+				      lsig->pin_count()+1, NetLogic::NOR);
+		  }
+		  break;
+
 		default:
 		  assert(0);
 	    }
@@ -188,7 +242,6 @@ NetNet* NetEBComp::synthesize(Design*des)
 
       NetScope*scope = lsig->scope();
       assert(scope);
-      string path = des->local_symbol(scope->name());
 
       unsigned width = lsig->pin_count();
       if (rsig->pin_count() > lsig->pin_count())
@@ -197,13 +250,14 @@ NetNet* NetEBComp::synthesize(Design*des)
       lsig = pad_to_width(des, lsig, width);
       rsig = pad_to_width(des, rsig, width);
 
-      NetNet*osig = new NetNet(scope, path, NetNet::IMPLICIT, 1);
+      NetNet*osig = new NetNet(scope, scope->local_symbol(),
+			       NetNet::IMPLICIT, 1);
       osig->local_flag(true);
 
 	/* Handle the special case of a single bit equality
 	   operation. Make an XNOR gate instead of a comparator. */
       if ((width == 1) && ((op_ == 'e') || (op_ == 'E'))) {
-	    NetLogic*gate = new NetLogic(scope, des->local_symbol(path),
+	    NetLogic*gate = new NetLogic(scope, scope->local_symbol(),
 					 3, NetLogic::XNOR);
 	    connect(gate->pin(0), osig->pin(0));
 	    connect(gate->pin(1), lsig->pin(0));
@@ -216,7 +270,7 @@ NetNet* NetEBComp::synthesize(Design*des)
 	   operation. This is similar to single bit equality, but uses
 	   an XOR instead of an XNOR gate. */
       if ((width == 1) && ((op_ == 'n') || (op_ == 'N'))) {
-	    NetLogic*gate = new NetLogic(scope, des->local_symbol(path),
+	    NetLogic*gate = new NetLogic(scope, scope->local_symbol(),
 					 3, NetLogic::XOR);
 	    connect(gate->pin(0), osig->pin(0));
 	    connect(gate->pin(1), lsig->pin(0));
@@ -226,7 +280,7 @@ NetNet* NetEBComp::synthesize(Design*des)
       }
 
 
-      NetCompare*dev = new NetCompare(scope, des->local_symbol(path), width);
+      NetCompare*dev = new NetCompare(scope, scope->local_symbol(), width);
       des->add_node(dev);
 
       for (unsigned idx = 0 ;  idx < lsig->pin_count() ;  idx += 1)
@@ -268,12 +322,102 @@ NetNet* NetEBComp::synthesize(Design*des)
       return osig;
 }
 
+NetNet* NetEBMult::synthesize(Design*des)
+{
+      NetNet*lsig = left_->synthesize(des);
+      NetNet*rsig = right_->synthesize(des);
+
+      if (lsig == 0)
+	    return 0;
+
+      if (rsig == 0)
+	    return 0;
+
+      NetScope*scope = lsig->scope();
+      assert(scope);
+
+      NetMult*mult = new NetMult(scope, scope->local_symbol(),
+				 expr_width(),
+				 lsig->pin_count(),
+				 rsig->pin_count());
+      des->add_node(mult);
+
+      mult->set_signed( has_sign() );
+      mult->set_line(*this);
+
+      for (unsigned idx = 0 ;  idx < lsig->pin_count() ;  idx += 1)
+	    connect(mult->pin_DataA(idx), lsig->pin(idx));
+
+      for (unsigned idx = 0 ;  idx < rsig->pin_count() ;  idx += 1)
+	    connect(mult->pin_DataB(idx), rsig->pin(idx));
+
+      NetNet*osig = new NetNet(scope, scope->local_symbol(),
+			       NetNet::IMPLICIT, expr_width());
+      osig->local_flag(true);
+
+      for (unsigned idx = 0 ;  idx < osig->pin_count() ;  idx += 1)
+	    connect(mult->pin_Result(idx), osig->pin(idx));
+
+      return osig;
+}
+
 NetNet* NetEBDiv::synthesize(Design*des)
 {
-      cerr << get_line() << ": internal error: cannot synthesize division: "
-	   << *this << endl;
-      des->errors += 1;
-      return 0;
+      NetNet*lsig = left_->synthesize(des);
+      NetNet*rsig  = right_->synthesize(des);
+
+      NetScope*scope = lsig->scope();
+
+      NetNet*osig = new NetNet(scope, scope->local_symbol(),
+			       NetNet::IMPLICIT, expr_width());
+      osig->local_flag(true);
+
+      switch (op()) {
+
+	  case '/': {
+		NetDivide*div = new NetDivide(scope, scope->local_symbol(),
+					      expr_width(),
+					      lsig->pin_count(),
+					      rsig->pin_count());
+		des->add_node(div);
+
+		for (unsigned idx = 0 ;  idx < lsig->pin_count() ;  idx += 1)
+		      connect(div->pin_DataA(idx), lsig->pin(idx));
+		for (unsigned idx = 0 ;  idx < rsig->pin_count() ;  idx += 1)
+		      connect(div->pin_DataB(idx), rsig->pin(idx));
+		for (unsigned idx = 0 ;  idx < osig->pin_count() ;  idx += 1)
+		      connect(div->pin_Result(idx), osig->pin(idx));
+		break;
+	  }
+
+	  case '%': {
+		NetModulo*div = new NetModulo(scope, scope->local_symbol(),
+					      expr_width(),
+					      lsig->pin_count(),
+					      rsig->pin_count());
+		des->add_node(div);
+
+		for (unsigned idx = 0 ;  idx < lsig->pin_count() ;  idx += 1)
+		      connect(div->pin_DataA(idx), lsig->pin(idx));
+		for (unsigned idx = 0 ;  idx < rsig->pin_count() ;  idx += 1)
+		      connect(div->pin_DataB(idx), rsig->pin(idx));
+		for (unsigned idx = 0 ;  idx < osig->pin_count() ;  idx += 1)
+		      connect(div->pin_Result(idx), osig->pin(idx));
+		break;
+	  }
+
+	  default: {
+		cerr << get_line() << ": internal error: "
+		     << "NetEBDiv has unexpeced op() code: "
+		     << op() << endl;
+		des->errors += 1;
+
+		delete osig;
+		return 0;
+	  }
+      }
+
+      return osig;
 }
 
 NetNet* NetEBLogic::synthesize(Design*des)
@@ -289,9 +433,9 @@ NetNet* NetEBLogic::synthesize(Design*des)
 
       NetScope*scope = lsig->scope();
       assert(scope);
-      string path = des->local_symbol(scope->name());
 
-      NetNet*osig = new NetNet(scope, path, NetNet::IMPLICIT, 1);
+      NetNet*osig = new NetNet(scope, scope->local_symbol(),
+			       NetNet::IMPLICIT, 1);
       osig->local_flag(true);
 
 
@@ -301,12 +445,11 @@ NetNet* NetEBLogic::synthesize(Design*des)
 		 comparison with a single wide OR gate. So handle this
 		 magically. */
 
-	    string oname = des->local_symbol(path);
-	    NetLogic*olog;
+	    perm_string oname = scope->local_symbol();
 
-	    olog = new NetLogic(scope, oname,
-				lsig->pin_count()+rsig->pin_count()+1,
-				NetLogic::OR);
+	    NetLogic*olog = new NetLogic(scope, oname,
+					 lsig->pin_count()+rsig->pin_count()+1,
+					 NetLogic::OR);
 
 	    connect(osig->pin(0), olog->pin(0));
 
@@ -326,7 +469,7 @@ NetNet* NetEBLogic::synthesize(Design*des)
 	      /* Create the logic AND gate. This is a single bit
 		 output, with inputs for each of the operands. */
 	    NetLogic*olog;
-	    string oname = des->local_symbol(path);
+	    perm_string oname = scope->local_symbol();
 
 	    olog = new NetLogic(scope, oname, 3, NetLogic::AND);
 
@@ -344,13 +487,105 @@ NetNet* NetEBLogic::synthesize(Design*des)
 	    connect(lsig->pin(0), olog->pin(1));
 
 	    assert(rsig->pin_count() == 1);
-	    connect(lsig->pin(0), olog->pin(2));
+	    connect(rsig->pin(0), olog->pin(2));
       }
 
 
       return osig;
 }
 
+NetNet* NetEBShift::synthesize(Design*des)
+{
+      if (! dynamic_cast<NetEConst*>(right_)) {
+	    NetExpr*tmp = right_->eval_tree();
+	    if (tmp) {
+		  delete right_;
+		  right_ = tmp;
+	    }
+      }
+
+      NetNet*lsig = left_->synthesize(des);
+      if (lsig == 0)
+	    return 0;
+
+      bool right_flag  =  op_ == 'r' || op_ == 'R';
+      bool signed_flag =  op_ == 'R';
+
+      NetScope*scope = lsig->scope();
+
+	/* Detect the special case where the shift amount is
+	   constant. Evaluate the shift amount, and simply reconnect
+	   the left operand to the output, but shifted. */
+      if (NetEConst*rcon = dynamic_cast<NetEConst*>(right_)) {
+	    verinum shift_v = rcon->value();
+	    long shift = shift_v.as_long();
+
+	    if (op() == 'r')
+		  shift = 0-shift;
+
+	    if (shift == 0)
+		  return lsig;
+
+	    NetNet*osig = new NetNet(scope, scope->local_symbol(),
+				     NetNet::IMPLICIT, expr_width());
+	    osig->local_flag(true);
+
+	    NetConst*zcon = new NetConst(scope, scope->local_symbol(),
+					 verinum::V0);
+	    des->add_node(zcon);
+	    NetNet*zsig = new NetNet(scope, scope->local_symbol(),
+				     NetNet::WIRE, 1);
+	    connect(zcon->pin(0), zsig->pin(0));
+
+	    if (shift > 0) {
+		  unsigned long ushift = shift;
+		  for (unsigned idx = 0; idx < osig->pin_count(); idx += 1)
+			if (idx < ushift) {
+			      connect(osig->pin(idx), zsig->pin(0));
+			} else {
+			      connect(osig->pin(idx), lsig->pin(idx-ushift));
+			}
+	    } else {
+		  unsigned long dshift = 0-shift;
+		  for (unsigned idx = 0; idx < osig->pin_count() ;  idx += 1)
+			if (idx+dshift < lsig->pin_count())
+			      connect(osig->pin(idx), lsig->pin(idx+dshift));
+			else
+			      connect(osig->pin(idx), zsig->pin(0));
+	    }
+
+
+	    return osig;
+      }
+
+      NetNet*rsig = right_->synthesize(des);
+      if (rsig == 0)
+	    return 0;
+
+      NetNet*osig = new NetNet(scope, scope->local_symbol(),
+			       NetNet::IMPLICIT, expr_width());
+      osig->local_flag(true);
+
+      assert(op() == 'l');
+      NetCLShift*dev = new NetCLShift(scope, scope->local_symbol(),
+				      osig->pin_count(),
+				      rsig->pin_count(),
+				      right_flag, signed_flag);
+      des->add_node(dev);
+
+      for (unsigned idx = 0 ; idx < dev->width() ;  idx += 1)
+	    connect(dev->pin_Result(idx), osig->pin(idx));
+
+      assert(lsig->pin_count() >= dev->width());
+      for (unsigned idx = 0 ;  idx < dev->width() ;  idx += 1)
+	    connect(dev->pin_Data(idx), lsig->pin(idx));
+
+      for (unsigned idx = 0 ;  idx < dev->width_dist() ;  idx += 1)
+	    connect(dev->pin_Distance(idx), rsig->pin(idx));
+
+      return osig;
+}
+
 NetNet* NetEConcat::synthesize(Design*des)
 {
 	/* First, synthesize the operands. */
@@ -363,7 +598,7 @@ NetNet* NetEConcat::synthesize(Design*des)
       assert(scope);
 
 	/* Make a NetNet object to carry the output vector. */
-      string path = scope->name() + "." + scope->local_symbol();
+      perm_string path = scope->local_symbol();
       NetNet*osig = new NetNet(scope, path, NetNet::IMPLICIT, expr_width());
       osig->local_flag(true);
 
@@ -378,7 +613,7 @@ NetNet* NetEConcat::synthesize(Design*des)
 		  obit += 1;
 	    }
 
-	    if (tmp[idx-1]->local_flag() && tmp[idx-1]->get_eref() == 0)
+	    if (tmp[idx-1]->local_flag() && tmp[idx-1]->get_refs() == 0)
 		  delete tmp[idx-1];
       }
 
@@ -391,12 +626,13 @@ NetNet* NetEConst::synthesize(Design*des)
       NetScope*scope = des->find_root_scope();
       assert(scope);
 
-      string path = scope->name() + "." + scope->local_symbol();
+      perm_string path = scope->local_symbol();
       unsigned width=expr_width();
 
       NetNet*osig = new NetNet(scope, path, NetNet::IMPLICIT, width);
       osig->local_flag(true);
-      NetConst*con = new NetConst(scope, des->local_symbol(path), value());
+      osig->set_signed(has_sign());
+      NetConst*con = new NetConst(scope, scope->local_symbol(), value());
       for (unsigned idx = 0 ;  idx < width;  idx += 1)
 	    connect(osig->pin(idx), con->pin(idx));
 
@@ -404,6 +640,14 @@ NetNet* NetEConst::synthesize(Design*des)
       return osig;
 }
 
+NetNet* NetECReal::synthesize(Design*des)
+{
+      cerr << get_line() << ": error: Real constants are "
+	   << "not synthesizable." << endl;
+      des->errors += 1;
+      return 0;
+}
+
 /*
  * The bitwise unary logic operator (there is only one) is turned
  * into discrete gates just as easily as the binary ones above.
@@ -414,14 +658,13 @@ NetNet* NetEUBits::synthesize(Design*des)
 
       NetScope*scope = isig->scope();
       assert(scope);
-      string path = des->local_symbol(scope->name());
 
-      NetNet*osig = new NetNet(scope, path, NetNet::IMPLICIT,
-			       isig->pin_count());
+      NetNet*osig = new NetNet(scope, scope->local_symbol(),
+			       NetNet::IMPLICIT, isig->pin_count());
       osig->local_flag(true);
 
       for (unsigned idx = 0 ;  idx < osig->pin_count() ;  idx += 1) {
-	    string oname = des->local_symbol(path);
+	    perm_string oname = scope->local_symbol();
 	    NetLogic*gate;
 
 	    switch (op()) {
@@ -447,12 +690,12 @@ NetNet* NetEUReduce::synthesize(Design*des)
 
       NetScope*scope = isig->scope();
       assert(scope);
-      string path = des->local_symbol(scope->name());
 
-      NetNet*osig = new NetNet(scope, path, NetNet::IMPLICIT, 1);
+      NetNet*osig = new NetNet(scope, scope->local_symbol(),
+			       NetNet::IMPLICIT, 1);
       osig->local_flag(true);
 
-      string oname = des->local_symbol(path);
+      perm_string oname = scope->local_symbol();
       NetLogic*gate;
 
       switch (op()) {
@@ -501,8 +744,46 @@ NetNet* NetEUReduce::synthesize(Design*des)
       return osig;
 }
 
+NetNet* NetESelect::synthesize(Design *des)
+{
+	// XXXX For now, only support pad form
+      assert(base_ == 0);
+
+      NetNet*sub = expr_->synthesize(des);
+      if (sub == 0)
+	    return 0;
+
+      NetScope*scope = sub->scope();
+      NetNet*net = new NetNet(scope, scope->local_symbol(),
+			      NetNet::IMPLICIT, expr_width());
+      if (has_sign()) {
+	    unsigned idx;
+
+	    for (idx = 0 ;  idx < sub->pin_count() ;  idx += 1)
+		  connect(sub->pin(idx), net->pin(idx));
+
+	    for ( ;  idx < net->pin_count(); idx += 1)
+		  connect(sub->pin(sub->pin_count()-1), net->pin(idx));
+
+      } else {
+	    unsigned idx;
+	    for (idx = 0 ;  idx < sub->pin_count() ;  idx += 1)
+		  connect(sub->pin(idx), net->pin(idx));
+
+	    NetConst*tmp = new NetConst(scope, scope->local_symbol(),
+					verinum::V0);
+
+	    for ( ;  idx < net->pin_count() ;  idx += 1)
+		  connect(net->pin(idx), tmp->pin(0));
+
+	    des->add_node(tmp);
+      }
+
+      return net;
+}
+
 /*
- * Synthesize a ?: operator an a NetMux device. Connect the condition
+ * Synthesize a ?: operator as a NetMux device. Connect the condition
  * expression to the select input, then connect the true and false
  * expressions to the B and A inputs. This way, when the select input
  * is one, the B input, which is the true expression, is selected.
@@ -513,15 +794,22 @@ NetNet* NetETernary::synthesize(Design *des)
       NetNet*tsig = true_val_->synthesize(des);
       NetNet*fsig = false_val_->synthesize(des);
 
-      string path = csig->scope()->name()+"."+csig->scope()->local_symbol();
+      perm_string path = csig->scope()->local_symbol();
 
       assert(csig->pin_count() == 1);
-      assert(tsig->pin_count() == fsig->pin_count());
-      unsigned width=tsig->pin_count();
+
+      unsigned width=expr_width();
       NetNet*osig = new NetNet(csig->scope(), path, NetNet::IMPLICIT, width);
       osig->local_flag(true);
 
-      string oname = des->local_symbol(path);
+	/* Make sure both value operands are the right width. */
+      tsig = pad_to_width(des, tsig, width);
+      fsig = pad_to_width(des, fsig, width);
+
+      assert(width <= tsig->pin_count());
+      assert(width <= fsig->pin_count());
+
+      perm_string oname = csig->scope()->local_symbol();
       NetMux *mux = new NetMux(csig->scope(), oname, width, 2, 1);
       for (unsigned idx = 0 ;  idx < width;  idx += 1) {
 	    connect(tsig->pin(idx), mux->pin_Data(idx, 1));
@@ -575,8 +863,8 @@ NetNet* NetESignal::synthesize(Design*des)
       NetScope*scope = net_->scope();
       assert(scope);
 
-      string name = scope->name() + "." + scope->local_symbol();
-      NetNet*tmp = new NetNet(scope, name, NetNet::NetNet::WIRE, wid);
+      perm_string name = scope->local_symbol();
+      NetNet*tmp = new NetNet(scope, name, NetNet::WIRE, wid);
       tmp->local_flag(true);
 
       for (unsigned idx = 0 ;  idx < wid ;  idx += 1)
@@ -587,122 +875,90 @@ NetNet* NetESignal::synthesize(Design*des)
 
 /*
  * $Log: expr_synth.cc,v $
- * Revision 1.31  2001/12/30 17:06:52  steve
- *  Synthesize reduction logic.
+ * Revision 1.59  2004/06/30 02:16:26  steve
+ *  Implement signed divide and signed right shift in nets.
  *
- * Revision 1.30  2001/12/18 05:34:02  steve
- *  Comments about MUX synthesis.
+ * Revision 1.58  2004/06/16 16:21:34  steve
+ *  Connect rsif of multiply to DataB.
  *
- * Revision 1.29  2001/11/29 01:58:18  steve
- *  Handle part selects in l-values of DFF devices.
+ * Revision 1.57  2004/06/12 15:00:02  steve
+ *  Support / and % in synthesized contexts.
  *
- * Revision 1.28  2001/10/28 01:14:53  steve
- *  NetObj constructor finally requires a scope.
+ * Revision 1.56  2004/06/01 01:04:57  steve
+ *  Fix synthesis method for logical and/or
  *
- * Revision 1.27  2001/10/20 05:21:51  steve
- *  Scope/module names are char* instead of string.
+ * Revision 1.55  2004/02/20 18:53:35  steve
+ *  Addtrbute keys are perm_strings.
  *
- * Revision 1.26  2001/08/31 22:59:48  steve
- *  synthesize the special case of compare with 0.
+ * Revision 1.54  2004/02/18 17:11:56  steve
+ *  Use perm_strings for named langiage items.
  *
- * Revision 1.25  2001/08/05 02:49:07  steve
- *  Properly synthesize part selects.
+ * Revision 1.53  2004/02/15 04:23:48  steve
+ *  Fix evaluation of compare to constant expression.
  *
- * Revision 1.24  2001/07/25 03:10:49  steve
- *  Create a config.h.in file to hold all the config
- *  junk, and support gcc 3.0. (Stephan Boettcher)
+ * Revision 1.52  2003/11/10 19:39:20  steve
+ *  Remove redundant scope tokens.
  *
- * Revision 1.23  2001/07/07 01:38:45  steve
- *  Put synthesized signals in proper scope.
+ * Revision 1.51  2003/10/27 06:04:21  steve
+ *  More flexible width handling for synthesized add.
  *
- * Revision 1.22  2001/06/15 04:14:18  steve
- *  Generate vvp code for GT and GE comparisons.
+ * Revision 1.50  2003/09/26 02:44:27  steve
+ *  Assure ternary arguments are wide enough.
  *
- * Revision 1.21  2001/06/07 02:12:43  steve
- *  Support structural addition.
+ * Revision 1.49  2003/09/03 23:31:36  steve
+ *  Support synthesis of constant downshifts.
  *
- * Revision 1.20  2001/02/15 06:59:36  steve
- *  FreeBSD port has a maintainer now.
+ * Revision 1.48  2003/08/28 04:11:18  steve
+ *  Spelling patch.
  *
- * Revision 1.19  2001/01/18 03:16:35  steve
- *  NetMux needs a scope. (PR#115)
+ * Revision 1.47  2003/08/09 03:23:40  steve
+ *  Add support for IVL_LPM_MULT device.
  *
- * Revision 1.18  2000/11/29 23:16:18  steve
- *  Do not delete synthesized signals used in expressions.
+ * Revision 1.46  2003/07/26 03:34:42  steve
+ *  Start handling pad of expressions in code generators.
  *
- * Revision 1.17  2000/11/29 05:24:00  steve
- *  synthesis for unary reduction ! and N operators.
+ * Revision 1.45  2003/06/24 01:38:02  steve
+ *  Various warnings fixed.
  *
- * Revision 1.16  2000/11/29 02:09:52  steve
- *  Add support for || synthesis (PR#53)
+ * Revision 1.44  2003/04/19 04:52:56  steve
+ *  Less picky about expression widths while synthesizing ternary.
  *
- * Revision 1.15  2000/10/07 19:45:43  steve
- *  Put logic devices into scopes.
+ * Revision 1.43  2003/04/08 05:07:15  steve
+ *  Detect constant shift distances in synthesis.
  *
- * Revision 1.14  2000/05/02 00:58:12  steve
- *  Move signal tables to the NetScope class.
+ * Revision 1.42  2003/04/08 04:33:55  steve
+ *  Synthesize shift expressions.
  *
- * Revision 1.13  2000/04/28 18:43:23  steve
- *  integer division in expressions properly get width.
+ * Revision 1.41  2003/03/06 00:28:41  steve
+ *  All NetObj objects have lex_string base names.
  *
- * Revision 1.12  2000/04/20 00:28:03  steve
- *  Catch some simple identity compareoptimizations.
+ * Revision 1.40  2003/02/26 01:29:24  steve
+ *  LPM objects store only their base names.
  *
- * Revision 1.11  2000/04/16 23:32:18  steve
- *  Synthesis of comparator in expressions.
+ * Revision 1.39  2003/01/30 16:23:07  steve
+ *  Spelling fixes.
  *
- *  Connect the NetEvent and related classes
- *  together better.
+ * Revision 1.38  2003/01/26 21:15:58  steve
+ *  Rework expression parsing and elaboration to
+ *  accommodate real/realtime values and expressions.
  *
- * Revision 1.10  2000/02/23 02:56:54  steve
- *  Macintosh compilers do not support ident.
+ * Revision 1.37  2002/11/17 23:37:55  steve
+ *  Magnitude compare to 0.
  *
- * Revision 1.9  2000/01/01 06:18:00  steve
- *  Handle synthesis of concatenation.
+ * Revision 1.36  2002/08/12 01:34:59  steve
+ *  conditional ident string using autoconfig.
  *
- * Revision 1.8  1999/12/17 06:18:15  steve
- *  Rewrite the cprop functor to use the functor_t interface.
+ * Revision 1.35  2002/07/07 22:31:39  steve
+ *  Smart synthesis of binary AND expressions.
  *
- * Revision 1.7  1999/12/17 03:38:46  steve
- *  NetConst can now hold wide constants.
+ * Revision 1.34  2002/07/05 21:26:17  steve
+ *  Avoid emitting to vvp local net symbols.
  *
- * Revision 1.6  1999/11/28 23:42:02  steve
- *  NetESignal object no longer need to be NetNode
- *  objects. Let them keep a pointer to NetNet objects.
- *
- * Revision 1.5  1999/11/27 19:07:57  steve
- *  Support the creation of scopes.
- *
- * Revision 1.4  1999/11/19 03:00:59  steve
- *  Whoops, created a signal with a duplicate name.
- *
- * Revision 1.3  1999/11/05 04:40:40  steve
- *  Patch to synthesize LPM_ADD_SUB from expressions,
- *  Thanks to Larry Doolittle. Also handle constants
- *  in expressions.
- *
- *  Synthesize adders in XNF, based on a patch from
- *  Larry. Accept synthesis of constants from Larry
- *  as is.
- *
- * Revision 1.2  1999/11/04 03:53:26  steve
- *  Patch to synthesize unary ~ and the ternary operator.
- *  Thanks to Larry Doolittle <LRDoolittle@lbl.gov>.
- *
- *  Add the LPM_MUX device, and integrate it with the
- *  ternary synthesis from Larry. Replace the lpm_mux
- *  generator in t-xnf.cc to use XNF EQU devices to
- *  put muxs into function units.
- *
- *  Rewrite elaborate_net for the PETernary class to
- *  also use the LPM_MUX device.
- *
- * Revision 1.1  1999/11/02 04:55:34  steve
- *  Add the synthesize method to NetExpr to handle
- *  synthesis of expressions, and use that method
- *  to improve r-value handling of LPM_FF synthesis.
- *
- *  Modify the XNF target to handle LPM_FF objects.
+ * Revision 1.33  2002/05/26 01:39:02  steve
+ *  Carry Verilog 2001 attributes with processes,
+ *  all the way through to the ivl_target API.
  *
+ *  Divide signal reference counts between rval
+ *  and lval references.
  */
 
