@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1998-2000 Stephen Williams (steve@icarus.com)
+ * Copyright (c) 1998-2004 Stephen Williams (steve@icarus.com)
  *
  *    This source code is free software; you can redistribute it
  *    and/or modify it in source code form under the terms of the GNU
@@ -16,8 +16,8 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT) && !defined(macintosh)
-#ident "$Id: pform_dump.cc,v 1.69 2002/01/26 05:28:28 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: pform_dump.cc,v 1.88 2004/10/04 01:10:55 steve Exp $"
 #endif
 
 # include "config.h"
@@ -29,6 +29,7 @@
  * module in question.
  */
 # include  "pform.h"
+# include  "PData.h"
 # include  "PEvent.h"
 # include  <iostream>
 # include  <iomanip>
@@ -251,11 +252,13 @@ void PWire::dump(ostream&out) const
       }
 
       out << ";" << endl;
-      for (map<string,string>::const_iterator idx = attributes.begin()
+      for (map<perm_string,PExpr*>::const_iterator idx = attributes.begin()
 		 ; idx != attributes.end()
 		 ; idx ++) {
-	    out << "        " << (*idx).first << " = \"" <<
-		  (*idx).second << "\"" << endl;
+	    out << "        " << (*idx).first;
+	    if ((*idx).second)
+		  out << " = " << *(*idx).second;
+	    out << endl;
       }
 }
 
@@ -263,7 +266,7 @@ void PGate::dump_pins(ostream&out) const
 {
       if (pin_count()) {
 	    if (pin(0)) out << *pin(0);
-			      
+
 	    for (unsigned idx = 1 ;  idx < pin_count() ;  idx += 1) {
 		  out << ", ";
 		  if (pin(idx)) out << *pin(idx);
@@ -434,7 +437,7 @@ void PAssignNB::dump(ostream&out, unsigned ind) const
 void PBlock::dump(ostream&out, unsigned ind) const
 {
       out << setw(ind) << "" << "begin";
-      if (name_.length())
+      if (name_ != 0)
 	    out << " : " << name_;
       out << endl;
 
@@ -492,7 +495,7 @@ void PCase::dump(ostream&out, unsigned ind) const
 
 	    } else {
 		  out << setw(ind+2) << "" << *cur->expr[0];
-			
+
 		  for(unsigned e = 1 ; e < cur->expr.count() ; e += 1)
 			out << ", " << *cur->expr[e];
 
@@ -555,13 +558,17 @@ void PDisable::dump(ostream&out, unsigned ind) const
 
 void PEventStatement::dump(ostream&out, unsigned ind) const
 {
-      out << setw(ind) << "" << "@(" << *(expr_[0]);
-      if (expr_.count() > 1)
-	    for (unsigned idx = 1 ;  idx < expr_.count() ;  idx += 1)
-		  out << " or " << *(expr_[idx]);
+      if (expr_.count() == 0) {
+	    out << setw(ind) << "" << "@* ";
 
-      out << ")";
+      } else {
+	    out << setw(ind) << "" << "@(" << *(expr_[0]);
+	    if (expr_.count() > 1)
+		  for (unsigned idx = 1 ;  idx < expr_.count() ;  idx += 1)
+			out << " or " << *(expr_[idx]);
 
+	    out << ")";
+      }
 
       if (statement_) {
 	    out << endl;
@@ -593,7 +600,35 @@ void PForStatement::dump(ostream&out, unsigned ind) const
 
 void PFunction::dump(ostream&out, unsigned ind) const
 {
-      out << setw(ind) << "" << "output " << out_->path() << ";" << endl;
+      out << setw(ind) << "" << "function ";
+      switch (return_type_.type) {
+	  case PTF_NONE:
+	    out << "?none? ";
+	    break;
+	  case PTF_REG:
+	    out << "reg ";
+	    break;
+	  case PTF_INTEGER:
+	    out << "integer ";
+	    break;
+	  case PTF_REAL:
+	    out << "real ";
+	    break;
+	  case PTF_REALTIME:
+	    out << "realtime ";
+	    break;
+	  case PTF_TIME:
+	    out << "time ";
+	    break;
+      }
+
+      if (return_type_.range) {
+	    out << "[";
+	    out << "] ";
+      }
+
+      out << name_ << ";" << endl;
+
       if (ports_)
 	    for (unsigned idx = 0 ;  idx < ports_->count() ;  idx += 1) {
 		  out << setw(ind) << "";
@@ -634,6 +669,9 @@ void PTask::dump(ostream&out, unsigned ind) const
 		      case NetNet::PINOUT:
 			out << "inout ";
 			break;
+		      default:
+			assert(0);
+			break;
 		  }
 		  out << (*ports_)[idx]->path() << ";" << endl;
 	    }
@@ -668,15 +706,40 @@ void PProcess::dump(ostream&out, unsigned ind) const
 
       out << " /* " << get_line() << " */" << endl;
 
+      for (map<perm_string,PExpr*>::const_iterator idx = attributes.begin()
+		 ; idx != attributes.end() ; idx++ ) {
+
+	    out << setw(ind+2) << "" << "(* " << (*idx).first;
+	    if ((*idx).second) {
+		  out << " = " << *(*idx).second;
+	    }
+	    out << " *)" << endl;
+      }
+
       statement_->dump(out, ind+2);
 }
 
 void Module::dump(ostream&out) const
 {
+      if (attributes.begin() != attributes.end()) {
+	    out << "(* ";
+	    for (map<perm_string,PExpr*>::const_iterator idx = attributes.begin()
+		 ; idx != attributes.end() ; idx++ ) {
+		    if (idx != attributes.begin()) {
+			out << " , ";
+		    }
+		    out << (*idx).first;
+		    if ((*idx).second) {
+			out << " = " << *(*idx).second;
+		    }
+	    }
+	    out << " *)  ";
+      }
+
       out << "module " << name_ << ";" << endl;
 
-      for (unsigned idx = 0 ;  idx < ports_.count() ;  idx += 1) {
-	    port_t*cur = ports_[idx];
+      for (unsigned idx = 0 ;  idx < ports.count() ;  idx += 1) {
+	    port_t*cur = ports[idx];
 
 	    if (cur == 0) {
 		  out << "    unconnected" << endl;
@@ -691,26 +754,43 @@ void Module::dump(ostream&out) const
 	    out << ")" << endl;
       }
 
-      typedef map<string,PExpr*>::const_iterator parm_iter_t;
+      typedef map<perm_string,param_expr_t>::const_iterator parm_iter_t;
       typedef map<hname_t,PExpr*>::const_iterator parm_hiter_t;
       for (parm_iter_t cur = parameters.begin()
 		 ; cur != parameters.end() ; cur ++) {
-	    out << "    parameter " << (*cur).first << " = ";
-	    if ((*cur).second)
-		  out << *(*cur).second << ";" << endl;
+	    out << "    parameter ";
+	    if ((*cur).second.signed_flag)
+		  out << "signed ";
+	    if ((*cur).second.msb)
+		  out << "[" << *(*cur).second.msb << ":"
+		      << *(*cur).second.lsb << "] ";
+	    out << (*cur).first << " = ";
+	    if ((*cur).second.expr)
+		  out << *(*cur).second.expr << ";" << endl;
 	    else
 		  out << "/* ERROR */;" << endl;
       }
 
       for (parm_iter_t cur = localparams.begin()
 		 ; cur != localparams.end() ; cur ++) {
-	    out << "    localparam " << (*cur).first << " = ";
-	    if ((*cur).second)
-		  out << *(*cur).second << ";" << endl;
+	    out << "    localparam ";
+	    if ((*cur).second.msb)
+		  out << "[" << *(*cur).second.msb << ":"
+		      << *(*cur).second.lsb << "] ";
+	    out << (*cur).first << " = ";
+	    if ((*cur).second.expr)
+		  out << *(*cur).second.expr << ";" << endl;
 	    else
 		  out << "/* ERROR */;" << endl;
       }
 
+      typedef map<perm_string,PExpr*>::const_iterator specparm_iter_t;
+      for (specparm_iter_t cur = specparams.begin()
+		 ; cur != specparams.end() ; cur ++) {
+	    out << "    specparam " << (*cur).first << " = "
+		<< *(*cur).second << ";" << endl;
+      }
+
       for (parm_hiter_t cur = defparms.begin()
 		 ; cur != defparms.end() ;  cur ++) {
 	    out << "    defparam " << (*cur).first << " = ";
@@ -720,13 +800,20 @@ void Module::dump(ostream&out) const
 		  out << "/* ERROR */;" << endl;
       }
 
-      for (map<string,PEvent*>::const_iterator cur = events.begin()
+      for (map<perm_string,PEvent*>::const_iterator cur = events.begin()
 		 ; cur != events.end() ;  cur ++ ) {
 	    PEvent*ev = (*cur).second;
 	    out << "    event " << ev->name() << "; // "
 		<< ev->get_line() << endl;
       }
 
+      for (map<hname_t,PData*>::const_iterator cur = datum.begin()
+		 ; cur != datum.end() ;  cur ++ ) {
+	    PData*tmp = (*cur).second;
+	    out << "    real " << tmp->name() << "; // "
+		<< tmp->get_line() << endl;
+      }
+
 	// Iterate through and display all the wires.
       for (map<hname_t,PWire*>::const_iterator wire = wires_.begin()
 		 ; wire != wires_.end()
@@ -736,7 +823,7 @@ void Module::dump(ostream&out) const
       }
 
 	// Dump the task definitions.
-      typedef map<string,PTask*>::const_iterator task_iter_t;
+      typedef map<perm_string,PTask*>::const_iterator task_iter_t;
       for (task_iter_t cur = tasks_.begin()
 		 ; cur != tasks_.end() ; cur ++) {
 	    out << "    task " << (*cur).first << ";" << endl;
@@ -745,7 +832,7 @@ void Module::dump(ostream&out) const
       }
 
 	// Dump the function definitions.
-      typedef map<string,PFunction*>::const_iterator func_iter_t;
+      typedef map<perm_string,PFunction*>::const_iterator func_iter_t;
       for (func_iter_t cur = funcs_.begin()
 		 ; cur != funcs_.end() ; cur ++) {
 	    out << "    function " << (*cur).first << ";" << endl;
@@ -805,256 +892,101 @@ void PUdp::dump(ostream&out) const
 	    out << "    initial " << ports[0] << " = 1'b" << initial
 		<< ";" << endl;
 
-      out << "endprimitive" << endl;
-
 	// Dump the attributes for the primitive as attribute
 	// statements.
-      for (map<string,string>::const_iterator idx = attributes.begin()
+      for (map<string,PExpr*>::const_iterator idx = attributes.begin()
 		 ; idx != attributes.end()
 		 ; idx ++) {
-	    out << "$attribute(" << name_ << ", \"" << (*idx).first <<
-		  "\", \"" << (*idx).second << "\")" << endl;
+	    out << "    attribute " << (*idx).first;
+	    if ((*idx).second)
+		  out << " = " << *(*idx).second;
+	    out << endl;
       }
+
+      out << "endprimitive" << endl;
 }
 
 
 /*
  * $Log: pform_dump.cc,v $
+ * Revision 1.88  2004/10/04 01:10:55  steve
+ *  Clean up spurious trailing white space.
+ *
+ * Revision 1.87  2004/05/31 23:34:39  steve
+ *  Rewire/generalize parsing an elaboration of
+ *  function return values to allow for better
+ *  speed and more type support.
+ *
+ * Revision 1.86  2004/05/25 19:21:07  steve
+ *  More identifier lists use perm_strings.
+ *
+ * Revision 1.85  2004/02/20 18:53:35  steve
+ *  Addtrbute keys are perm_strings.
+ *
+ * Revision 1.84  2004/02/20 06:22:58  steve
+ *  parameter keys are per_strings.
+ *
+ * Revision 1.83  2004/02/18 17:11:57  steve
+ *  Use perm_strings for named langiage items.
+ *
+ * Revision 1.82  2003/07/05 20:42:08  steve
+ *  Fix some enumeration warnings.
+ *
+ * Revision 1.81  2003/06/20 00:53:19  steve
+ *  Module attributes from the parser
+ *  through to elaborated form.
+ *
+ * Revision 1.80  2003/06/13 19:10:46  steve
+ *  Properly manage real variables in subscopes.
+ *
+ * Revision 1.79  2003/02/27 06:45:11  steve
+ *  specparams as far as pform.
+ *
+ * Revision 1.78  2003/01/26 21:15:59  steve
+ *  Rework expression parsing and elaboration to
+ *  accommodate real/realtime values and expressions.
+ *
+ * Revision 1.77  2002/10/19 22:59:49  steve
+ *  Redo the parameter vector support to allow
+ *  parameter names in range expressions.
+ *
+ * Revision 1.76  2002/08/19 02:39:17  steve
+ *  Support parameters with defined ranges.
+ *
+ * Revision 1.75  2002/08/12 01:35:00  steve
+ *  conditional ident string using autoconfig.
+ *
+ * Revision 1.74  2002/05/26 01:39:02  steve
+ *  Carry Verilog 2001 attributes with processes,
+ *  all the way through to the ivl_target API.
+ *
+ *  Divide signal reference counts between rval
+ *  and lval references.
+ *
+ * Revision 1.73  2002/05/24 04:36:23  steve
+ *  Verilog 2001 attriubtes on nets/wires.
+ *
+ * Revision 1.72  2002/05/23 03:08:51  steve
+ *  Add language support for Verilog-2001 attribute
+ *  syntax. Hook this support into existing $attribute
+ *  handling, and add number and void value types.
+ *
+ *  Add to the ivl_target API new functions for access
+ *  of complex attributes attached to gates.
+ *
+ * Revision 1.71  2002/05/19 23:37:28  steve
+ *  Parse port_declaration_lists from the 2001 Standard.
+ *
+ * Revision 1.70  2002/04/21 04:59:08  steve
+ *  Add support for conbinational events by finding
+ *  the inputs to expressions and some statements.
+ *  Get case and assignment statements working.
+ *
  * Revision 1.69  2002/01/26 05:28:28  steve
  *  Detect scalar/vector declarion mismatch.
  *
  * Revision 1.68  2001/12/03 04:47:15  steve
  *  Parser and pform use hierarchical names as hname_t
  *  objects instead of encoded strings.
- *
- * Revision 1.67  2001/07/25 03:10:49  steve
- *  Create a config.h.in file to hold all the config
- *  junk, and support gcc 3.0. (Stephan Boettcher)
- *
- * Revision 1.66  2001/01/13 22:20:08  steve
- *  Parse parameters within nested scopes.
- *
- * Revision 1.65  2000/12/11 00:31:43  steve
- *  Add support for signed reg variables,
- *  simulate in t-vvm signed comparisons.
- *
- * Revision 1.64  2000/12/10 22:01:36  steve
- *  Support decimal constants in behavioral delays.
- *
- * Revision 1.63  2000/11/11 01:52:09  steve
- *  change set for support of nmos, pmos, rnmos, rpmos, notif0, and notif1
- *  change set to correct behavior of bufif0 and bufif1
- *  (Tim Leight)
- *
- *  Also includes fix for PR#27
- *
- * Revision 1.62  2000/10/14 02:23:02  steve
- *  Check for missing concat subexpressions (PR#11)
- *
- * Revision 1.61  2000/07/26 05:08:07  steve
- *  Parse disable statements to pform.
- *
- * Revision 1.60  2000/05/23 16:03:13  steve
- *  Better parsing of expressions lists will empty expressoins.
- *
- * Revision 1.59  2000/05/16 04:05:16  steve
- *  Module ports are really special PEIdent
- *  expressions, because a name can be used
- *  many places in the port list.
- *
- * Revision 1.58  2000/05/11 23:37:27  steve
- *  Add support for procedural continuous assignment.
- *
- * Revision 1.57  2000/05/06 15:41:57  steve
- *  Carry assignment strength to pform.
- *
- * Revision 1.56  2000/05/04 03:37:59  steve
- *  Add infrastructure for system functions, move
- *  $time to that structure and add $random.
- *
- * Revision 1.55  2000/04/22 04:20:19  steve
- *  Add support for force assignment.
- *
- * Revision 1.54  2000/04/12 20:02:53  steve
- *  Finally remove the NetNEvent and NetPEvent classes,
- *  Get synthesis working with the NetEvWait class,
- *  and get started supporting multiple events in a
- *  wait in vvm.
- *
- * Revision 1.53  2000/04/12 04:23:58  steve
- *  Named events really should be expressed with PEIdent
- *  objects in the pform,
- *
- *  Handle named events within the mix of net events
- *  and edges. As a unified lot they get caught together.
- *  wait statements are broken into more complex statements
- *  that include a conditional.
- *
- *  Do not generate NetPEvent or NetNEvent objects in
- *  elaboration. NetEvent, NetEvWait and NetEvProbe
- *  take over those functions in the netlist.
- *
- * Revision 1.52  2000/04/01 19:31:57  steve
- *  Named events as far as the pform.
- *
- * Revision 1.51  2000/03/12 17:09:41  steve
- *  Support localparam.
- *
- * Revision 1.50  2000/03/08 04:36:54  steve
- *  Redesign the implementation of scopes and parameters.
- *  I now generate the scopes and notice the parameters
- *  in a separate pass over the pform. Once the scopes
- *  are generated, I can process overrides and evalutate
- *  paremeters before elaboration begins.
- *
- * Revision 1.49  2000/02/23 02:56:55  steve
- *  Macintosh compilers do not support ident.
- *
- * Revision 1.48  2000/02/18 05:15:03  steve
- *  Catch module instantiation arrays.
- *
- * Revision 1.47  2000/01/09 20:37:57  steve
- *  Careful with wires connected to multiple ports.
- *
- * Revision 1.46  2000/01/09 05:50:49  steve
- *  Support named parameter override lists.
- *
- * Revision 1.45  1999/10/10 01:59:55  steve
- *  Structural case equals device.
- *
- * Revision 1.44  1999/09/30 02:43:02  steve
- *  Elaborate ~^ and ~| operators.
- *
- * Revision 1.43  1999/09/30 00:48:50  steve
- *  Cope with errors during ternary operator elaboration.
- *
- * Revision 1.42  1999/09/29 21:15:58  steve
- *  Handle some mor missing names.
- *
- * Revision 1.41  1999/09/29 20:23:53  steve
- *  Handle empty named ports in the dump.
- *
- * Revision 1.40  1999/09/29 18:36:04  steve
- *  Full case support
- *
- * Revision 1.39  1999/09/17 02:06:26  steve
- *  Handle unconnected module ports.
- *
- * Revision 1.38  1999/09/08 02:24:39  steve
- *  Empty conditionals (pmonta@imedia.com)
- *
- * Revision 1.37  1999/09/04 19:11:46  steve
- *  Add support for delayed non-blocking assignments.
- *
- * Revision 1.36  1999/08/25 22:22:41  steve
- *  elaborate some aspects of functions.
- *
- * Revision 1.35  1999/08/23 16:48:39  steve
- *  Parameter overrides support from Peter Monta
- *  AND and XOR support wide expressions.
- *
- * Revision 1.34  1999/08/03 04:49:13  steve
- *  Proper port type names.
- *
- * Revision 1.33  1999/08/03 04:14:49  steve
- *  Parse into pform arbitrarily complex module
- *  port declarations.
- *
- * Revision 1.32  1999/08/01 16:34:50  steve
- *  Parse and elaborate rise/fall/decay times
- *  for gates, and handle the rules for partial
- *  lists of times.
- *
- * Revision 1.31  1999/07/31 19:14:47  steve
- *  Add functions up to elaboration (Ed Carter)
- *
- * Revision 1.30  1999/07/30 00:43:17  steve
- *  Handle dumping tasks with no ports.
- *
- * Revision 1.29  1999/07/24 02:11:20  steve
- *  Elaborate task input ports.
- *
- * Revision 1.28  1999/07/17 19:51:00  steve
- *  netlist support for ternary operator.
- *
- * Revision 1.27  1999/07/12 00:59:36  steve
- *  procedural blocking assignment delays.
- *
- * Revision 1.26  1999/07/03 02:12:52  steve
- *  Elaborate user defined tasks.
- *
- * Revision 1.25  1999/06/24 04:24:18  steve
- *  Handle expression widths for EEE and NEE operators,
- *  add named blocks and scope handling,
- *  add registers declared in named blocks.
- *
- * Revision 1.24  1999/06/19 21:06:16  steve
- *  Elaborate and supprort to vvm the forever
- *  and repeat statements.
- *
- * Revision 1.23  1999/06/17 05:34:42  steve
- *  Clean up interface of the PWire class,
- *  Properly match wire ranges.
- *
- * Revision 1.22  1999/06/15 05:38:39  steve
- *  Support case expression lists.
- *
- * Revision 1.21  1999/06/15 03:44:53  steve
- *  Get rid of the STL vector template.
- *
- * Revision 1.20  1999/06/13 23:51:16  steve
- *  l-value part select for procedural assignments.
- *
- * Revision 1.19  1999/06/10 04:03:53  steve
- *  Add support for the Ternary operator,
- *  Add support for repeat concatenation,
- *  Correct some seg faults cause by elaboration
- *  errors,
- *  Parse the casex anc casez statements.
- *
- * Revision 1.18  1999/06/06 20:45:39  steve
- *  Add parse and elaboration of non-blocking assignments,
- *  Replace list<PCase::Item*> with an svector version,
- *  Add integer support.
- *
- * Revision 1.17  1999/05/29 02:36:17  steve
- *  module parameter bind by name.
- *
- * Revision 1.16  1999/05/10 00:16:58  steve
- *  Parse and elaborate the concatenate operator
- *  in structural contexts, Replace vector<PExpr*>
- *  and list<PExpr*> with svector<PExpr*>, evaluate
- *  constant expressions with parameters, handle
- *  memories as lvalues.
- *
- *  Parse task declarations, integer types.
- *
- * Revision 1.15  1999/05/05 03:04:46  steve
- *  Fix handling of null delay statements.
- *
- * Revision 1.14  1999/05/01 02:57:53  steve
- *  Handle much more complex event expressions.
- *
- * Revision 1.13  1999/04/29 02:16:26  steve
- *  Parse OR of event expressions.
- *
- * Revision 1.12  1999/04/19 01:59:37  steve
- *  Add memories to the parse and elaboration phases.
- *
- * Revision 1.11  1999/02/21 17:01:57  steve
- *  Add support for module parameters.
- *
- * Revision 1.10  1999/02/15 02:06:15  steve
- *  Elaborate gate ranges.
- *
- * Revision 1.9  1999/02/03 04:20:11  steve
- *  Parse and elaborate the Verilog CASE statement.
- *
- * Revision 1.8  1999/02/01 00:26:49  steve
- *  Carry some line info to the netlist,
- *  Dump line numbers for processes.
- *  Elaborate prints errors about port vector
- *  width mismatch
- *  Emit better handles null statements.
  */
 
