@@ -16,19 +16,23 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT) && !defined(macintosh)
-#ident "$Id: t-dll-api.cc,v 1.75 2002/01/28 00:52:41 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: t-dll-api.cc,v 1.108 2004/10/04 01:10:55 steve Exp $"
 #endif
 
 # include "config.h"
-
 # include  "t-dll.h"
+# include  <stdlib.h>
+# include  <string.h>
+#ifdef HAVE_MALLOC_H
+# include  <malloc.h>
+#endif
 
 /* THE FOLLOWING ARE FUNCTIONS THAT ARE CALLED FROM THE TARGET. */
 
 extern "C" const char*ivl_design_flag(ivl_design_t des, const char*key)
 {
-      return des->self->get_flag(key).c_str();
+      return des->self->get_flag(key);
 }
 
 extern "C" int ivl_design_process(ivl_design_t des,
@@ -89,14 +93,15 @@ inline static const char *basename(ivl_scope_t scope, const char *inst)
       return inst+1;
 }
 
-extern "C" const char*ivl_memory_name(ivl_memory_t net)
-{
-      return net->name_;
-}
-
 extern "C" const char* ivl_memory_basename(ivl_memory_t net)
 {
-      return basename(net->scope_, net->name_);
+      return net->basename_;
+}
+
+extern "C" ivl_scope_t ivl_memory_scope(ivl_memory_t net)
+{
+      assert(net);
+      return net->scope_;
 }
 
 extern "C" int ivl_memory_root(ivl_memory_t net)
@@ -143,17 +148,36 @@ extern "C" unsigned ivl_const_pins(ivl_net_const_t net)
 extern "C" int ivl_const_signed(ivl_net_const_t net)
 {
       assert(net);
-      return net->signed_ == 1;
+      return net->signed_;
 }
 
 extern "C" const char* ivl_event_name(ivl_event_t net)
 {
-      return net->name;
+      static char*name_buffer = 0;
+      static unsigned name_size = 0;
+
+      ivl_scope_t scope = net->scope;
+      const char*sn = ivl_scope_name(scope);
+
+      unsigned need = strlen(sn) + 1 + strlen(net->name) + 1;
+      if (need > name_size) {
+	    name_buffer = (char*)realloc(name_buffer, need);
+	    name_size = need;
+      }
+
+      strcpy(name_buffer, sn);
+      char*tmp = name_buffer + strlen(sn);
+      *tmp++ = '.';
+      strcpy(tmp, net->name);
+
+      cerr << "ANACHRONISM: Call to anachronistic ivl_event_name." << endl;
+
+      return name_buffer;
 }
 
 extern "C" const char* ivl_event_basename(ivl_event_t net)
 {
-      return basename(net->scope, net->name);
+      return net->name;
 }
 
 
@@ -218,6 +242,12 @@ extern "C" ivl_scope_t ivl_expr_def(ivl_expr_t net)
       return 0;
 }
 
+extern "C" double ivl_expr_dvalue(ivl_expr_t net)
+{
+      assert(net->type_ == IVL_EX_REALNUM);
+      return net->u_.real_.value;
+}
+
 extern "C" unsigned ivl_expr_lsi(ivl_expr_t net)
 {
       switch (net->type_) {
@@ -242,8 +272,8 @@ extern "C" const char* ivl_expr_name(ivl_expr_t net)
 	  case IVL_EX_SIGNAL:
 	    return net->u_.signal_.sig->name_;
 
-	  case IVL_EX_MEMORY:
-	    return net->u_.memory_.mem_->name_;
+	  case IVL_EX_VARIABLE:
+	    return net->u_.variable_.var->name;
 
 	  default:
 	    assert(0);
@@ -326,6 +356,20 @@ extern "C" ivl_expr_t ivl_expr_oper3(ivl_expr_t net)
       return 0;
 }
 
+extern "C" ivl_parameter_t ivl_expr_parameter(ivl_expr_t net)
+{
+      switch (net->type_) {
+	  case IVL_EX_NUMBER:
+	    return net->u_.number_.parameter;
+	  case IVL_EX_STRING:
+	    return net->u_.string_.parameter;
+	  case IVL_EX_REALNUM:
+	    return net->u_.real_.parameter;
+	  default:
+	    return 0;
+      }
+}
+
 extern "C" ivl_expr_t ivl_expr_parm(ivl_expr_t net, unsigned idx)
 {
       assert(net);
@@ -376,6 +420,13 @@ extern "C" unsigned ivl_expr_repeat(ivl_expr_t net)
       return net->u_.concat_.rept;
 }
 
+extern "C" ivl_event_t ivl_expr_event(ivl_expr_t net)
+{
+      assert(net);
+      assert(net->type_ == IVL_EX_EVENT);
+      return net->u_.event_.event;
+}
+
 extern "C" ivl_scope_t ivl_expr_scope(ivl_expr_t net)
 {
       assert(net);
@@ -413,8 +464,31 @@ extern "C" const char* ivl_expr_string(ivl_expr_t net)
 
 extern "C" unsigned long ivl_expr_uvalue(ivl_expr_t net)
 {
-      assert(net->type_ == IVL_EX_ULONG);
-      return net->u_.ulong_.value;
+      switch (net->type_) {
+
+	  case IVL_EX_ULONG:
+	    return net->u_.ulong_.value;
+
+	  case IVL_EX_NUMBER: {
+		unsigned long val = 0;
+		for (unsigned long idx = 0 ;  idx < net->width_ ;  idx += 1) {
+		      if (net->u_.number_.bits_[idx] == '1')
+			    val |= 1UL << idx;
+		}
+
+		return val;
+	  }
+
+	  default:
+	    assert(0);
+      }
+
+}
+
+extern "C" ivl_variable_type_t ivl_expr_value(ivl_expr_t net)
+{
+      assert(net);
+      return net->value_;
 }
 
 extern "C" unsigned ivl_expr_width(ivl_expr_t net)
@@ -429,29 +503,51 @@ extern "C" ivl_memory_t ivl_expr_memory(ivl_expr_t net)
       return net->u_.memory_.mem_;
 }
 
+extern "C" ivl_variable_t ivl_expr_variable(ivl_expr_t net)
+{
+      assert(net->type_ == IVL_EX_VARIABLE);
+      return net->u_.variable_.var;
+}
+
 extern "C" const char* ivl_logic_attr(ivl_net_logic_t net, const char*key)
 {
       assert(net);
       unsigned idx;
 
-      for (idx = 0 ;  idx < net->nattr_ ;  idx += 1) {
+      for (idx = 0 ;  idx < net->nattr ;  idx += 1) {
 
-	    if (strcmp(net->akey_[idx], key) == 0)
-		  return net->aval_[idx];
+	    if (strcmp(net->attr[idx].key, key) == 0)
+		  return net->attr[idx].type == IVL_ATT_STR
+			? net->attr[idx].val.str
+			: 0;
       }
 
       return 0;
 }
 
+extern "C" unsigned ivl_logic_attr_cnt(ivl_net_logic_t net)
+{
+      return net->nattr;
+}
+
+extern "C" ivl_attribute_t ivl_logic_attr_val(ivl_net_logic_t net,
+					      unsigned idx)
+{
+      assert(idx < net->nattr);
+      return net->attr + idx;
+}
+
 extern "C" const char* ivl_logic_name(ivl_net_logic_t net)
 {
       assert(net);
+      cerr << "ANACHRONISM: Call to anachronistic ivl_logic_name." << endl;
       return net->name_;
 }
 
 extern "C" const char* ivl_logic_basename(ivl_net_logic_t net)
 {
-      return basename(net->scope_, net->name_);
+      assert(net);
+      return net->name_;
 }
 
 extern "C" ivl_scope_t ivl_logic_scope(ivl_net_logic_t net)
@@ -526,9 +622,56 @@ extern "C" const char* ivl_udp_name(ivl_udp_t net)
 
 extern "C" const char* ivl_lpm_basename(ivl_lpm_t net)
 {
-      return basename(net->scope, net->name);
+      return net->name;
 }
 
+extern "C" ivl_nexus_t ivl_lpm_async_clr(ivl_lpm_t net)
+{
+      assert(net);
+      switch(net->type) {
+	  case IVL_LPM_FF:
+	    return net->u_.ff.aclr;
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
+
+extern "C" ivl_nexus_t ivl_lpm_sync_clr(ivl_lpm_t net)
+{
+      assert(net);
+      switch(net->type) {
+	  case IVL_LPM_FF:
+	    return net->u_.ff.sclr;
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
+
+extern "C" ivl_nexus_t ivl_lpm_async_set(ivl_lpm_t net)
+{
+      assert(net);
+      switch(net->type) {
+	  case IVL_LPM_FF:
+	    return net->u_.ff.aset;
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
+
+extern "C" ivl_nexus_t ivl_lpm_sync_set(ivl_lpm_t net)
+{
+      assert(net);
+      switch(net->type) {
+	  case IVL_LPM_FF:
+	    return net->u_.ff.sset;
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
 
 extern "C" ivl_nexus_t ivl_lpm_clk(ivl_lpm_t net)
 {
@@ -543,6 +686,43 @@ extern "C" ivl_nexus_t ivl_lpm_clk(ivl_lpm_t net)
       }
 }
 
+extern "C" ivl_expr_t ivl_lpm_aset_value(ivl_lpm_t net)
+{
+      assert(net);
+      switch (net->type) {
+	  case IVL_LPM_FF:
+	  case IVL_LPM_RAM:
+	    return net->u_.ff.aset_value;
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
+extern "C" ivl_expr_t ivl_lpm_sset_value(ivl_lpm_t net)
+{
+      assert(net);
+      switch (net->type) {
+	  case IVL_LPM_FF:
+	  case IVL_LPM_RAM:
+	    return net->u_.ff.sset_value;
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
+
+extern "C" ivl_scope_t ivl_lpm_define(ivl_lpm_t net)
+{
+      assert(net);
+      switch (net->type) {
+	  case IVL_LPM_UFUNC:
+	    return net->u_.ufunc.def;
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
+
 extern "C" ivl_nexus_t ivl_lpm_enable(ivl_lpm_t net)
 {
       assert(net);
@@ -623,17 +803,66 @@ extern "C" ivl_nexus_t ivl_lpm_data2(ivl_lpm_t net, unsigned sdx, unsigned idx)
 	    assert(idx < net->u_.mux.width);
 	    return net->u_.mux.d[sdx*net->u_.mux.width + idx];
 
+	  case IVL_LPM_UFUNC: {
+		sdx += 1; /* skip the output port. */
+		assert(sdx < net->u_.ufunc.ports);
+		assert(idx < net->u_.ufunc.port_wid[sdx]);
+		unsigned base = 0;
+		for (unsigned i = 0 ;  i < sdx ;  i += 1)
+		      base += net->u_.ufunc.port_wid[i];
+		return net->u_.ufunc.pins[base+idx];
+	  }
+
 	  default:
 	    assert(0);
 	    return 0;
       }
 }
 
+extern "C" unsigned ivl_lpm_data2_width(ivl_lpm_t net, unsigned sdx)
+{
+      assert(net);
+      switch (net->type) {
+	  case IVL_LPM_UFUNC:
+	    sdx += 1; /* skip the output port. */
+	    assert(sdx < net->u_.ufunc.ports);
+	    return net->u_.ufunc.port_wid[sdx];
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
+
+/*
+ * This function returns the hierarchical name for the LPM device. The
+ * name needs to be built up from the scope name and the lpm base
+ * name.
+ *
+ * Anachronism: This function is provided for
+ * compatibility. Eventually, it will be removed.
+ */
 extern "C" const char* ivl_lpm_name(ivl_lpm_t net)
 {
-      return net->name;
+      static char*name_buffer = 0;
+      static unsigned name_size = 0;
+
+      ivl_scope_t scope = ivl_lpm_scope(net);
+      const char*sn = ivl_scope_name(scope);
+
+      unsigned need = strlen(sn) + 1 + strlen(net->name) + 1;
+      if (need > name_size) {
+	    name_buffer = (char*)realloc(name_buffer, need);
+	    name_size = need;
+      }
+
+      strcpy(name_buffer, sn);
+      char*tmp = name_buffer + strlen(sn);
+      *tmp++ = '.';
+      strcpy(tmp, net->name);
+      return name_buffer;
 }
 
+
 extern "C" ivl_nexus_t ivl_lpm_q(ivl_lpm_t net, unsigned idx)
 {
       assert(net);
@@ -674,6 +903,10 @@ extern "C" ivl_nexus_t ivl_lpm_q(ivl_lpm_t net, unsigned idx)
 	    assert(idx < net->u_.shift.width);
 	    return net->u_.shift.q[idx];
 
+	  case IVL_LPM_UFUNC:
+	    assert(idx < net->u_.ufunc.port_wid[0]);
+	    return net->u_.ufunc.pins[idx];
+
 	  default:
 	    assert(0);
 	    return 0;
@@ -730,11 +963,43 @@ extern "C" unsigned ivl_lpm_selects(ivl_lpm_t net)
       }
 }
 
+extern "C" int ivl_lpm_signed(ivl_lpm_t net)
+{
+      assert(net);
+      switch (net->type) {
+	  case IVL_LPM_FF:
+	  case IVL_LPM_RAM:
+	  case IVL_LPM_MUX:
+	    return 0;
+	  case IVL_LPM_ADD:
+	  case IVL_LPM_CMP_EQ:
+	  case IVL_LPM_CMP_GE:
+	  case IVL_LPM_CMP_GT:
+	  case IVL_LPM_CMP_NE:
+	  case IVL_LPM_DIVIDE:
+	  case IVL_LPM_MOD:
+	  case IVL_LPM_MULT:
+	  case IVL_LPM_SUB:
+	    return net->u_.arith.signed_flag;
+	  case IVL_LPM_SHIFTL:
+	  case IVL_LPM_SHIFTR:
+	    return net->u_.shift.signed_flag;
+	    return 0;
+	  case IVL_LPM_UFUNC:
+	    return 0;
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
+
 extern "C" unsigned ivl_lpm_size(ivl_lpm_t net)
 {
       switch (net->type) {
 	  case IVL_LPM_MUX:
 	    return net->u_.mux.size;
+	  case IVL_LPM_UFUNC:
+	    return net->u_.ufunc.ports - 1;
 	  default:
 	    assert(0);
 	    return 0;
@@ -768,6 +1033,8 @@ extern "C" unsigned ivl_lpm_width(ivl_lpm_t net)
 	  case IVL_LPM_SHIFTL:
 	  case IVL_LPM_SHIFTR:
 	    return net->u_.shift.width;
+	  case IVL_LPM_UFUNC:
+	    return net->u_.ufunc.port_wid[0];
 	  default:
 	    assert(0);
 	    return 0;
@@ -783,7 +1050,7 @@ extern "C" ivl_memory_t ivl_lpm_memory(ivl_lpm_t net)
 	  default:
 	    assert(0);
 	    return 0;
-      }      
+      }
 }
 
 extern "C" ivl_expr_t ivl_lval_mux(ivl_lval_t net)
@@ -810,6 +1077,14 @@ extern "C" ivl_memory_t ivl_lval_mem(ivl_lval_t net)
       return 0x0;
 }
 
+extern "C" ivl_variable_t ivl_lval_var(ivl_lval_t net)
+{
+      assert(net);
+      if (net->type_ == IVL_LVAL_VAR)
+	    return net->n.var;
+      return 0x0;
+}
+
 extern "C" unsigned ivl_lval_part_off(ivl_lval_t net)
 {
       assert(net);
@@ -921,6 +1196,24 @@ extern "C" ivl_signal_t ivl_nexus_ptr_sig(ivl_nexus_ptr_t net)
       return net->l.sig;
 }
 
+extern "C" const char* ivl_parameter_basename(ivl_parameter_t net)
+{
+      assert(net);
+      return net->basename;
+}
+
+extern "C" ivl_expr_t ivl_parameter_expr(ivl_parameter_t net)
+{
+      assert(net);
+      return net->value;
+}
+
+extern "C" ivl_scope_t ivl_parameter_scope(ivl_parameter_t net)
+{
+      assert(net);
+      return net->scope;
+}
+
 extern "C" ivl_process_type_t ivl_process_type(ivl_process_t net)
 {
       return net->type_;
@@ -936,14 +1229,36 @@ extern "C" ivl_statement_t ivl_process_stmt(ivl_process_t net)
       return net->stmt_;
 }
 
+extern "C" unsigned ivl_process_attr_cnt(ivl_process_t net)
+{
+      return net->nattr;
+}
+
+extern "C" ivl_attribute_t ivl_process_attr_val(ivl_process_t net,
+						unsigned idx)
+{
+      assert(idx < net->nattr);
+      return net->attr + idx;
+}
+
+extern "C" unsigned ivl_scope_attr_cnt(ivl_scope_t net)
+{
+      assert(net);
+      return net->nattr;
+}
+
+extern "C" ivl_attribute_t ivl_scope_attr_val(ivl_scope_t net,
+					      unsigned idx)
+{
+      assert(idx < net->nattr);
+      return net->attr + idx;
+}
+
 extern "C" const char* ivl_scope_basename(ivl_scope_t net)
 {
       assert(net);
 
-      if (net->parent == 0)
-	    return net->name_;
-
-      return basename(net->parent, net->name_);
+      return net->name_;
 }
 
 extern "C" int ivl_scope_children(ivl_scope_t net,
@@ -978,6 +1293,19 @@ extern "C" ivl_event_t ivl_scope_event(ivl_scope_t net, unsigned idx)
       return net->event_[idx];
 }
 
+extern "C" unsigned ivl_scope_vars(ivl_scope_t net)
+{
+      assert(net);
+      return net->nvar_;
+}
+
+extern "C" ivl_variable_t ivl_scope_var(ivl_scope_t net, unsigned idx)
+{
+      assert(net);
+      assert(idx < net->nvar_);
+      return net->var_[idx];
+}
+
 extern "C" unsigned ivl_scope_logs(ivl_scope_t net)
 {
       assert(net);
@@ -1017,9 +1345,60 @@ extern "C" ivl_memory_t ivl_scope_mem(ivl_scope_t net, unsigned idx)
       return net->mem_[idx];
 }
 
+static unsigned scope_name_len(ivl_scope_t net)
+{
+      unsigned len = 0;
+
+      for (ivl_scope_t cur = net ;  cur ;  cur = cur->parent)
+	    len += strlen(cur->name_) + 1;
+
+      return len;
+}
+
+static void push_scope_basename(ivl_scope_t net, char*buf)
+{
+      if (net->parent == 0) {
+	    strcpy(buf, net->name_);
+	    return;
+      }
+
+      push_scope_basename(net->parent, buf);
+      strcat(buf, ".");
+      strcat(buf, net->name_);
+}
+
 extern "C" const char* ivl_scope_name(ivl_scope_t net)
 {
-      return net->name_;
+      static char*name_buffer = 0;
+      static unsigned name_size = 0;
+
+      if (net->parent == 0)
+	    return net->name_;
+
+      unsigned needlen = scope_name_len(net);
+
+      if (name_size < needlen) {
+	    name_buffer = (char*)realloc(name_buffer, needlen);
+	    name_size = needlen;
+      }
+
+
+      push_scope_basename(net, name_buffer);
+
+      return name_buffer;
+}
+
+extern "C" unsigned ivl_scope_params(ivl_scope_t net)
+{
+      assert(net);
+      return net->nparam_;
+}
+
+extern "C" ivl_parameter_t ivl_scope_param(ivl_scope_t net, unsigned idx)
+{
+      assert(net);
+      assert(idx < net->nparam_);
+      return net->param_ + idx;
 }
 
 extern "C" ivl_scope_t ivl_scope_parent(ivl_scope_t net)
@@ -1056,6 +1435,12 @@ extern "C" ivl_signal_t ivl_scope_sig(ivl_scope_t net, unsigned idx)
       return net->sigs_[idx];
 }
 
+extern "C" int ivl_scope_time_units(ivl_scope_t net)
+{
+      assert(net);
+      return net->time_units;
+}
+
 extern "C" ivl_scope_type_t ivl_scope_type(ivl_scope_t net)
 {
       assert(net);
@@ -1070,28 +1455,53 @@ extern "C" const char* ivl_scope_tname(ivl_scope_t net)
 
 extern "C" const char* ivl_signal_attr(ivl_signal_t net, const char*key)
 {
-      if (net->nattr_ == 0)
+      if (net->nattr == 0)
 	    return 0;
 
-      assert(net->akey_);
-      assert(net->aval_);
+      for (unsigned idx = 0 ;  idx < net->nattr ;  idx += 1)
 
-      for (unsigned idx = 0 ;  idx < net->nattr_ ;  idx += 1)
-
-	    if (strcmp(key, net->akey_[idx]) == 0)
-		  return net->aval_[idx];
+	    if (strcmp(key, net->attr[idx].key) == 0)
+		  return net->attr[idx].type == IVL_ATT_STR
+			? net->attr[idx].val.str
+			: 0;
 
       return 0;
 }
 
+extern "C" unsigned ivl_signal_attr_cnt(ivl_signal_t net)
+{
+      return net->nattr;
+}
+
+extern "C" ivl_attribute_t ivl_signal_attr_val(ivl_signal_t net, unsigned idx)
+{
+      assert(idx < net->nattr);
+      return net->attr + idx;
+}
+
 extern "C" const char* ivl_signal_basename(ivl_signal_t net)
 {
-      return basename(net->scope_, net->name_);
+      return net->name_;
 }
 
 extern "C" const char* ivl_signal_name(ivl_signal_t net)
 {
-      return net->name_;
+      static char*name_buffer = 0;
+      static unsigned name_size = 0;
+
+      unsigned needlen = scope_name_len(net->scope_);
+      needlen += strlen(net->name_) + 2;
+
+      if (name_size < needlen) {
+	    name_buffer = (char*)realloc(name_buffer, needlen);
+	    name_size = needlen;
+      }
+
+      push_scope_basename(net->scope_, name_buffer);
+      strcat(name_buffer, ".");
+      strcat(name_buffer, net->name_);
+
+      return name_buffer;
 }
 
 extern "C" ivl_nexus_t ivl_signal_pin(ivl_signal_t net, unsigned idx)
@@ -1111,14 +1521,35 @@ extern "C" unsigned ivl_signal_pins(ivl_signal_t net)
       return net->width_;
 }
 
+extern "C" int ivl_signal_msb(ivl_signal_t net)
+{
+      assert(net->lsb_dist == 1 || net->lsb_dist == -1);
+      return net->lsb_index + net->lsb_dist * (net->width_ - 1);
+}
+
+extern "C" int ivl_signal_lsb(ivl_signal_t net)
+{
+      return net->lsb_index;
+}
+
 extern "C" ivl_signal_port_t ivl_signal_port(ivl_signal_t net)
 {
       return net->port_;
 }
 
+extern "C" int ivl_signal_local(ivl_signal_t net)
+{
+      return net->local_;
+}
+
 extern "C" int ivl_signal_signed(ivl_signal_t net)
 {
-      return net->signed_ != 0;
+      return net->signed_;
+}
+
+extern "C" int ivl_signal_integer(ivl_signal_t net)
+{
+      return net->isint_;
 }
 
 extern "C" ivl_signal_type_t ivl_signal_type(ivl_signal_t net)
@@ -1131,6 +1562,18 @@ extern "C" ivl_statement_type_t ivl_statement_type(ivl_statement_t net)
       return net->type_;
 }
 
+extern "C" ivl_scope_t ivl_stmt_block_scope(ivl_statement_t net)
+{
+      switch (net->type_) {
+	  case IVL_ST_BLOCK:
+	  case IVL_ST_FORK:
+	    return net->u_.block_.scope;
+	  default:
+	    assert(0);
+	    return 0;
+      }
+}
+
 extern "C" unsigned ivl_stmt_block_count(ivl_statement_t net)
 {
       switch (net->type_) {
@@ -1174,6 +1617,7 @@ extern "C" unsigned ivl_stmt_case_count(ivl_statement_t net)
 {
       switch (net->type_) {
 	  case IVL_ST_CASE:
+	  case IVL_ST_CASER:
 	  case IVL_ST_CASEX:
 	  case IVL_ST_CASEZ:
 	    return net->u_.case_.ncase;
@@ -1187,6 +1631,7 @@ extern "C" ivl_expr_t ivl_stmt_case_expr(ivl_statement_t net, unsigned idx)
 {
       switch (net->type_) {
 	  case IVL_ST_CASE:
+	  case IVL_ST_CASER:
 	  case IVL_ST_CASEX:
 	  case IVL_ST_CASEZ:
 	    assert(idx < net->u_.case_.ncase);
@@ -1202,6 +1647,7 @@ extern "C" ivl_statement_t ivl_stmt_case_stmt(ivl_statement_t net, unsigned idx)
 {
       switch (net->type_) {
 	  case IVL_ST_CASE:
+	  case IVL_ST_CASER:
 	  case IVL_ST_CASEX:
 	  case IVL_ST_CASEZ:
 	    assert(idx < net->u_.case_.ncase);
@@ -1220,6 +1666,7 @@ extern "C" ivl_expr_t ivl_stmt_cond_expr(ivl_statement_t net)
 	    return net->u_.condit_.cond_;
 
 	  case IVL_ST_CASE:
+	  case IVL_ST_CASER:
 	  case IVL_ST_CASEX:
 	  case IVL_ST_CASEZ:
 	    return net->u_.case_.cond;
@@ -1274,13 +1721,34 @@ extern "C" unsigned long ivl_stmt_delay_val(ivl_statement_t net)
       return net->u_.delay_.delay_;
 }
 
-extern "C" ivl_event_t ivl_stmt_event(ivl_statement_t net)
+extern "C" unsigned ivl_stmt_nevent(ivl_statement_t net)
 {
       switch (net->type_) {
 	  case IVL_ST_WAIT:
-	    return net->u_.wait_.event_;
+	    return net->u_.wait_.nevent;
+
 	  case IVL_ST_TRIGGER:
-	    return net->u_.trig_.event_;
+	    return 1;
+
+	  default:
+	    assert(0);
+      }
+      return 0;
+}
+
+extern "C" ivl_event_t ivl_stmt_events(ivl_statement_t net, unsigned idx)
+{
+      switch (net->type_) {
+	  case IVL_ST_WAIT:
+	    assert(idx < net->u_.wait_.nevent);
+	    if (net->u_.wait_.nevent == 1)
+		  return net->u_.wait_.event;
+	    else
+		  return net->u_.wait_.events[idx];
+
+	  case IVL_ST_TRIGGER:
+	    assert(idx == 0);
+	    return net->u_.wait_.event;
 	  default:
 	    assert(0);
       }
@@ -1345,6 +1813,9 @@ extern "C" unsigned ivl_stmt_lwidth(ivl_statement_t net)
 		case IVL_LVAL_MEM:
 		  sum += ivl_memory_width(ivl_lval_mem(cur));
 		  break;
+		case IVL_LVAL_VAR:
+		  sum += 0;
+		  break;
 		default:
 		  assert(0);
 	    }
@@ -1450,265 +1921,121 @@ extern "C" ivl_statement_t ivl_stmt_sub_stmt(ivl_statement_t net)
       return 0;
 }
 
+extern "C" const char* ivl_variable_name(ivl_variable_t net)
+{
+      assert(net);
+      return net->name;
+}
+
+extern "C" ivl_variable_type_t ivl_variable_type(ivl_variable_t net)
+{
+      assert(net);
+      return net->type;
+}
+
 /*
  * $Log: t-dll-api.cc,v $
- * Revision 1.75  2002/01/28 00:52:41  steve
- *  Add support for bit select of parameters.
- *  This leads to a NetESelect node and the
- *  vvp code generator to support that.
+ * Revision 1.108  2004/10/04 01:10:55  steve
+ *  Clean up spurious trailing white space.
  *
- * Revision 1.74  2002/01/03 04:19:01  steve
- *  Add structural modulus support down to vvp.
+ * Revision 1.107  2004/06/30 02:16:27  steve
+ *  Implement signed divide and signed right shift in nets.
  *
- * Revision 1.73  2001/12/06 03:11:00  steve
- *  Add ivl_logic_delay function to ivl_target.
+ * Revision 1.106  2003/12/03 02:46:24  steve
+ *  Add support for wait on list of named events.
  *
- * Revision 1.72  2001/11/14 03:28:49  steve
- *  DLL target support for force and release.
+ * Revision 1.105  2003/11/10 20:59:03  steve
+ *  Design::get_flag returns const char* instead of string.
  *
- * Revision 1.71  2001/11/01 04:25:31  steve
- *  ivl_target support for cassign.
+ * Revision 1.104  2003/09/03 23:33:30  steve
+ *  Pass FF synchronous set values to code generator.
  *
- * Revision 1.70  2001/10/31 05:24:52  steve
- *  ivl_target support for assign/deassign.
+ * Revision 1.103  2003/08/22 23:14:26  steve
+ *  Preserve variable ranges all the way to the vpi.
  *
- * Revision 1.69  2001/10/19 21:53:24  steve
- *  Support multiple root modules (Philip Blundell)
+ * Revision 1.102  2003/08/15 02:23:53  steve
+ *  Add synthesis support for synchronous reset.
  *
- * Revision 1.68  2001/10/16 02:19:27  steve
- *  Support IVL_LPM_DIVIDE for structural divide.
+ * Revision 1.101  2003/06/24 01:38:03  steve
+ *  Various warnings fixed.
  *
- * Revision 1.67  2001/09/16 22:19:42  steve
- *  Support attributes to logic gates.
+ * Revision 1.100  2003/06/23 01:25:44  steve
+ *  Module attributes make it al the way to ivl_target.
  *
- * Revision 1.66  2001/09/01 01:57:31  steve
- *  Make constants available through the design root
+ * Revision 1.99  2003/05/14 05:26:41  steve
+ *  Support real expressions in case statements.
  *
- * Revision 1.65  2001/08/31 22:58:40  steve
- *  Support DFF CE inputs.
+ * Revision 1.98  2003/04/22 04:48:30  steve
+ *  Support event names as expressions elements.
  *
- * Revision 1.64  2001/08/28 04:07:18  steve
- *  Add some ivl_target convenience functions.
+ * Revision 1.97  2003/04/11 05:18:08  steve
+ *  Handle signed magnitude compare all the
+ *  way through to the vvp code generator.
  *
- * Revision 1.63  2001/08/25 23:50:03  steve
- *  Change the NetAssign_ class to refer to the signal
- *  instead of link into the netlist. This is faster
- *  and uses less space. Make the NetAssignNB carry
- *  the delays instead of the NetAssign_ lval objects.
+ * Revision 1.96  2003/03/10 23:40:53  steve
+ *  Keep parameter constants for the ivl_target API.
  *
- *  Change the vvp code generator to support multiple
- *  l-values, i.e. concatenations of part selects.
+ * Revision 1.95  2003/03/06 04:32:40  steve
+ *  Wrong sense of need compared to have.
  *
- * Revision 1.62  2001/08/10 00:40:45  steve
- *  tgt-vvp generates code that skips nets as inputs.
+ * Revision 1.94  2003/03/06 01:24:37  steve
+ *  Obsolete the ivl_event_name function.
  *
- * Revision 1.61  2001/07/28 01:17:40  steve
- *  Support getting the signal from IVL_EX_SIGNAL expressions.
+ * Revision 1.93  2003/03/06 00:28:42  steve
+ *  All NetObj objects have lex_string base names.
  *
- * Revision 1.60  2001/07/27 04:51:44  steve
- *  Handle part select expressions as variants of
- *  NetESignal/IVL_EX_SIGNAL objects, instead of
- *  creating new and useless temporary signals.
+ * Revision 1.92  2003/03/03 02:22:41  steve
+ *  Scope names stored only as basename.
  *
- * Revision 1.59  2001/07/27 02:41:55  steve
- *  Fix binding of dangling function ports. do not elide them.
+ * Revision 1.91  2003/02/26 01:29:24  steve
+ *  LPM objects store only their base names.
  *
- * Revision 1.58  2001/07/25 03:10:49  steve
- *  Create a config.h.in file to hold all the config
- *  junk, and support gcc 3.0. (Stephan Boettcher)
+ * Revision 1.90  2003/01/26 21:15:59  steve
+ *  Rework expression parsing and elaboration to
+ *  accommodate real/realtime values and expressions.
  *
- * Revision 1.57  2001/07/22 00:17:49  steve
- *  Support the NetESubSignal expressions in vvp.tgt.
+ * Revision 1.89  2002/12/21 00:55:58  steve
+ *  The $time system task returns the integer time
+ *  scaled to the local units. Change the internal
+ *  implementation of vpiSystemTime the $time functions
+ *  to properly account for this. Also add $simtime
+ *  to get the simulation time.
  *
- * Revision 1.56  2001/07/19 04:55:06  steve
- *  Support calculated delays in vvp.tgt.
+ * Revision 1.88  2002/10/23 01:47:17  steve
+ *  Fix synth2 handling of aset/aclr signals where
+ *  flip-flops are split by begin-end blocks.
  *
- * Revision 1.55  2001/07/07 20:20:10  steve
- *  Pass parameters to system functions.
+ * Revision 1.87  2002/09/26 03:18:04  steve
+ *  Generate vvp code for asynch set/reset of NetFF.
  *
- * Revision 1.54  2001/07/07 03:01:37  steve
- *  Detect and make available to t-dll the right shift.
+ * Revision 1.86  2002/08/12 01:35:00  steve
+ *  conditional ident string using autoconfig.
  *
- * Revision 1.53  2001/07/04 22:59:25  steve
- *  handle left shifter in dll output.
+ * Revision 1.85  2002/08/07 02:17:14  steve
+ *  Dead code.
  *
- * Revision 1.52  2001/06/30 23:03:16  steve
- *  support fast programming by only writing the bits
- *  that are listed in the input file.
+ * Revision 1.84  2002/08/05 04:18:45  steve
+ *  Store only the base name of memories.
  *
- * Revision 1.51  2001/06/16 23:45:05  steve
- *  Add support for structural multiply in t-dll.
- *  Add code generators and vvp support for both
- *  structural and behavioral multiply.
+ * Revision 1.83  2002/07/05 21:26:17  steve
+ *  Avoid emitting to vvp local net symbols.
  *
- * Revision 1.50  2001/06/16 02:41:41  steve
- *  Generate code to support memory access in continuous
- *  assignment statements. (Stephan Boettcher)
+ * Revision 1.82  2002/06/21 04:59:35  steve
+ *  Carry integerness throughout the compilation.
  *
- * Revision 1.49  2001/06/15 04:14:19  steve
- *  Generate vvp code for GT and GE comparisons.
+ * Revision 1.81  2002/05/27 00:08:45  steve
+ *  Support carrying the scope of named begin-end
+ *  blocks down to the code generator, and have
+ *  the vvp code generator use that to support disable.
  *
- * Revision 1.48  2001/06/07 03:09:37  steve
- *  support subtraction in tgt-vvp.
+ * Revision 1.80  2002/05/26 01:39:03  steve
+ *  Carry Verilog 2001 attributes with processes,
+ *  all the way through to the ivl_target API.
  *
- * Revision 1.47  2001/06/07 02:12:43  steve
- *  Support structural addition.
- *
- * Revision 1.46  2001/05/20 01:06:16  steve
- *  stub ivl_expr_parms for sfunctions.
- *
- * Revision 1.45  2001/05/17 04:37:02  steve
- *  Behavioral ternary operators for vvp.
- *
- * Revision 1.44  2001/05/08 23:59:33  steve
- *  Add ivl and vvp.tgt support for memories in
- *  expressions and l-values. (Stephan Boettcher)
- *
- * Revision 1.43  2001/05/06 17:48:20  steve
- *  Support memory objects. (Stephan Boettcher)
- *
- * Revision 1.42  2001/04/29 23:17:38  steve
- *  Carry drive strengths in the ivl_nexus_ptr_t, and
- *  handle constant devices in targets.'
- *
- * Revision 1.41  2001/04/26 05:12:02  steve
- *  Implement simple MUXZ for ?: operators.
- *
- * Revision 1.40  2001/04/22 23:09:46  steve
- *  More UDP consolidation from Stephan Boettcher.
- *
- * Revision 1.39  2001/04/21 00:55:46  steve
- *  Generate code for disable.
- *
- * Revision 1.38  2001/04/15 02:58:11  steve
- *  vvp support for <= with internal delay.
- *
- * Revision 1.37  2001/04/06 02:28:02  steve
- *  Generate vvp code for functions with ports.
- *
- * Revision 1.36  2001/04/05 03:20:57  steve
- *  Generate vvp code for the repeat statement.
- *
- * Revision 1.35  2001/04/05 01:12:28  steve
- *  Get signed compares working correctly in vvp.
- *
- * Revision 1.34  2001/04/04 04:50:35  steve
- *  Support forever loops in the tgt-vvp target.
- *
- * Revision 1.33  2001/04/03 04:50:37  steve
- *  Support non-blocking assignments.
- *
- * Revision 1.32  2001/04/02 02:28:12  steve
- *  Generate code for task calls.
- *
- * Revision 1.31  2001/04/02 00:28:35  steve
- *  Support the scope expression node.
- *
- * Revision 1.30  2001/04/01 06:52:27  steve
- *  support the NetWhile statement.
- *
- * Revision 1.29  2001/04/01 01:48:21  steve
- *  Redesign event information to support arbitrary edge combining.
- *
- * Revision 1.28  2001/03/31 17:36:38  steve
- *  Generate vvp code for case statements.
- *
- * Revision 1.27  2001/03/30 05:49:52  steve
- *  Generate code for fork/join statements.
- *
- * Revision 1.26  2001/03/29 03:47:38  steve
- *  Behavioral trigger statements.
- *
- * Revision 1.25  2001/03/29 02:52:39  steve
- *  Add unary ~ operator to tgt-vvp.
- *
- * Revision 1.24  2001/03/28 06:07:39  steve
- *  Add the ivl_event_t to ivl_target, and use that to generate
- *  .event statements in vvp way ahead of the thread that uses it.
- *
- * Revision 1.23  2001/03/27 06:27:40  steve
- *  Generate code for simple @ statements.
- *
- * Revision 1.22  2001/03/20 01:44:13  steve
- *  Put processes in the proper scope.
- *
- * Revision 1.21  2001/01/15 00:47:02  steve
- *  Pass scope type information to the target module.
- *
- * Revision 1.20  2001/01/15 00:05:39  steve
- *  Add client data pointer for scope and process scanners.
- *
- * Revision 1.19  2000/12/05 06:29:33  steve
- *  Make signal attributes available to ivl_target API.
- *
- * Revision 1.18  2000/11/12 17:47:29  steve
- *  flip-flop pins for ivl_target API.
- *
- * Revision 1.17  2000/11/11 00:03:36  steve
- *  Add support for the t-dll backend grabing flip-flops.
- *
- * Revision 1.16  2000/10/28 22:32:34  steve
- *  API for concatenation expressions.
- *
- * Revision 1.15  2000/10/25 05:41:24  steve
- *  Get target signal from nexus_ptr.
- *
- * Revision 1.14  2000/10/18 20:04:39  steve
- *  Add ivl_lval_t and support for assignment l-values.
- *
- * Revision 1.13  2000/10/16 22:44:54  steve
- *  Stubs so that cygwin port will link ivl.
- *
- * Revision 1.12  2000/10/15 04:46:23  steve
- *  Scopes and processes are accessible randomly from
- *  the design, and signals and logic are accessible
- *  from scopes. Remove the target calls that are no
- *  longer needed.
- *
- *  Add the ivl_nexus_ptr_t and the means to get at
- *  them from nexus objects.
- *
- *  Give names to methods that manipulate the ivl_design_t
- *  type more consistent names.
- *
- * Revision 1.11  2000/10/08 04:01:54  steve
- *  Back pointers in the nexus objects into the devices
- *  that point to it.
- *
- *  Collect threads into a list in the design.
- *
- * Revision 1.10  2000/10/06 23:46:50  steve
- *  ivl_target updates, including more complete
- *  handling of ivl_nexus_t objects. Much reduced
- *  dependencies on pointers to netlist objects.
- *
- * Revision 1.9  2000/10/05 05:03:01  steve
- *  xor and constant devices.
- *
- * Revision 1.8  2000/09/30 02:18:15  steve
- *  ivl_expr_t support for binary operators,
- *  Create a proper ivl_scope_t object.
- *
- * Revision 1.7  2000/09/26 00:30:07  steve
- *  Add EX_NUMBER and ST_TRIGGER to dll-api.
- *
- * Revision 1.6  2000/09/24 15:46:00  steve
- *  API access to signal type and port type.
- *
- * Revision 1.5  2000/09/24 02:21:53  steve
- *  Add support for signal expressions.
- *
- * Revision 1.4  2000/09/23 05:15:07  steve
- *  Add enough tgt-verilog code to support hello world.
- *
- * Revision 1.3  2000/09/22 03:58:30  steve
- *  Access to the name of a system task call.
- *
- * Revision 1.2  2000/09/19 04:15:27  steve
- *  Introduce the means to get statement types.
- *
- * Revision 1.1  2000/09/18 01:24:32  steve
- *  Get the structure for ivl_statement_t worked out.
+ *  Divide signal reference counts between rval
+ *  and lval references.
  *
+ * Revision 1.79  2002/05/24 04:36:23  steve
+ *  Verilog 2001 attriubtes on nets/wires.
  */
 
