@@ -17,8 +17,8 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT) && !defined(macintosh)
-#ident "$Id: t-dll-proc.cc,v 1.41 2002/01/19 19:02:08 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: t-dll-proc.cc,v 1.65 2004/10/04 01:10:55 steve Exp $"
 #endif
 
 # include "config.h"
@@ -54,6 +54,8 @@ bool dll_target::process(const NetProcTop*net)
 	/* Save the scope of the process. */
       obj->scope_ = lookup_scope_(net->scope());
 
+      obj->nattr = net->attr_cnt();
+      obj->attr = fill_in_attributes(net);
 
 	/* This little bit causes the process to be completely
 	   generated so that it can be passed to the DLL. The
@@ -97,7 +99,7 @@ void dll_target::task_def(const NetScope*net)
 
 }
 
-void dll_target::func_def(const NetScope*net)
+bool dll_target::func_def(const NetScope*net)
 {
       ivl_scope_t scope = lookup_scope_(net);
       const NetFuncDef*def = net->func_def();
@@ -111,12 +113,35 @@ void dll_target::func_def(const NetScope*net)
       scope->def = stmt_cur_;
       stmt_cur_ = 0;
 
-      scope->ports = def->port_count();
+      scope->ports = def->port_count() + 1;
       if (scope->ports > 0) {
 	    scope->port = new ivl_signal_t[scope->ports];
-	    for (unsigned idx = 0 ;  idx < scope->ports ;  idx += 1)
-		  scope->port[idx] = find_signal(des_, def->port(idx));
+	    for (unsigned idx = 1 ;  idx < scope->ports ;  idx += 1)
+		  scope->port[idx] = find_signal(des_, def->port(idx-1));
       }
+
+	/* FIXME: the ivl_target API expects port-0 to be the output
+	   port. This assumes that the return value is a signal, which
+	   is *not* correct. Someday, I'm going to have to change
+	   this, but that will break code generators that use this
+	   result. */
+      if (const NetNet*ret_sig = def->return_sig()) {
+	    scope->port[0] = find_signal(des_, ret_sig);
+	    return true;
+      }
+
+      if (const NetVariable*ret_var = def->return_var()) {
+	    cerr << ret_var->get_line() << ": internal error: "
+		 << "Function " << net->name() << " has an unsupported "
+		 << "return type." << endl;
+	    return false;
+      }
+
+      cerr << "?:0" << ": internal error: "
+	   << "Function " << net->name() << " has a return type"
+	   << " that I do not understand." << endl;
+
+      return false;
 }
 
 /*
@@ -134,20 +159,47 @@ void dll_target::proc_assign(const NetAssign*net)
       stmt_cur_->u_.assign_.lval_ = new struct ivl_lval_s[cnt];
       stmt_cur_->u_.assign_.delay = 0;
 
+	/* The assignment may have multiple concatenated
+	   l-values. Scan them and accumulate an ivl_lval_t list. */
       for (unsigned idx = 0 ;  idx < cnt ;  idx += 1) {
 	    struct ivl_lval_s*cur = stmt_cur_->u_.assign_.lval_ + idx;
 	    const NetAssign_*asn = net->l_val(idx);
 
 	    cur->width_ = asn->lwidth();
 	    cur->loff_  = asn->get_loff();
-	    cur->type_  = IVL_LVAL_REG;
-	    cur->n.sig  = find_signal(des_, asn->sig());
+	    if (asn->sig()) {
+		  cur->type_  = IVL_LVAL_REG;
+		  cur->n.sig  = find_signal(des_, asn->sig());
+
+		  cur->idx = 0;
+		  if (asn->bmux()) {
+			assert(expr_ == 0);
+			asn->bmux()->expr_scan(this);
+
+			if (cur->n.sig->lsb_index != 0)
+			      sub_off_from_expr_(asn->sig()->lsb());
+			if (cur->n.sig->lsb_dist != 1)
+			      mul_expr_by_const_(cur->n.sig->lsb_dist);
+
+			cur->type_ = IVL_LVAL_MUX;
+			cur->idx = expr_;
+			expr_ = 0;
+		  }
+
+	    } else if (asn->var()) {
+		  cur->type_ = IVL_LVAL_VAR;
+		  cur->idx = 0;
+		  cur->n.var = find_variable(des_, asn->var());
+
+	    } else {
+		  assert(asn->mem());
+		  cur->type_ = IVL_LVAL_MEM;
+		  cur->n.mem = find_memory(des_, asn->mem());
+		  assert(cur->n.mem);
+		  cur->width_ = ivl_memory_width(cur->n.mem);
 
-	    cur->idx = 0;
-	    if (asn->bmux()) {
 		  assert(expr_ == 0);
 		  asn->bmux()->expr_scan(this);
-		  cur->type_ = IVL_LVAL_MUX;
 		  cur->idx = expr_;
 		  expr_ = 0;
 	    }
@@ -157,6 +209,13 @@ void dll_target::proc_assign(const NetAssign*net)
       net->rval()->expr_scan(this);
       stmt_cur_->u_.assign_.rval_ = expr_;
       expr_ = 0;
+
+      const NetExpr*del = net->get_delay();
+      if (del) {
+	    del->expr_scan(this);
+	    stmt_cur_->u_.assign_.delay = expr_;
+	    expr_ = 0;
+      }
 }
 
 
@@ -164,7 +223,7 @@ void dll_target::proc_assign_nb(const NetAssignNB*net)
 {
       unsigned cnt = net->l_val_count();
 
-      unsigned long delay_val = net->rise_time();
+      const NetExpr* delay_exp = net->get_delay();
       assert(stmt_cur_);
       assert(stmt_cur_->type_ == IVL_ST_NONE);
 
@@ -178,18 +237,43 @@ void dll_target::proc_assign_nb(const NetAssignNB*net)
 	    struct ivl_lval_s*cur = stmt_cur_->u_.assign_.lval_ + idx;
 	    const NetAssign_*asn = net->l_val(idx);
 
-	    cur->type_ = IVL_LVAL_REG;
 	    cur->width_ = asn->lwidth();
 	    cur->loff_  = asn->get_loff();
-	    cur->n.sig = find_signal(des_, asn->sig());
+	    if (asn->sig()) {
+		  cur->type_ = IVL_LVAL_REG;
+		  cur->n.sig = find_signal(des_, asn->sig());
+
+		  cur->idx = 0;
+		  if (asn->bmux()) {
+			assert(expr_ == 0);
+			asn->bmux()->expr_scan(this);
+
+			if (cur->n.sig->lsb_index != 0)
+			      sub_off_from_expr_(asn->sig()->lsb());
+			if (cur->n.sig->lsb_dist != 1)
+			      mul_expr_by_const_(cur->n.sig->lsb_dist);
+
+			cur->type_ = IVL_LVAL_MUX;
+			cur->idx = expr_;
+			expr_ = 0;
+		  }
+	    } else if (asn->mem()) {
+		  assert(asn->mem());
+		  cur->type_ = IVL_LVAL_MEM;
+		  cur->n.mem = find_memory(des_, asn->mem());
+		  assert(cur->n.mem);
+		  cur->width_ = ivl_memory_width(cur->n.mem);
 
-	    cur->idx = 0;
-	    if (asn->bmux()) {
 		  assert(expr_ == 0);
 		  asn->bmux()->expr_scan(this);
-		  cur->type_ = IVL_LVAL_MUX;
 		  cur->idx = expr_;
 		  expr_ = 0;
+	    } else {
+		  assert(asn->var());
+		  cur->type_ = IVL_LVAL_VAR;
+		  cur->idx = 0;
+		  cur->n.var = find_variable(des_, asn->var());
+
 	    }
       }
 
@@ -198,81 +282,19 @@ void dll_target::proc_assign_nb(const NetAssignNB*net)
       stmt_cur_->u_.assign_.rval_ = expr_;
       expr_ = 0;
 
-      if (delay_val > 0) {
+      if (const NetEConst*delay_num = dynamic_cast<const NetEConst*>(delay_exp)) {
+	    verinum val = delay_num->value();
 	    ivl_expr_t de = new struct ivl_expr_s;
 	    de->type_ = IVL_EX_ULONG;
 	    de->width_  = 8 * sizeof(unsigned long);
 	    de->signed_ = 0;
-	    de->u_.ulong_.value = delay_val;
+	    de->u_.ulong_.value = val.as_ulong();
 	    stmt_cur_->u_.assign_.delay = de;
-      }
-}
 
-void dll_target::proc_assign_mem(const NetAssignMem*net)
-{
-      assert(stmt_cur_);
-      assert(stmt_cur_->type_ == IVL_ST_NONE);
-
-      stmt_cur_->type_ = IVL_ST_ASSIGN;
-
-      stmt_cur_->u_.assign_.lvals_ = 1;
-      stmt_cur_->u_.assign_.lval_  = new struct ivl_lval_s[1];
-      stmt_cur_->u_.assign_.delay  = 0;
-      struct ivl_lval_s*cur = stmt_cur_->u_.assign_.lval_;
-
-      cur->type_ = IVL_LVAL_MEM;
-      cur->n.mem = lookup_memory_(net->memory());
-      assert(cur->n.mem);
-      cur->width_ = ivl_memory_width(cur->n.mem);
-      
-      assert(expr_ == 0);
-
-      net->index()->expr_scan(this);
-      cur->type_ = IVL_LVAL_MEM;
-      cur->idx = expr_;
-      expr_ = 0;
-
-      net->rval()->expr_scan(this);
-      stmt_cur_->u_.assign_.rval_ = expr_;
-      expr_ = 0;
-}
-
-void dll_target::proc_assign_mem_nb(const NetAssignMemNB*net)
-{
-      assert(stmt_cur_);
-      assert(stmt_cur_->type_ == IVL_ST_NONE);
-
-      stmt_cur_->type_ = IVL_ST_ASSIGN_NB;
-
-      stmt_cur_->u_.assign_.lvals_ = 1;
-      stmt_cur_->u_.assign_.lval_  = new struct ivl_lval_s[1];
-      stmt_cur_->u_.assign_.delay  = 0;
-      struct ivl_lval_s*cur = stmt_cur_->u_.assign_.lval_;
-
-      cur->type_ = IVL_LVAL_MEM;
-      cur->n.mem = lookup_memory_(net->memory());
-      cur->width_ = ivl_memory_width(cur->n.mem);
-      
-      assert(expr_ == 0);
-
-      net->index()->expr_scan(this);
-      cur->type_ = IVL_LVAL_MEM;
-      cur->idx = expr_;
-      expr_ = 0;
-
-      net->rval()->expr_scan(this);
-      stmt_cur_->u_.assign_.rval_ = expr_;
-      expr_ = 0;
-
-      unsigned long delay_val = 0;
-      
-      if (delay_val > 0) {
-	    ivl_expr_t de = new struct ivl_expr_s;
-	    de->type_ = IVL_EX_ULONG;
-	    de->width_  = 8 * sizeof(unsigned long);
-	    de->signed_ = 0;
-	    de->u_.ulong_.value = delay_val;
-	    stmt_cur_->u_.assign_.delay = de;
+      } else if (delay_exp != 0) {
+	    delay_exp->expr_scan(this);
+	    stmt_cur_->u_.assign_.delay = expr_;
+	    expr_ = 0;
       }
 }
 
@@ -295,7 +317,7 @@ bool dll_target::proc_block(const NetBlock*net)
 
 	/* If there is exactly one statement, there is no need for the
 	   block wrapper, generate the contained statement instead. */
-      if (count == 1) {
+      if ((count == 1) && (net->subscope() == 0)) {
 	    return net->proc_first()->emit_proc(this);
       }
 
@@ -311,6 +333,11 @@ bool dll_target::proc_block(const NetBlock*net)
       stmt_cur_->u_.block_.stmt_ = (struct ivl_statement_s*)
 	    calloc(count, sizeof(struct ivl_statement_s));
 
+      if (net->subscope())
+	    stmt_cur_->u_.block_.scope = lookup_scope_(net->subscope());
+      else
+	    stmt_cur_->u_.block_.scope = 0;
+
       struct ivl_statement_s*save_cur_ = stmt_cur_;
       unsigned idx = 0;
       bool flag = true;
@@ -358,6 +385,11 @@ void dll_target::proc_case(const NetCase*net)
       stmt_cur_->u_.case_.cond = expr_;
       expr_ = 0;
 
+	/* If the condition expression is a real valued expression,
+	   then change the case statement to a CASER statement. */
+      if (stmt_cur_->u_.case_.cond->value_ == IVL_VT_REAL)
+	    stmt_cur_->type_ = IVL_ST_CASER;
+
       unsigned ncase = net->nitems();
       stmt_cur_->u_.case_.ncase = ncase;
 
@@ -535,18 +567,27 @@ bool dll_target::proc_force(const NetForce*net)
       assert(lsig);
       ivl_signal_t sig = find_signal(des_, lsig);
       assert(sig);
-      if (sig->type_ != IVL_SIT_REG) {
+
+      ivl_lval_type_t ltype;
+      switch (sig->type_) {
+	  case IVL_SIT_REG:
+	    ltype = IVL_LVAL_REG;
+	    break;
+	  case IVL_SIT_TRI:
+	  case IVL_SIT_TRI0:
+	  case IVL_SIT_TRI1:
+	    ltype = IVL_LVAL_NET;
+	    break;
+	  default:
 	    cerr << net->get_line() << ": internal error: Sorry, "
 		 << "force to nets not supported by this target."
 		 << endl;
 	    return false;
       }
 
-      assert(sig->type_ == IVL_SIT_REG);
-
       stmt_cur_->u_.cassign_.lval[0].width_ = lsig->pin_count();
       stmt_cur_->u_.cassign_.lval[0].loff_  = 0;
-      stmt_cur_->u_.cassign_.lval[0].type_  = IVL_LVAL_REG;
+      stmt_cur_->u_.cassign_.lval[0].type_  = ltype;
       stmt_cur_->u_.cassign_.lval[0].idx    = 0;
       stmt_cur_->u_.cassign_.lval[0].n.sig  = sig;
 
@@ -589,18 +630,45 @@ bool dll_target::proc_release(const NetRelease*net)
 
       stmt_cur_->type_ = IVL_ST_RELEASE;
 
+	/* If there is no signal attached to the release, then it is
+	   the victim of an elided net. In that case, simply state
+	   that there are no lvals, and that's all. */
+      const NetNet*lsig = net->lval();
+      if (lsig == 0) {
+	    stmt_cur_->u_.cassign_.lvals = 0;
+	    return true;
+      }
+
+      assert(lsig);
       stmt_cur_->u_.cassign_.lvals = 1;
       stmt_cur_->u_.cassign_.lval = (struct ivl_lval_s*)
 	    calloc(1, sizeof(struct ivl_lval_s));
 
-      const NetNet*lsig = net->lval();
+
       ivl_signal_t sig = find_signal(des_, lsig);
       assert(sig);
-      assert(sig->type_ == IVL_SIT_REG);
+
+      ivl_lval_type_t ltype;
+      switch (sig->type_) {
+	  case IVL_SIT_REG:
+	    ltype = IVL_LVAL_REG;
+	    break;
+	  case IVL_SIT_TRI:
+	  case IVL_SIT_TRI0:
+	  case IVL_SIT_TRI1:
+	    ltype = IVL_LVAL_NET;
+	    break;
+	  default:
+	    cerr << net->get_line() << ": internal error: Sorry, "
+		 << "force/release to nets not supported by this target."
+		 << endl;
+	    return false;
+      }
+
 
       stmt_cur_->u_.cassign_.lval[0].width_ = lsig->pin_count();
       stmt_cur_->u_.cassign_.lval[0].loff_  = 0;
-      stmt_cur_->u_.cassign_.lval[0].type_  = IVL_LVAL_REG;
+      stmt_cur_->u_.cassign_.lval[0].type_  = ltype;
       stmt_cur_->u_.cassign_.lval[0].idx    = 0;
       stmt_cur_->u_.cassign_.lval[0].n.sig  = sig;
 
@@ -638,7 +706,8 @@ void dll_target::proc_stask(const NetSTask*net)
       assert(stmt_cur_->type_ == IVL_ST_NONE);
 
       stmt_cur_->type_ = IVL_ST_STASK;
-      stmt_cur_->u_.stask_.name_ = strdup(net->name());
+	/* System task names are lex_strings strings. */
+      stmt_cur_->u_.stask_.name_ = net->name();
       stmt_cur_->u_.stask_.nparm_= nparms;
       stmt_cur_->u_.stask_.parms_= (ivl_expr_t*)
 	    calloc(nparms, sizeof(ivl_expr_t));
@@ -658,6 +727,7 @@ bool dll_target::proc_trigger(const NetEvTrig*net)
       assert(stmt_cur_->type_ == IVL_ST_NONE);
 
       stmt_cur_->type_ = IVL_ST_TRIGGER;
+      stmt_cur_->u_.wait_.nevent = 1;
 
 	/* Locate the event by name. Save the ivl_event_t in the
 	   statement so that the generator can find it easily. */
@@ -667,7 +737,7 @@ bool dll_target::proc_trigger(const NetEvTrig*net)
       for (unsigned idx = 0 ;  idx < ev_scope->nevent_ ;  idx += 1) {
 	    const char*ename = ivl_event_basename(ev_scope->event_[idx]);
 	    if (strcmp(ev->name(), ename) == 0) {
-		  stmt_cur_->u_.wait_.event_ = ev_scope->event_[idx];
+		  stmt_cur_->u_.wait_.event = ev_scope->event_[idx];
 		  break;
 	    }
       }
@@ -694,65 +764,74 @@ bool dll_target::proc_wait(const NetEvWait*net)
       stmt_cur_->u_.wait_.stmt_ = (struct ivl_statement_s*)
 	    calloc(1, sizeof(struct ivl_statement_s));
 
-      if (net->nevents() != 1) {
-	    cerr << "internal error: multiple events not supported." << endl;
-	    return false;
+      stmt_cur_->u_.wait_.nevent = net->nevents();
+      if (net->nevents() > 1) {
+	    stmt_cur_->u_.wait_.events = (ivl_event_t*)
+		  calloc(net->nevents(), sizeof(ivl_event_t*));
       }
 
-	/* Locate the event by name. Save the ivl_event_t in the
-	   statement so that the generator can find it easily. */
-      const NetEvent*ev = net->event(0);
-      ivl_scope_t ev_scope = lookup_scope_(ev->scope());
+      for (unsigned edx = 0 ;  edx < net->nevents() ;  edx += 1) {
 
-      for (unsigned idx = 0 ;  idx < ev_scope->nevent_ ;  idx += 1) {
-	    const char*ename = ivl_event_basename(ev_scope->event_[idx]);
-	    if (strcmp(ev->name(), ename) == 0) {
-		  stmt_cur_->u_.wait_.event_ = ev_scope->event_[idx];
-		  break;
+	      /* Locate the event by name. Save the ivl_event_t in the
+		 statement so that the generator can find it easily. */
+	    const NetEvent*ev = net->event(edx);
+	    ivl_scope_t ev_scope = lookup_scope_(ev->scope());
+	    ivl_event_t ev_tmp=0;
+
+	    for (unsigned idx = 0 ;  idx < ev_scope->nevent_ ;  idx += 1) {
+		  const char*ename = ivl_event_basename(ev_scope->event_[idx]);
+		  if (strcmp(ev->name(), ename) == 0) {
+			ev_tmp = ev_scope->event_[idx];
+			break;
+		  }
+	    }
+	    // XXX should we assert(ev_tmp)?
+
+	    if (net->nevents() == 1)
+		  stmt_cur_->u_.wait_.event = ev_tmp;
+	    else
+		  stmt_cur_->u_.wait_.events[edx] = ev_tmp;
+
+	      /* If this is an event with a probe, then connect up the
+		 pins. This wasn't done during the ::event method because
+		 the signals weren't scanned yet. */
+
+	    if (ev->nprobe() >= 1) {
+		  unsigned iany = 0;
+		  unsigned ineg = ev_tmp->nany;
+		  unsigned ipos = ineg + ev_tmp->nneg;
+
+		  for (unsigned idx = 0 ;  idx < ev->nprobe() ;  idx += 1) {
+			const NetEvProbe*pr = ev->probe(idx);
+			unsigned base = 0;
+
+			switch (pr->edge()) {
+			    case NetEvProbe::ANYEDGE:
+			      base = iany;
+			      iany += pr->pin_count();
+			      break;
+			    case NetEvProbe::NEGEDGE:
+			      base = ineg;
+			      ineg += pr->pin_count();
+			      break;
+			    case NetEvProbe::POSEDGE:
+			      base = ipos;
+			      ipos += pr->pin_count();
+			      break;
+			}
+
+			for (unsigned bit = 0
+				   ; bit < pr->pin_count()
+				   ; bit += 1) {
+			      ivl_nexus_t nex = (ivl_nexus_t)
+				    pr->pin(bit).nexus()->t_cookie();
+			      assert(nex);
+			      ev_tmp->pins[base+bit] = nex;
+			}
+		  }
 	    }
       }
 
-	/* If this is an event with a probe, then connect up the
-	   pins. This wasn't done during the ::event method because
-	   the signals weren't scanned yet. */
-
-      if (ev->nprobe() >= 1) {
-	    const NetEvProbe*pr = ev->probe(0);
-	    ivl_event_t evnt = stmt_cur_->u_.wait_.event_;
-
-	    unsigned iany = 0;
-	    unsigned ineg = evnt->nany;
-	    unsigned ipos = ineg + evnt->nneg;
-
-	    for (unsigned idx = 0 ;  idx < ev->nprobe() ;  idx += 1) {
-		  const NetEvProbe*pr = ev->probe(idx);
-		  unsigned base = 0;
-
-		  switch (pr->edge()) {
-		      case NetEvProbe::ANYEDGE:
-			base = iany;
-			iany += pr->pin_count();
-			break;
-		      case NetEvProbe::NEGEDGE:
-			base = ineg;
-			ineg += pr->pin_count();
-			break;
-		      case NetEvProbe::POSEDGE:
-			base = ipos;
-			ipos += pr->pin_count();
-			break;
-		  }
-
-		  for (unsigned bit = 0;  bit < pr->pin_count(); bit += 1) {
-			ivl_nexus_t nex = (ivl_nexus_t)
-			      pr->pin(bit).nexus()->t_cookie();
-			assert(nex);
-			evnt->pins[base+bit] = nex;
-		  }
-	    }
-
-      }	    
-
 	/* The ivl_statement_t for the wait statement is not complete
 	   until we calculate the sub-statement. */
 
@@ -793,6 +872,99 @@ void dll_target::proc_while(const NetWhile*net)
 
 /*
  * $Log: t-dll-proc.cc,v $
+ * Revision 1.65  2004/10/04 01:10:55  steve
+ *  Clean up spurious trailing white space.
+ *
+ * Revision 1.64  2004/05/31 23:34:39  steve
+ *  Rewire/generalize parsing an elaboration of
+ *  function return values to allow for better
+ *  speed and more type support.
+ *
+ * Revision 1.63  2004/05/19 03:18:40  steve
+ *  Add ivl_target support for non-blocking assign of real.
+ *
+ * Revision 1.62  2003/12/19 01:27:10  steve
+ *  Fix various unsigned compare warnings.
+ *
+ * Revision 1.61  2003/12/03 02:46:24  steve
+ *  Add support for wait on list of named events.
+ *
+ * Revision 1.60  2003/06/24 01:38:03  steve
+ *  Various warnings fixed.
+ *
+ * Revision 1.59  2003/05/14 05:26:41  steve
+ *  Support real expressions in case statements.
+ *
+ * Revision 1.58  2003/05/07 19:56:20  steve
+ *  Improve internal error message.
+ *
+ * Revision 1.57  2003/03/01 06:25:30  steve
+ *  Add the lex_strings string handler, and put
+ *  scope names and system task/function names
+ *  into this table. Also, permallocate event
+ *  names from the beginning.
+ *
+ * Revision 1.56  2003/01/30 16:23:08  steve
+ *  Spelling fixes.
+ *
+ * Revision 1.55  2003/01/26 21:15:59  steve
+ *  Rework expression parsing and elaboration to
+ *  accommodate real/realtime values and expressions.
+ *
+ * Revision 1.54  2002/08/19 00:06:12  steve
+ *  Allow release to handle removal of target net.
+ *
+ * Revision 1.53  2002/08/13 05:35:00  steve
+ *  Do not elide named blocks.
+ *
+ * Revision 1.52  2002/08/12 01:35:00  steve
+ *  conditional ident string using autoconfig.
+ *
+ * Revision 1.51  2002/08/07 00:54:39  steve
+ *  Add force to nets.
+ *
+ * Revision 1.50  2002/08/04 18:28:15  steve
+ *  Do not use hierarchical names of memories to
+ *  generate vvp labels. -tdll target does not
+ *  used hierarchical name string to look up the
+ *  memory objects in the design.
+ *
+ * Revision 1.49  2002/06/16 20:39:12  steve
+ *  Normalize run-time index expressions for bit selects
+ *
+ * Revision 1.48  2002/06/16 19:19:16  steve
+ *  Generate runtime code to normalize indices.
+ *
+ * Revision 1.47  2002/06/05 03:44:25  steve
+ *  Add support for memory words in l-value of
+ *  non-blocking assignments, and remove the special
+ *  NetAssignMem_ and NetAssignMemNB classes.
+ *
+ * Revision 1.46  2002/06/04 05:38:44  steve
+ *  Add support for memory words in l-value of
+ *  blocking assignments, and remove the special
+ *  NetAssignMem class.
+ *
+ * Revision 1.45  2002/05/29 22:05:55  steve
+ *  Offset lvalue index expressions.
+ *
+ * Revision 1.44  2002/05/27 00:08:45  steve
+ *  Support carrying the scope of named begin-end
+ *  blocks down to the code generator, and have
+ *  the vvp code generator use that to support disable.
+ *
+ * Revision 1.43  2002/05/26 01:39:03  steve
+ *  Carry Verilog 2001 attributes with processes,
+ *  all the way through to the ivl_target API.
+ *
+ *  Divide signal reference counts between rval
+ *  and lval references.
+ *
+ * Revision 1.42  2002/04/21 22:31:02  steve
+ *  Redo handling of assignment internal delays.
+ *  Leave it possible for them to be calculated
+ *  at run time.
+ *
  * Revision 1.41  2002/01/19 19:02:08  steve
  *  Pass back target errors processing conditionals.
  *
