@@ -1,7 +1,7 @@
 #ifndef __t_dll_H
 #define __t_dll_H
 /*
- * Copyright (c) 2000 Stephen Williams (steve@icarus.com)
+ * Copyright (c) 2000-2004 Stephen Williams (steve@icarus.com)
  *
  *    This source code is free software; you can redistribute it
  *    and/or modify it in source code form under the terms of the GNU
@@ -18,12 +18,14 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT) && !defined(macintosh)
-#ident "$Id: t-dll.h,v 1.75 2002/01/28 00:52:41 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: t-dll.h,v 1.115 2004/10/04 01:10:56 steve Exp $"
 #endif
 
 # include  "target.h"
 # include  "ivl_target.h"
+# include  "StringHeap.h"
+# include  "netlist.h"
 
 #if defined(__MINGW32__)
 #include <windows.h>
@@ -67,6 +69,7 @@ struct dll_target  : public target_t, public expr_scan_t {
 
       bool bufz(const NetBUFZ*);
       void event(const NetEvent*);
+      void variable(const NetVariable*);
       void logic(const NetLogic*);
       void net_case_cmp(const NetCaseCmp*);
       void udp(const NetUDP*);
@@ -82,6 +85,7 @@ struct dll_target  : public target_t, public expr_scan_t {
       void net_assign(const NetAssign_*);
       bool net_cassign(const NetCAssign*);
       bool net_force(const NetForce*);
+      bool net_function(const NetUserFunc*);
       bool net_const(const NetConst*);
       void net_probe(const NetEvProbe*);
 
@@ -91,7 +95,6 @@ struct dll_target  : public target_t, public expr_scan_t {
       void memory(const NetMemory*);
 
       ivl_dll_t dll_;
-      string dll_path_;
 
       ivl_design_s des_;
 
@@ -102,9 +105,7 @@ struct dll_target  : public target_t, public expr_scan_t {
 	   statements of a thread. */
       struct ivl_statement_s*stmt_cur_;
       void proc_assign(const NetAssign*);
-      void proc_assign_mem(const NetAssignMem*);
       void proc_assign_nb(const NetAssignNB*);
-      void proc_assign_mem_nb(const NetAssignMemNB*net);
       bool proc_block(const NetBlock*);
       void proc_case(const NetCase*);
       bool proc_cassign(const NetCAssign*);
@@ -122,7 +123,7 @@ struct dll_target  : public target_t, public expr_scan_t {
       bool proc_wait(const NetEvWait*);
       void proc_while(const NetWhile*);
 
-      void func_def(const NetScope*);
+      bool func_def(const NetScope*);
       void task_def(const NetScope*);
 
       struct ivl_expr_s*expr_;
@@ -130,6 +131,9 @@ struct dll_target  : public target_t, public expr_scan_t {
       void expr_concat(const NetEConcat*);
       void expr_memory(const NetEMemory*);
       void expr_const(const NetEConst*);
+      void expr_creal(const NetECReal*);
+      void expr_param(const NetEConstParam*);
+      void expr_event(const NetEEvent*);
       void expr_scope(const NetEScope*);
       void expr_select(const NetESelect*);
       void expr_sfunc(const NetESFunc*);
@@ -138,14 +142,32 @@ struct dll_target  : public target_t, public expr_scan_t {
       void expr_ufunc(const NetEUFunc*);
       void expr_unary(const NetEUnary*);
       void expr_signal(const NetESignal*);
+      void expr_variable(const NetEVariable*);
 
       ivl_scope_t lookup_scope_(const NetScope*scope);
-      ivl_memory_t lookup_memory_(const NetMemory*mem);
+
+      ivl_attribute_s* fill_in_attributes(const Attrib*net);
+      void logic_attributes(struct ivl_net_logic_s *obj, const NetNode*net);
 
     private:
+      StringHeap strings_;
+
       static ivl_scope_t find_scope(ivl_design_s &des, const NetScope*cur);
       static ivl_signal_t find_signal(ivl_design_s &des, const NetNet*net);
+      static ivl_memory_t find_memory(ivl_design_s &des, const NetMemory*net);
+      static ivl_variable_t find_variable(ivl_design_s &des, const NetVariable*net);
+
+      static ivl_parameter_t scope_find_param(ivl_scope_t scope,
+					      const char*name);
+
       void add_root(ivl_design_s &des_, const NetScope *s);
+
+      void sub_off_from_expr_(long);
+      void mul_expr_by_const_(long);
+
+      void make_scope_parameters(ivl_scope_t scope, const NetScope*net);
+
+      static ivl_expr_t expr_from_value_(const verinum&that);
 };
 
 /*
@@ -153,9 +175,9 @@ struct dll_target  : public target_t, public expr_scan_t {
  */
 
 struct ivl_event_s {
-      char*name;
+      perm_string name;
       ivl_scope_t scope;
-      unsigned short nany, nneg, npos;
+      unsigned nany, nneg, npos;
       ivl_nexus_t*pins;
 };
 
@@ -167,8 +189,9 @@ struct ivl_event_s {
  */
 struct ivl_expr_s {
       ivl_expr_type_t type_;
+      ivl_variable_type_t value_;
 
-      unsigned width_  :24;
+      unsigned width_;
       unsigned signed_ : 1;
 
       union {
@@ -184,15 +207,20 @@ struct ivl_expr_s {
 	    } bitsel_;
 
 	    struct {
-		  unsigned   rept  :16;
-		  unsigned   parms :16;
+		  unsigned   rept;
+		  unsigned   parms;
 		  ivl_expr_t*parm;
 	    } concat_;
 
 	    struct {
 		  char*bits_;
+		  ivl_parameter_t parameter;
 	    } number_;
 
+	    struct {
+		  ivl_event_t event;
+	    } event_;
+
 	    struct {
 		  ivl_scope_t scope;
 	    } scope_;
@@ -203,13 +231,14 @@ struct ivl_expr_s {
 	    } signal_;
 
 	    struct {
-		  char*name_;
-		  ivl_expr_t    *parm;
-		  unsigned short parms;
+		  const char *name_;
+		  ivl_expr_t *parm;
+		  unsigned   parms;
 	    } sfunc_;
 
 	    struct {
 		  char*value_;
+		  ivl_parameter_t parameter;
 	    } string_;
 
 	    struct {
@@ -225,19 +254,27 @@ struct ivl_expr_s {
 
 	    struct {
 		  ivl_scope_t def;
-		  ivl_expr_t    *parm;
-		  unsigned short parms;
+		  ivl_expr_t  *parm;
+		  unsigned    parms;
 	    } ufunc_;
 
 	    struct {
 		  unsigned long value;
 	    } ulong_;
 
+	    struct {
+		  double value;
+		  ivl_parameter_t parameter;
+	    } real_;
+
 	    struct {
 		  char op_;
 		  ivl_expr_t sub_;
 	    } unary_;
 
+	    struct {
+		  ivl_variable_t var;
+	    } variable_;
       } u_;
 };
 
@@ -251,14 +288,18 @@ struct ivl_expr_s {
 struct ivl_lpm_s {
       ivl_lpm_type_t type;
       ivl_scope_t scope;
-      char* name;
+      perm_string name;
 
       union {
 	    struct ivl_lpm_ff_s {
-		  unsigned short width;
-		  unsigned short swid; // ram only
+		  unsigned width;
+		  unsigned swid; // ram only
 		  ivl_nexus_t clk;
 		  ivl_nexus_t we;
+		  ivl_nexus_t aclr;
+		  ivl_nexus_t aset;
+		  ivl_nexus_t sclr;
+		  ivl_nexus_t sset;
 		  union {
 			ivl_nexus_t*pins;
 			ivl_nexus_t pin;
@@ -272,12 +313,14 @@ struct ivl_lpm_s {
 			ivl_nexus_t pin;
 		  } s;
 		  ivl_memory_t mem; // ram only
+		  ivl_expr_t aset_value;
+		  ivl_expr_t sset_value;
 	    } ff;
 
 	    struct ivl_lpm_mux_s {
-		  unsigned short width;
-		  unsigned short size;
-		  unsigned short swid;
+		  unsigned width;
+		  unsigned size;
+		  unsigned swid;
 		  ivl_nexus_t*d;
 		  union {
 			ivl_nexus_t*pins;
@@ -290,17 +333,26 @@ struct ivl_lpm_s {
 	    } mux;
 
 	    struct ivl_lpm_shift_s {
-		  unsigned short width;
-		  unsigned short select;
+		  unsigned width;
+		  unsigned select;
+		  unsigned signed_flag :1;
 		  ivl_nexus_t*q;
 		  ivl_nexus_t*d;
 		  ivl_nexus_t*s;
 	    } shift;
 
 	    struct ivl_lpm_arith_s {
-		  unsigned short width;
+		  unsigned width;
+		  unsigned signed_flag :1;
 		  ivl_nexus_t*q, *a, *b;
 	    } arith;
+
+	    struct ivl_lpm_ufunc_s {
+		  ivl_scope_t def;
+		  unsigned ports;
+		  unsigned *port_wid;
+		  ivl_nexus_t*pins;
+	    } ufunc;
       } u_;
 };
 
@@ -314,16 +366,19 @@ enum ivl_lval_type_t {
       IVL_LVAL_REG = 0,
       IVL_LVAL_MUX = 1,
       IVL_LVAL_MEM = 2,
+      IVL_LVAL_NET = 3, /* Only force can have NET l-values */
+      IVL_LVAL_VAR = 4
 };
 
 struct ivl_lval_s {
-      unsigned width_  :24;
-      unsigned loff_   :24;
+      unsigned width_;
+      unsigned loff_;
       unsigned type_   : 8;
       ivl_expr_t idx;
       union {
 	    ivl_signal_t sig;
 	    ivl_memory_t mem;
+	    ivl_variable_t var;
       } n;
 };
 
@@ -332,7 +387,7 @@ struct ivl_lval_s {
  * structural context.
  */
 struct ivl_net_const_s {
-      unsigned width_  :24;
+      unsigned width_;
       unsigned signed_ : 1;
 
       union {
@@ -354,15 +409,14 @@ struct ivl_net_logic_s {
       ivl_logic_t type_;
       ivl_udp_t udp;
 
-      char* name_;
+      perm_string name_;
       ivl_scope_t scope_;
 
       unsigned npins_;
       ivl_nexus_t*pins_;
 
-      char**akey_;
-      char**aval_;
-      unsigned nattr_;
+      struct ivl_attribute_s*attr;
+      unsigned nattr;
 
       unsigned delay[3];
 };
@@ -372,12 +426,13 @@ struct ivl_net_logic_s {
  * UDP definition.
  */
 struct ivl_udp_s {
-      char* name;
+      perm_string name;
       unsigned nin;
-      unsigned short sequ;
+      unsigned sequ;
       char init;
       unsigned nrows;
-      char **table; // zero terminated array of pointers
+      typedef const char*ccharp_t;
+      ccharp_t*table; // zero terminated array of pointers
 };
 
 /*
@@ -393,7 +448,7 @@ struct ivl_udp_s {
  * input to the device, then the drives are both HiZ.
  */
 struct ivl_nexus_ptr_s {
-      unsigned pin_  :24;
+      unsigned pin_;
       unsigned type_ : 8;
       unsigned drive0 : 3;
       unsigned drive1 : 3;
@@ -412,24 +467,32 @@ struct ivl_nexus_ptr_s {
 struct ivl_nexus_s {
       unsigned nptr_;
       struct ivl_nexus_ptr_s*ptrs_;
-      char*name_;
+      const char*name_;
       void*private_data;
 };
 
 
-/* 
+/*
  * Memory.
  */
 struct ivl_memory_s {
-      char*name_;
+      perm_string basename_;
       ivl_scope_t scope_;
-      unsigned width_  :24;
+      unsigned width_;
       unsigned signed_ : 1;
       unsigned size_;
       int root_;
 };
 
-
+/*
+ * This is the implementation of a parameter. Each scope has a list of
+ * these.
+ */
+struct ivl_parameter_s {
+      perm_string basename;
+      ivl_scope_t scope;
+      ivl_expr_t  value;
+};
 /*
  * All we know about a process it its type (initial or always) and the
  * single statement that is it. A process also has a scope, although
@@ -440,6 +503,9 @@ struct ivl_process_s {
       ivl_scope_t scope_;
       ivl_statement_t stmt_;
 
+      struct ivl_attribute_s*attr;
+      unsigned nattr;
+
       ivl_process_t next_;
 };
 
@@ -452,8 +518,8 @@ struct ivl_process_s {
 struct ivl_scope_s {
       ivl_scope_t child_, sibling_, parent;
 
-      char* name_;
-      const char* tname_;
+      perm_string name_;
+      perm_string tname_;
       ivl_scope_type_t type_;
 
       unsigned nsigs_;
@@ -471,11 +537,22 @@ struct ivl_scope_s {
       unsigned nmem_;
       ivl_memory_t* mem_;
 
+      unsigned nvar_;
+      ivl_variable_t* var_;
+
+      unsigned nparam_;
+      ivl_parameter_t param_;
+
 	/* Scopes that are tasks/functions have a definition. */
       ivl_statement_t def;
 
       unsigned ports;
       ivl_signal_t*port;
+
+      signed int time_units :8;
+
+      struct ivl_attribute_s*attr;
+      unsigned nattr;
 };
 
 /*
@@ -488,10 +565,17 @@ struct ivl_signal_s {
       ivl_signal_type_t type_;
       ivl_signal_port_t port_;
 
-      unsigned width_  :24;
+      unsigned width_;
       unsigned signed_ : 1;
+      unsigned isint_  : 1;
+      unsigned local_  : 1;
 
-      char*name_;
+	/* These encode the run-time index for the least significant
+	   bit, and the distance to the second bit. */
+      signed lsb_index;
+      signed lsb_dist;
+
+      perm_string name_;
       ivl_scope_t scope_;
 
       union {
@@ -499,9 +583,8 @@ struct ivl_signal_s {
 	    ivl_nexus_t*pins_;
       } n;
 
-      char**akey_;
-      char**aval_;
-      unsigned nattr_;
+      struct ivl_attribute_s*attr;
+      unsigned nattr;
 };
 
 /*
@@ -522,6 +605,7 @@ struct ivl_statement_s {
 	    struct { /* IVL_ST_BLOCK, IVL_ST_FORK */
 		  struct ivl_statement_s*stmt_;
 		  unsigned nstmt_;
+		  ivl_scope_t scope;
 	    } block_;
 
 	    struct { /* IVL_ST_CASE, IVL_ST_CASEX, IVL_ST_CASEZ */
@@ -564,21 +648,21 @@ struct ivl_statement_s {
 	    } forever_;
 
 	    struct { /* IVL_ST_STASK */
-		  char* name_;
-		  unsigned nparm_;
+		  const char*name_;
+		  unsigned   nparm_;
 		  ivl_expr_t*parms_;
 	    } stask_;
 
-	    struct { /* IVL_ST_TRIGGER */
-		  ivl_event_t event_;
-	    } trig_;
-
 	    struct { /* IVL_ST_UTASK */
 		  ivl_scope_t def;
 	    } utask_;
 
-	    struct { /* IVL_ST_WAIT */
-		  ivl_event_t event_;
+	    struct { /* IVL_ST_TRIGGER IVL_ST_WAIT */
+		  unsigned nevent;
+		  union {
+			ivl_event_t event;
+			ivl_event_t*events;
+		  };
 		  ivl_statement_t stmt_;
 	    } wait_;
 
@@ -589,113 +673,72 @@ struct ivl_statement_s {
       } u_;
 };
 
+/*
+ * This holds the details about a variable object.
+ */
+struct ivl_variable_s {
+      ivl_variable_type_t type;
+      perm_string name;
+      ivl_scope_t scope;
+};
+
 /*
  * $Log: t-dll.h,v $
- * Revision 1.75  2002/01/28 00:52:41  steve
- *  Add support for bit select of parameters.
- *  This leads to a NetESelect node and the
- *  vvp code generator to support that.
+ * Revision 1.115  2004/10/04 01:10:56  steve
+ *  Clean up spurious trailing white space.
  *
- * Revision 1.74  2002/01/19 19:02:08  steve
- *  Pass back target errors processing conditionals.
+ * Revision 1.114  2004/06/30 02:16:27  steve
+ *  Implement signed divide and signed right shift in nets.
  *
- * Revision 1.73  2002/01/03 04:19:01  steve
- *  Add structural modulus support down to vvp.
+ * Revision 1.113  2004/05/31 23:34:39  steve
+ *  Rewire/generalize parsing an elaboration of
+ *  function return values to allow for better
+ *  speed and more type support.
  *
- * Revision 1.72  2001/12/06 03:11:01  steve
- *  Add ivl_logic_delay function to ivl_target.
+ * Revision 1.112  2004/02/20 06:22:58  steve
+ *  parameter keys are per_strings.
  *
- * Revision 1.71  2001/11/14 03:28:49  steve
- *  DLL target support for force and release.
+ * Revision 1.111  2004/02/19 07:06:57  steve
+ *  LPM, logic and Variables have perm_string names.
  *
- * Revision 1.70  2001/11/04 05:03:21  steve
- *  MacOSX 10.1 updates.
+ * Revision 1.110  2004/02/19 06:57:11  steve
+ *  Memory and Event names use perm_string.
  *
- * Revision 1.69  2001/11/01 04:25:31  steve
- *  ivl_target support for cassign.
+ * Revision 1.109  2004/02/18 17:11:58  steve
+ *  Use perm_strings for named langiage items.
  *
- * Revision 1.68  2001/10/31 05:24:52  steve
- *  ivl_target support for assign/deassign.
+ * Revision 1.108  2003/12/03 02:46:24  steve
+ *  Add support for wait on list of named events.
  *
- * Revision 1.67  2001/10/30 02:52:07  steve
- *  Stubs for assign/deassign for t-dll.
+ * Revision 1.107  2003/11/10 20:59:04  steve
+ *  Design::get_flag returns const char* instead of string.
  *
- * Revision 1.66  2001/10/19 21:53:24  steve
- *  Support multiple root modules (Philip Blundell)
+ * Revision 1.106  2003/09/03 23:33:29  steve
+ *  Pass FF synchronous set values to code generator.
  *
- * Revision 1.65  2001/10/16 02:19:27  steve
- *  Support IVL_LPM_DIVIDE for structural divide.
+ * Revision 1.105  2003/08/15 02:23:53  steve
+ *  Add synthesis support for synchronous reset.
  *
- * Revision 1.64  2001/09/16 22:19:42  steve
- *  Support attributes to logic gates.
+ * Revision 1.104  2003/06/23 01:25:44  steve
+ *  Module attributes make it al the way to ivl_target.
  *
- * Revision 1.63  2001/09/01 01:57:31  steve
- *  Make constants available through the design root
+ * Revision 1.103  2003/06/17 21:28:59  steve
+ *  Remove short int restrictions from vvp opcodes. (part 2)
  *
- * Revision 1.62  2001/08/31 22:58:40  steve
- *  Support DFF CE inputs.
+ * Revision 1.102  2003/04/22 04:48:30  steve
+ *  Support event names as expressions elements.
  *
- * Revision 1.61  2001/08/28 04:07:41  steve
- *  Add some ivl_target convenience functions.
+ * Revision 1.101  2003/04/11 05:18:08  steve
+ *  Handle signed magnitude compare all the
+ *  way through to the vvp code generator.
  *
- * Revision 1.60  2001/08/25 23:50:03  steve
- *  Change the NetAssign_ class to refer to the signal
- *  instead of link into the netlist. This is faster
- *  and uses less space. Make the NetAssignNB carry
- *  the delays instead of the NetAssign_ lval objects.
+ * Revision 1.100  2003/03/10 23:40:54  steve
+ *  Keep parameter constants for the ivl_target API.
  *
- *  Change the vvp code generator to support multiple
- *  l-values, i.e. concatenations of part selects.
- *
- * Revision 1.59  2001/08/10 00:40:45  steve
- *  tgt-vvp generates code that skips nets as inputs.
- *
- * Revision 1.58  2001/07/27 04:51:44  steve
- *  Handle part select expressions as variants of
- *  NetESignal/IVL_EX_SIGNAL objects, instead of
- *  creating new and useless temporary signals.
- *
- * Revision 1.57  2001/07/27 02:41:56  steve
- *  Fix binding of dangling function ports. do not elide them.
- *
- * Revision 1.56  2001/07/22 00:17:50  steve
- *  Support the NetESubSignal expressions in vvp.tgt.
- *
- * Revision 1.55  2001/07/19 04:55:06  steve
- *  Support calculated delays in vvp.tgt.
- *
- * Revision 1.54  2001/07/07 20:20:10  steve
- *  Pass parameters to system functions.
- *
- * Revision 1.53  2001/07/04 22:59:25  steve
- *  handle left shifter in dll output.
- *
- * Revision 1.52  2001/06/30 23:03:16  steve
- *  support fast programming by only writing the bits
- *  that are listed in the input file.
- *
- * Revision 1.51  2001/06/30 21:07:26  steve
- *  Support non-const right shift (unsigned).
- *
- * Revision 1.50  2001/06/19 03:01:10  steve
- *  Add structural EEQ gates (Stephan Boettcher)
- *
- * Revision 1.49  2001/06/16 23:45:05  steve
- *  Add support for structural multiply in t-dll.
- *  Add code generators and vvp support for both
- *  structural and behavioral multiply.
- *
- * Revision 1.48  2001/06/16 02:41:42  steve
- *  Generate code to support memory access in continuous
- *  assignment statements. (Stephan Boettcher)
- *
- * Revision 1.47  2001/06/15 04:14:19  steve
- *  Generate vvp code for GT and GE comparisons.
- *
- * Revision 1.46  2001/06/07 02:12:43  steve
- *  Support structural addition.
- *
- * Revision 1.45  2001/05/20 15:09:39  steve
- *  Mingw32 support (Venkat Iyer)
+ * Revision 1.99  2003/03/01 06:25:30  steve
+ *  Add the lex_strings string handler, and put
+ *  scope names and system task/function names
+ *  into this table. Also, permallocate event
+ *  names from the beginning.
  */
 #endif
