@@ -16,8 +16,8 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT) && !defined(macintosh)
-#ident "$Id: t-xnf.cc,v 1.43 2001/07/25 03:10:50 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: t-xnf.cc,v 1.52 2004/02/20 18:53:36 steve Exp $"
 #endif
 
 # include "config.h"
@@ -29,7 +29,7 @@
  * use by Xilinx tools, and other tools that accepts Xilinx designs.
  *
  * The code generator automatically detects ports to top level modules
- * and generates SIG records that make the XNF useable as a schematic.
+ * and generates SIG records that make the XNF usable as a schematic.
  *
  * FLAGS
  * The XNF backend uses the following flags from the command line to
@@ -61,7 +61,7 @@
  *      Specify the LCA library part type for the gate. The lname
  *      is the name of the symbol to use (i.e. DFF) and the comma
  *      separated list is the names of the pins, in the order they
- *      appear in the verilog source. If the name is prefixed with a
+ *      appear in the Verilog source. If the name is prefixed with a
  *      tilde (~) then the pin is inverted, and the proper "INV" token
  *      will be added to the PIN record.
  *
@@ -77,7 +77,7 @@
 # include  "netlist.h"
 # include  "target.h"
 # include  <fstream>
-# include  <strstream>
+# include  <sstream>
 
 verinum::V link_get_ival(const Link&lnk)
 {
@@ -87,7 +87,7 @@ verinum::V link_get_ival(const Link&lnk)
 	    if (cur == &lnk)
 		  continue;
 
-	    if (const NetNet*sig = dynamic_cast<const NetNet*>(cur->get_obj()))
+	    if (dynamic_cast<const NetNet*>(cur->get_obj()))
 		  return cur->nexus()->get_init();
 
       }
@@ -119,6 +119,7 @@ class target_xnf  : public target_t {
 
     private:
       static string mangle(const string&);
+      static string mangle(perm_string);
       static string choose_sig_name(const Link*lnk);
       static void draw_pin(ostream&os, const string&name,
 			   const Link&lnk);
@@ -153,6 +154,11 @@ string target_xnf::mangle(const string&name)
       return result;
 }
 
+string target_xnf::mangle(perm_string name)
+{
+      return mangle(string(name));
+}
+
 /*
  * This method takes a signal and pin number as a nexus. Scan the
  * nexus to decide which name to use if there are lots of attached
@@ -160,7 +166,7 @@ string target_xnf::mangle(const string&name)
  */
 string target_xnf::choose_sig_name(const Link*lnk)
 {
-      return mangle( lnk->nexus()->name() );
+      return mangle( string(lnk->nexus()->name()) );
 }
 
 void target_xnf::draw_pin(ostream&os, const string&name,
@@ -173,7 +179,7 @@ void target_xnf::draw_pin(ostream&os, const string&name,
 	    use_name = use_name.substr(1);
       }
 
-      char type;
+      char type=0;
       switch (lnk.get_dir()) {
 	  case Link::INPUT:
 	  case Link::PASSIVE:
@@ -183,6 +189,7 @@ void target_xnf::draw_pin(ostream&os, const string&name,
 	    type = 'O';
 	    break;
       }
+      assert(type);
 
       os << "    PIN, " << use_name << ", " << type << ", " <<
 	    choose_sig_name(&lnk);
@@ -224,17 +231,17 @@ void target_xnf::draw_sym_with_lcaname(ostream&os, string lca,
 
 bool target_xnf::start_design(const Design*des)
 {
-      out_.open(des->get_flag("-o").c_str(), ios::out | ios::trunc);
+      out_.open(des->get_flag("-o"), ios::out | ios::trunc);
 
       string ncfpath = des->get_flag("ncf");
       if (ncfpath != "")
 	    ncf_.open(ncfpath.c_str());
 
       out_ << "LCANET,6" << endl;
-      out_ << "PROG,verilog,$Name: v0_6 $,\"Icarus Verilog\"" << endl;
-      ncf_ << "# Generated by Icarus Verilog $Name: v0_6 $" << endl;
+      out_ << "PROG,verilog,$Name: v0_8 $,\"Icarus Verilog\"" << endl;
+      ncf_ << "# Generated by Icarus Verilog $Name: v0_8 $" << endl;
 
-      if (des->get_flag("part") != "") {
+      if (des->get_flag("part") != 0) {
 	    out_ << "PART," << des->get_flag("part") << endl;
 	    ncf_ << "CONFIG PART=" << des->get_flag("part") << ";" << endl;
       }
@@ -340,7 +347,7 @@ void target_xnf::signal(const NetNet*net)
 	/* Now look to see if a PAD attribute is attached, and if so
 	   write out PAD information to the XNF and the ncf files. */
 
-      string pad = net->attribute("PAD");
+      string pad = net->attribute(perm_string::literal("PAD")).as_string();
       if (pad == "")
 	    return;
 
@@ -424,6 +431,9 @@ void target_xnf::draw_carry(ostream &os, const NetAddSub*gate, unsigned idx,
 	  case LOWER_W_CO:
 	    draw_pin(os, "COUT0", gate->pin_Cout());
 	    break;
+
+	  default:
+	    assert(0);
       }
 
       // Connect the Cout, this will connect to the next Cin
@@ -672,19 +682,19 @@ void target_xnf::lpm_compare_le_(ostream&os, const NetCompare*dev)
 
 void target_xnf::lpm_ff(const NetFF*net)
 {
-      string type = net->attribute("LPM_FFType");
+      string type = net->attribute(perm_string::literal("LPM_FFType")).as_string();
       if (type == "") type = "DFF";
 
 	// XXXX For now, only support DFF
       assert(type == "DFF");
 
-      string lcaname = net->attribute("XNF-LCA");
+      string lcaname = net->attribute(perm_string::literal("XNF-LCA")).as_string();
       if (lcaname != "") {
 	    draw_sym_with_lcaname(out_, lcaname, net);
 	    return;
       }
 
-      assert(net->attribute("XNF-LCA") == "");
+      assert(net->attribute(perm_string::literal("XNF-LCA")) == verinum(""));
 
 	/* Create a DFF object for each bit of width. The symbol name
 	   has the index number appended so that read XNF may be able
@@ -705,13 +715,16 @@ void target_xnf::lpm_ff(const NetFF*net)
 		case verinum::V1:
 		  out_ << "INIT=S, ";
 		  break;
+
+		default:
+		  break;
 	    }
 
 	    out_ << "LIBVER=2.0.0" << endl;
 	    draw_pin(out_, "Q", net->pin_Q(idx));
 	    draw_pin(out_, "D", net->pin_Data(idx));
 
-	    if (net->attribute("Clock:LPM_Polarity") == "INVERT")
+	    if (net->attribute(perm_string::literal("Clock:LPM_Polarity")) == verinum("INVERT"))
 		  draw_pin(out_, "~C", net->pin_Clock());
 	    else
 		  draw_pin(out_, "C", net->pin_Clock());
@@ -754,8 +767,6 @@ void target_xnf::lpm_ram_dq(const NetRamDq*ram)
 {
       assert(ram->count_partners() == 1);
 
-      const NetMemory*mem = ram->mem();
-
       for (unsigned idx = 0 ;  idx < ram->width() ;  idx += 1) {
 	    out_ << "SYM, " << mangle(ram->name())
 	       << "<" << idx << ">, RAMS" << endl;
@@ -765,8 +776,8 @@ void target_xnf::lpm_ram_dq(const NetRamDq*ram)
 	    draw_pin(out_, "WE", ram->pin_WE());
 	    draw_pin(out_, "WCLK", ram->pin_InClock());
 	    for (unsigned adr = 0 ;  adr < ram->awidth() ;  adr += 1) {
-		  strstream tmp;
-		  tmp << "A" << adr << ends;
+		  ostringstream tmp;
+		  tmp << "A" << adr;
 		  draw_pin(out_, tmp.str(), ram->pin_Address(adr));
 	    }
 
@@ -811,7 +822,7 @@ void target_xnf::logic(const NetLogic*net)
 {
 	// The XNF-LCA attribute overrides anything I might guess
 	// about this object.
-      string lca = net->attribute("XNF-LCA");
+      string lca = net->attribute(perm_string::literal("XNF-LCA")).as_string();
       if (lca != "") {
 	    draw_sym_with_lcaname(out_, lca, net);
 	    return;
@@ -909,7 +920,7 @@ bool target_xnf::bufz(const NetBUFZ*net)
 
 void target_xnf::udp(const NetUDP*net)
 {
-      string lca = net->attribute("XNF-LCA");
+      string lca = net->attribute(perm_string::literal("XNF-LCA")).as_string();
 
 	// I only know how to draw a UDP if it has the XNF-LCA
 	// attribute attached to it.
@@ -927,6 +938,38 @@ extern const struct target tgt_xnf = { "xnf", &target_xnf_obj };
 
 /*
  * $Log: t-xnf.cc,v $
+ * Revision 1.52  2004/02/20 18:53:36  steve
+ *  Addtrbute keys are perm_strings.
+ *
+ * Revision 1.51  2004/02/18 17:11:58  steve
+ *  Use perm_strings for named langiage items.
+ *
+ * Revision 1.50  2003/11/10 20:59:04  steve
+ *  Design::get_flag returns const char* instead of string.
+ *
+ * Revision 1.49  2003/07/05 20:42:08  steve
+ *  Fix some enumeration warnings.
+ *
+ * Revision 1.48  2003/06/24 01:38:03  steve
+ *  Various warnings fixed.
+ *
+ * Revision 1.47  2003/01/30 16:23:08  steve
+ *  Spelling fixes.
+ *
+ * Revision 1.46  2003/01/14 21:16:18  steve
+ *  Move strstream to ostringstream for compatibility.
+ *
+ * Revision 1.45  2002/08/12 01:35:01  steve
+ *  conditional ident string using autoconfig.
+ *
+ * Revision 1.44  2002/05/23 03:08:52  steve
+ *  Add language support for Verilog-2001 attribute
+ *  syntax. Hook this support into existing $attribute
+ *  handling, and add number and void value types.
+ *
+ *  Add to the ivl_target API new functions for access
+ *  of complex attributes attached to gates.
+ *
  * Revision 1.43  2001/07/25 03:10:50  steve
  *  Create a config.h.in file to hold all the config
  *  junk, and support gcc 3.0. (Stephan Boettcher)
