@@ -16,8 +16,8 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT) && !defined(macintosh)
-#ident "$Id: verinum.cc,v 1.31 2002/02/01 05:09:14 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: verinum.cc,v 1.43 2004/05/18 18:43:15 steve Exp $"
 #endif
 
 # include "config.h"
@@ -45,6 +45,22 @@ verinum::verinum(const string&str)
 : has_len_(true), has_sign_(false), string_flag_(true)
 {
       nbits_ = str.length() * 8;
+
+	// Special case: The string "" is 8 bits of 0.
+      if (nbits_ == 0) {
+	    nbits_ = 8;
+	    bits_ = new V [nbits_];
+	    bits_[0] = V0;
+	    bits_[1] = V0;
+	    bits_[2] = V0;
+	    bits_[3] = V0;
+	    bits_[4] = V0;
+	    bits_[5] = V0;
+	    bits_[6] = V0;
+	    bits_[7] = V0;
+	    return;
+      }
+
       bits_ = new V [nbits_];
 
       unsigned idx, cp;
@@ -95,14 +111,27 @@ verinum::verinum(const verinum&that)
 
 verinum::verinum(const verinum&that, unsigned nbits)
 {
-      assert(nbits <= that.nbits_);
       string_flag_ = false;
       nbits_ = nbits;
       bits_ = new V[nbits_];
       has_len_ = true;
-      has_sign_ = false;
-      for (unsigned idx = 0 ;  idx < nbits_ ;  idx += 1)
+      has_sign_ = that.has_sign_;
+
+      unsigned copy = nbits;
+      if (copy > that.nbits_)
+	    copy = that.nbits_;
+      for (unsigned idx = 0 ;  idx < copy ;  idx += 1)
 	    bits_[idx] = that.bits_[idx];
+
+      if (copy < nbits_) {
+	    if (has_sign_) {
+		  for (unsigned idx = copy ;  idx < nbits_ ;  idx += 1)
+			bits_[idx] = bits_[idx-1];
+	    } else {
+		  for (unsigned idx = copy ;  idx < nbits_ ;  idx += 1)
+			bits_[idx] = verinum::V0;
+	    }
+      }
 }
 
 verinum::verinum(long that)
@@ -188,7 +217,7 @@ unsigned long verinum::as_ulong() const
  *
  * This function presumes that the native format is 2s compliment
  * (pretty safe these days) and masks/sets bits accordingly. If the
- * value is too large for the lative form, it truncates the high bits.
+ * value is too large for the native form, it truncates the high bits.
  */
 signed long verinum::as_long() const
 {
@@ -207,10 +236,13 @@ signed long verinum::as_long() const
 
 	    val = -1;
 	    signed long mask = ~1L;
-	    for (unsigned idx = 0 ;  idx < top ;  idx += 1, mask <<= 1)
+	    for (unsigned idx = 0 ;  idx < top ;  idx += 1) {
 		  if (bits_[idx] == V0)
 			val &= mask;
 
+		  mask = (mask << 1) | 1L;
+	    }
+
       } else {
 	    unsigned top = nbits_;
 	    if (top > (8 * sizeof(long) - 1))
@@ -233,7 +265,7 @@ string verinum::as_string() const
 
       char*tmp = new char[nbits_/8+1];
       char*cp = tmp;
-      for (unsigned idx = nbits_ ;  idx > 0 ;  idx -= 8, cp += 1) {
+      for (unsigned idx = nbits_ ;  idx > 0 ;  idx -= 8) {
 	    V*bp = bits_+idx;
 	    *cp = 0;
 	    if (*(--bp) == V1) *cp |= 0x80;
@@ -244,6 +276,10 @@ string verinum::as_string() const
 	    if (*(--bp) == V1) *cp |= 0x04;
 	    if (*(--bp) == V1) *cp |= 0x02;
 	    if (*(--bp) == V1) *cp |= 0x01;
+	    if (*cp != 0) {
+		  cp += 1;
+		  *cp = 0;
+	    }
       }
 
       tmp[nbits_/8] = 0;
@@ -273,6 +309,14 @@ bool verinum::is_defined() const
       return true;
 }
 
+bool verinum::is_zero() const
+{
+      for (unsigned idx = 0 ;  idx < nbits_ ;  idx += 1)
+	    if (bits_[idx] != V0) return false;
+
+      return true;
+}
+
 /*
  * This function returns a version of the verinum that has only as
  * many bits as are needed to accurately represent the value. It takes
@@ -281,7 +325,7 @@ bool verinum::is_defined() const
  * If the input value has a definite length, then that value is just
  * returned as is.
  */
-static verinum trim_vnum(const verinum&that)
+verinum trim_vnum(const verinum&that)
 {
       unsigned tlen;
 
@@ -314,7 +358,7 @@ static verinum trim_vnum(const verinum&that)
 	    tlen = 1;
 	    for (unsigned idx = tlen ;  idx < that.len() ;  idx += 1)
 		  if (that.get(idx) != verinum::V0)
-			tlen = idx;
+			tlen = idx+1;
 
       }
 
@@ -351,6 +395,11 @@ ostream& operator<< (ostream&o, verinum::V v)
  */
 ostream& operator<< (ostream&o, const verinum&v)
 {
+      if (v.is_string()) {
+	    o << "\"" << v.as_string() << "\"";
+	    return o;
+      }
+
 	/* If the verinum number has a fixed length, dump all the bits
 	   literally. This is how we express the fixed length in the
 	   output. */
@@ -376,7 +425,7 @@ ostream& operator<< (ostream&o, const verinum&v)
 	   out as a decimal number. */
       if (v.is_defined()) {
 	    if (v.has_sign())
-		  o << "'sd" << v.as_ulong();
+		  o << "'sd" << v.as_long();
 	    else
 		  o << "'d" << v.as_ulong();
 	    return o;
@@ -434,12 +483,16 @@ verinum::V operator <= (const verinum&left, const verinum&right)
 	    if (right[idx-1] != verinum::V0) return verinum::V1;
       }
 
+      idx = right.len();
+      if (left.len() < idx) idx = left.len();
+
       while (idx > 0) {
 	    if (left[idx-1] == verinum::Vx) return verinum::Vx;
 	    if (left[idx-1] == verinum::Vz) return verinum::Vx;
 	    if (right[idx-1] == verinum::Vx) return verinum::Vx;
 	    if (right[idx-1] == verinum::Vz) return verinum::Vx;
 	    if (left[idx-1] > right[idx-1]) return verinum::V0;
+	    if (left[idx-1] < right[idx-1]) return verinum::V1;
 	    idx -= 1;
       }
 
@@ -538,8 +591,10 @@ verinum v_not(const verinum&left)
 
 /*
  * Addition works a bit at a time, from the least significant up to
- * the most significant. The result is signed only if either of the
- * operands is signed.
+ * the most significant. The result is signed only if both of the
+ * operands are signed. The result is also expanded as needed to
+ * prevent overflow. It is up to the caller to shrink the result back
+ * down if that is the desire.
  */
 verinum operator + (const verinum&left, const verinum&right)
 {
@@ -549,21 +604,38 @@ verinum operator + (const verinum&left, const verinum&right)
       unsigned max = left.len();
       if (right.len() > max) max = right.len();
 
-      verinum val (verinum::V0, max);
-      val.has_sign( left.has_sign() || right.has_sign() );
+      bool signed_flag = left.has_sign() && right.has_sign();
+      verinum::V*val_bits = new verinum::V[max+1];
 
       verinum::V carry = verinum::V0;
       for (unsigned idx = 0 ;  idx < min ;  idx += 1)
-	    val.set(idx, add_with_carry(left[idx], right[idx], carry));
+	    val_bits[idx] = add_with_carry(left[idx], right[idx], carry);
+
+      verinum::V rpad = signed_flag? right[right.len()-1] : verinum::V0;
+      verinum::V lpad = signed_flag? left[left.len()-1] : verinum::V0;
 
       if (left.len() > right.len()) {
-	    for (unsigned idx = min ;  idx < max ;  idx += 1)
-		  val.set(idx,add_with_carry(left[idx], verinum::V0, carry));
+
+	    for (unsigned idx = min ;  idx < left.len() ;  idx += 1)
+		  val_bits[idx] = add_with_carry(left[idx], rpad, carry);
+
       } else {
-	    for (unsigned idx = min ;  idx < max ;  idx += 1)
-		  val.set(idx, add_with_carry(verinum::V0, right[idx], carry));
+
+	    for (unsigned idx = min ;  idx < right.len() ;  idx += 1)
+		  val_bits[idx] = add_with_carry(lpad, right[idx], carry);
       }
 
+      if (signed_flag) {
+	    val_bits[max] = add_with_carry(lpad, rpad, carry);
+	    if (val_bits[max] != val_bits[max-1])
+		  max += 1;
+      }
+
+      verinum val (val_bits, max, false);
+      val.has_sign(signed_flag);
+
+      delete[]val_bits;
+
       return val;
 }
 
@@ -604,7 +676,7 @@ verinum operator - (const verinum&left, const verinum&r)
  * result. The resulting number is as large as the sum of the sizes of
  * the operand.
  *
- * The algorithm used is sucessive shift and add operations,
+ * The algorithm used is successive shift and add operations,
  * implemented as the nested loops.
  *
  * If either value is not completely defined, then the result is not
@@ -642,6 +714,68 @@ verinum operator * (const verinum&left, const verinum&right)
       return trim_vnum(result);
 }
 
+verinum operator << (const verinum&that, unsigned shift)
+{
+      verinum result(verinum::V0, that.len() + shift, that.has_len());
+
+      for (unsigned idx = 0 ;  idx < that.len() ;  idx += 1)
+	    result.set(idx+shift, that.get(idx));
+
+      return result;
+}
+
+verinum operator >> (const verinum&that, unsigned shift)
+{
+      if (shift >= that.len()) {
+	    if (that.has_sign()) {
+		  verinum result (that.get(that.len()-1), 1);
+		  result.has_sign(true);
+		  return result;
+	    } else {
+		  verinum result(verinum::V0, 1);
+		  return result;
+	    }
+      }
+
+      verinum result(that.has_sign()? that.get(that.len()-1) : verinum::V0,
+		     that.len() - shift, that.has_len());
+
+      for (unsigned idx = shift ;  idx < that.len() ;  idx += 1)
+	    result.set(idx-shift, that.get(idx));
+
+      return result;
+}
+
+static verinum unsigned_divide(verinum num, verinum den)
+{
+      unsigned nwid = num.len();
+      while (nwid > 0 && (num.get(nwid-1) == verinum::V0))
+	    nwid -= 1;
+
+      unsigned dwid = den.len();
+      while (dwid > 0 && (den.get(dwid-1) == verinum::V0))
+	    dwid -= 1;
+
+      if (dwid > nwid)
+	    return verinum(verinum::V0, 1);
+
+      den = den << (nwid-dwid);
+
+      unsigned idx = nwid - dwid + 1;
+      verinum result (verinum::V0, idx);
+      while (idx > 0) {
+	    if (den <= num) {
+		  verinum dif = num - den;
+		  num = dif;
+		  result.set(idx-1, verinum::V1);
+	    }
+	    den = den >> 1;
+	    idx -= 1;
+      }
+
+      return result;
+}
+
 /*
  * This operator divides the left number by the right number. If
  * either value is signed, the result is signed. If both values have a
@@ -664,7 +798,7 @@ verinum operator / (const verinum&left, const verinum&right)
 
 	/* If the right expression is a zero value, then the result is
 	   filled with 'bx bits. */
-      if (right.as_ulong() == 0) {
+      if (right.is_zero()) {
 	    verinum result (verinum::Vx, use_len, has_len_flag);
 	    result.has_sign(left.has_sign() || right.has_sign());
 	    return result;
@@ -691,16 +825,19 @@ verinum operator / (const verinum&left, const verinum&right)
 
       } else {
 
-	      /* XXXX FIXME XXXX Use native unsigned division to do
-		 the work. This does not work if the result is too
-		 large for the native integer. */
-	    assert(use_len <= 8*sizeof(unsigned long));
-	    unsigned long l = left.as_ulong();
-	    unsigned long r = right.as_ulong();
-	    unsigned long v = l / r;
-	    for (unsigned idx = 0 ;  idx < use_len ;  idx += 1) {
-		  result.set(idx,  (v & 1)? verinum::V1 : verinum::V0);
-		  v >>= 1;
+	    if (use_len <= 8 * sizeof(unsigned long)) {
+		    /* Use native unsigned division to do the work. */
+
+		  unsigned long l = left.as_ulong();
+		  unsigned long r = right.as_ulong();
+		  unsigned long v = l / r;
+		  for (unsigned idx = 0 ;  idx < use_len ;  idx += 1) {
+			result.set(idx,  (v & 1)? verinum::V1 : verinum::V0);
+			v >>= 1;
+		  }
+
+	    } else {
+		  result = unsigned_divide(left, right);
 	    }
       }
 
@@ -791,8 +928,57 @@ verinum::V operator & (verinum::V l, verinum::V r)
       return verinum::V1;
 }
 
+verinum::V operator ^ (verinum::V l, verinum::V r)
+{
+      if (l == verinum::V0)
+	    return r;
+      if (r == verinum::V0)
+	    return l;
+      if ((l == verinum::V1) && (r == verinum::V1))
+	    return verinum::V0;
+
+      return verinum::Vx;
+}
+
 /*
  * $Log: verinum.cc,v $
+ * Revision 1.43  2004/05/18 18:43:15  steve
+ *  Handle null string as a single nul character.
+ *
+ * Revision 1.42  2004/02/17 06:52:55  steve
+ *  Support unsigned divide of huge numbers.
+ *
+ * Revision 1.41  2003/10/26 04:54:56  steve
+ *  Support constant evaluation of binary ^ operator.
+ *
+ * Revision 1.40  2003/05/25 03:01:19  steve
+ *  Get length of trimed unsigned value right.
+ *
+ * Revision 1.39  2003/04/14 03:40:21  steve
+ *  Make some effort to preserve bits while
+ *  operating on constant values.
+ *
+ * Revision 1.38  2003/04/03 04:30:00  steve
+ *  Prevent overrun comparing verinums to zero.
+ *
+ * Revision 1.37  2003/02/02 00:43:16  steve
+ *  Fix conversion of signed numbes to long
+ *
+ * Revision 1.36  2003/01/30 16:23:08  steve
+ *  Spelling fixes.
+ *
+ * Revision 1.35  2002/08/19 02:39:17  steve
+ *  Support parameters with defined ranges.
+ *
+ * Revision 1.34  2002/08/12 01:35:01  steve
+ *  conditional ident string using autoconfig.
+ *
+ * Revision 1.33  2002/04/27 23:26:24  steve
+ *  Trim leading nulls from string forms.
+ *
+ * Revision 1.32  2002/04/27 04:48:43  steve
+ *  Display string verinums as strings.
+ *
  * Revision 1.31  2002/02/01 05:09:14  steve
  *  Propagate sign in unary minus.
  *
@@ -818,47 +1004,5 @@ verinum::V operator & (verinum::V l, verinum::V r)
  *
  * Revision 1.24  2001/02/07 21:47:13  steve
  *  Fix expression widths for rvalues and parameters (PR#131,132)
- *
- * Revision 1.23  2001/02/07 02:46:31  steve
- *  Support constant evaluation of / and % (PR#124)
- *
- * Revision 1.22  2001/01/02 03:23:40  steve
- *  Evaluate constant &, | and unary ~.
- *
- * Revision 1.21  2000/12/10 22:01:36  steve
- *  Support decimal constants in behavioral delays.
- *
- * Revision 1.20  2000/09/28 03:55:55  steve
- *  handel, by truncation, verinums that are to long for long integers.
- *
- * Revision 1.19  2000/09/27 18:28:37  steve
- *  multiply in parameter expressions.
- *
- * Revision 1.18  2000/09/07 22:37:10  steve
- *  The + operator now preserves signedness.
- *
- * Revision 1.17  2000/06/12 03:56:51  steve
- *  Fix subract of short value form long one.
- *
- * Revision 1.16  2000/02/23 04:43:43  steve
- *  Some compilers do not accept the not symbol.
- *
- * Revision 1.15  2000/02/23 02:56:56  steve
- *  Macintosh compilers do not support ident.
- *
- * Revision 1.14  2000/01/07 03:45:49  steve
- *  Initial support for signed constants.
- *
- * Revision 1.13  2000/01/06 05:57:06  steve
- *  Only sign-extend unsized numbers.
- *
- * Revision 1.12  1999/11/06 16:00:17  steve
- *  Put number constants into a static table.
- *
- * Revision 1.11  1999/10/22 23:57:53  steve
- *  do the <= in bits, not numbers.
- *
- * Revision 1.10  1999/10/10 23:29:37  steve
- *  Support evaluating + operator at compile time.
  */
 
