@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2001 Stephen Williams (steve@icarus.com)
+ * Copyright (c) 2001-2003 Stephen Williams (steve@icarus.com)
  *
  *    This source code is free software; you can redistribute it
  *    and/or modify it in source code form under the terms of the GNU
@@ -16,22 +16,24 @@
  *    along with this program; if not, write to the Free Software
  *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
-#if !defined(WINNT)
-#ident "$Id: compile.cc,v 1.120 2002/01/11 05:21:47 steve Exp $"
+#ifdef HAVE_CVS_IDENT
+#ident "$Id: compile.cc,v 1.175 2004/10/04 01:10:58 steve Exp $"
 #endif
 
 # include  "arith.h"
 # include  "compile.h"
 # include  "functor.h"
+# include  "logic.h"
 # include  "resolv.h"
-# include  "udp.h" 
+# include  "udp.h"
 # include  "memory.h"
-# include  "force.h" 
+# include  "force.h"
 # include  "symbols.h"
 # include  "codes.h"
 # include  "schedule.h"
 # include  "vpi_priv.h"
 # include  "parse_misc.h"
+# include  "statistics.h"
 #ifdef HAVE_MALLOC_H
 # include  <malloc.h>
 #endif
@@ -68,6 +70,8 @@ enum operand_e {
       OA_FUNC_PTR2,
        /* The operand is a pointer to a memory */
       OA_MEM_PTR,
+	/* The operand is a VPI handle */
+      OA_VPI_PTR,
 };
 
 struct opcode_table_s {
@@ -80,21 +84,34 @@ struct opcode_table_s {
 
 const static struct opcode_table_s opcode_table[] = {
       { "%add",    of_ADD,    3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%add/wr", of_ADD_WR, 2,  {OA_BIT1,     OA_BIT2,     OA_NONE} },
+      { "%addi",   of_ADDI,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%and",    of_AND,    3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%and/r",  of_ANDR,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%assign", of_ASSIGN, 3,  {OA_FUNC_PTR, OA_BIT1,     OA_BIT2} },
+      { "%assign/d", of_ASSIGN_D, 3,  {OA_FUNC_PTR, OA_BIT1, OA_BIT2} },
       { "%assign/m",of_ASSIGN_MEM,3,{OA_MEM_PTR,OA_BIT1,     OA_BIT2} },
+      { "%assign/v0",of_ASSIGN_V0,3,{OA_FUNC_PTR,OA_BIT1,    OA_BIT2} },
+      { "%assign/wr",of_ASSIGN_WR,3,{OA_VPI_PTR,OA_BIT1,     OA_BIT2} },
       { "%assign/x0",of_ASSIGN_X0,3,{OA_FUNC_PTR,OA_BIT1,    OA_BIT2} },
+      { "%blend",  of_BLEND,  3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%breakpoint", of_BREAKPOINT, 0,  {OA_NONE, OA_NONE, OA_NONE} },
       { "%cassign",of_CASSIGN,2,  {OA_FUNC_PTR, OA_FUNC_PTR2,OA_NONE} },
       { "%cmp/s",  of_CMPS,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%cmp/u",  of_CMPU,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%cmp/wr", of_CMPWR,  2,  {OA_BIT1,     OA_BIT2,     OA_NONE} },
       { "%cmp/x",  of_CMPX,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%cmp/z",  of_CMPZ,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%cmpi/u", of_CMPIU,  3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%cvt/ir", of_CVT_IR, 2,  {OA_BIT1,     OA_BIT2,     OA_NONE} },
+      { "%cvt/ri", of_CVT_RI, 2,  {OA_BIT1,     OA_BIT2,     OA_NONE} },
+      { "%cvt/vr", of_CVT_VR, 3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%deassign",of_DEASSIGN,2,{OA_FUNC_PTR, OA_BIT1,     OA_NONE} },
       { "%delay",  of_DELAY,  1,  {OA_NUMBER,   OA_NONE,     OA_NONE} },
       { "%delayx", of_DELAYX, 1,  {OA_NUMBER,   OA_NONE,     OA_NONE} },
       { "%div",    of_DIV,    3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%div/s",  of_DIV_S,  3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%div/wr", of_DIV_WR, 2,  {OA_BIT1,     OA_BIT2,     OA_NONE} },
       { "%end",    of_END,    0,  {OA_NONE,     OA_NONE,     OA_NONE} },
       { "%force",  of_FORCE,  2,  {OA_FUNC_PTR, OA_BIT1,     OA_NONE} },
       { "%inv",    of_INV,    2,  {OA_BIT1,     OA_BIT2,     OA_NONE} },
@@ -110,22 +127,37 @@ const static struct opcode_table_s opcode_table[] = {
       { "%join",   of_JOIN,   0,  {OA_NONE,     OA_NONE,     OA_NONE} },
       { "%load",   of_LOAD,   2,  {OA_BIT1,     OA_FUNC_PTR, OA_NONE} },
       { "%load/m", of_LOAD_MEM,2, {OA_BIT1,     OA_MEM_PTR,  OA_NONE} },
+      { "%load/nx",of_LOAD_NX,3,  {OA_BIT1,     OA_VPI_PTR,  OA_BIT2} },
+      { "%load/v", of_LOAD_VEC,3, {OA_BIT1,     OA_FUNC_PTR, OA_BIT2} },
+      { "%load/wr",of_LOAD_WR,2,  {OA_BIT1,     OA_VPI_PTR,  OA_BIT2} },
       { "%load/x", of_LOAD_X, 3,  {OA_BIT1,     OA_FUNC_PTR, OA_BIT2} },
+      { "%loadi/wr",of_LOADI_WR,3,{OA_BIT1,     OA_NUMBER,   OA_BIT2} },
       { "%mod",    of_MOD,    3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%mod/s",  of_MOD_S,  3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%mov",    of_MOV,    3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%mul",    of_MUL,    3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%mul/wr", of_MUL_WR, 2,  {OA_BIT1,     OA_BIT2,     OA_NONE} },
+      { "%muli",   of_MULI,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%nand",   of_NAND,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%nand/r", of_NANDR,  3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%noop",   of_NOOP,   0,  {OA_NONE,     OA_NONE,     OA_NONE} },
+      { "%nor",    of_NOR,    3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%nor/r",  of_NORR,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%or",     of_OR,     3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%or/r",   of_ORR,    3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%release",of_RELEASE,1,  {OA_FUNC_PTR, OA_NONE,     OA_NONE} },
       { "%set",    of_SET,    2,  {OA_FUNC_PTR, OA_BIT1,     OA_NONE} },
       { "%set/m",  of_SET_MEM,2,  {OA_MEM_PTR,  OA_BIT1,     OA_NONE} },
-      { "%set/x",  of_SET_X,  3,  {OA_FUNC_PTR, OA_BIT1,     OA_BIT2} },
+      { "%set/v",  of_SET_VEC,3,  {OA_FUNC_PTR, OA_BIT1,     OA_BIT2} },
+      { "%set/wr", of_SET_WORDR,2,{OA_VPI_PTR,  OA_BIT1,     OA_NONE} },
+      { "%set/x0", of_SET_X0, 3,  {OA_FUNC_PTR, OA_BIT1,     OA_BIT2} },
+      { "%set/x0/x",of_SET_X0_X,3,{OA_FUNC_PTR, OA_BIT1,     OA_BIT2} },
       { "%shiftl/i0", of_SHIFTL_I0, 2, {OA_BIT1,OA_NUMBER,   OA_NONE} },
       { "%shiftr/i0", of_SHIFTR_I0, 2, {OA_BIT1,OA_NUMBER,   OA_NONE} },
+      { "%shiftr/s/i0", of_SHIFTR_S_I0,2,{OA_BIT1,OA_NUMBER, OA_NONE} },
       { "%sub",    of_SUB,    3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
+      { "%sub/wr", of_SUB_WR, 2,  {OA_BIT1,     OA_BIT2,     OA_NONE} },
+      { "%subi",   of_SUBI,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%wait",   of_WAIT,   1,  {OA_FUNC_PTR, OA_NONE,     OA_NONE} },
       { "%xnor",   of_XNOR,   3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
       { "%xnor/r", of_XNORR,  3,  {OA_BIT1,     OA_BIT2,     OA_NUMBER} },
@@ -190,22 +222,35 @@ void define_functor_symbol(const char*label, vvp_ipoint_t ipt)
 
 static vvp_ipoint_t lookup_functor_symbol(const char*label)
 {
+      assert(sym_functors);
       symbol_value_t val = sym_get_value(sym_functors, label);
       return val.num;
 }
 
 static vvp_ipoint_t ipoint_lookup(const char *label, unsigned idx)
 {
-        /* First, look to see if the symbol is a signal, whether net
-	   or reg. If so, get the correct bit out. */
+        /* First, look to see if the symbol is a vpi object of some
+	   sort. If it is, then get the vvp_ipoint_t pointer out of
+	   the vpiHandle. */
       symbol_value_t val = sym_get_value(sym_vpi, label);
       if (val.ptr) {
 	    vpiHandle vpi = (vpiHandle) val.ptr;
-	    assert((vpi->vpi_type->type_code == vpiNet)
-		   || (vpi->vpi_type->type_code == vpiReg));
+	    switch (vpi->vpi_type->type_code) {
+		case vpiNet:
+		case vpiReg:
+		case vpiIntegerVar: {
+		      __vpiSignal*sig = (__vpiSignal*)vpi;
+		      return vvp_fvector_get(sig->bits, idx);
+		}
 
-	    __vpiSignal*sig = (__vpiSignal*)vpi;
-	    return vvp_fvector_get(sig->bits, idx);
+		case vpiNamedEvent: {
+		      __vpiNamedEvent*tmp = (__vpiNamedEvent*)vpi;
+		      return tmp->funct;
+		}
+
+		default:
+		  assert(0);
+	    }
       }
 
 	/* Failing that, look for a general functor. */
@@ -281,11 +326,11 @@ bool functor_resolv_list_s::resolve(bool mes)
 
       if (mes)
 	    fprintf(stderr, "unresolved functor reference: %s\n", source);
-      
+
       return false;
 }
 
-inline static 
+inline static
 void postpone_functor_input(vvp_ipoint_t ptr, char*lab, unsigned idx)
 {
       struct functor_resolv_list_s*res = new struct functor_resolv_list_s;
@@ -321,14 +366,13 @@ bool functor_gen_resolv_list_s::resolve(bool mes)
 
       if (mes)
 	    fprintf(stderr, "unresolved functor reference: %s\n", source);
-      
+
       return false;
 }
 
-inline static 
 void functor_ref_lookup(vvp_ipoint_t *ref, char*lab, unsigned idx)
 {
-      struct functor_gen_resolv_list_s*res = 
+      struct functor_gen_resolv_list_s*res =
 	    new struct functor_gen_resolv_list_s;
 
       res->ref    = ref;
@@ -354,12 +398,39 @@ bool vpi_handle_resolv_list_s::resolve(bool mes)
       if (!val.ptr) {
 	    // check for thread vector  T<base,wid>
 	    unsigned base, wid;
-	    unsigned n;
-	    if (2 <= sscanf(label, "T<%u,%u>%n", &base, &wid, &n) 
+	    unsigned n = 0;
+	    char ss[32];
+	    if (2 <= sscanf(label, "T<%u,%u>%n", &base, &wid, &n)
 		&& n == strlen(label)) {
-		  val.ptr = vpip_make_vthr_vector(base, wid);
+		  val.ptr = vpip_make_vthr_vector(base, wid, false);
+		  sym_set_value(sym_vpi, label, val);
+
+	    } else if (3 <= sscanf(label, "T<%u,%u,%[su]>%n", &base,
+				   &wid, ss, &n)
+		       && n == strlen(label)) {
+
+		  bool signed_flag = false;
+		  for (char*fp = ss ;  *fp ;  fp += 1) switch (*fp) {
+		      case 's':
+			signed_flag = true;
+			break;
+		      case 'u':
+			signed_flag = false;
+			break;
+		      default:
+			break;
+		  }
+
+		  val.ptr = vpip_make_vthr_vector(base, wid, signed_flag);
+		  sym_set_value(sym_vpi, label, val);
+
+	    } else if (2 == sscanf(label, "W<%u,%[r]>%n", &base, ss, &n)
+		       && n == strlen(label)) {
+
+		  val.ptr = vpip_make_vthr_word(base, ss);
 		  sym_set_value(sym_vpi, label, val);
 	    }
+
       }
 
       if (!val.ptr) {
@@ -371,15 +442,39 @@ bool vpi_handle_resolv_list_s::resolve(bool mes)
 	    free(label);
 	    return true;
       }
-      
+
       if (mes)
 	    fprintf(stderr, "unresolved vpi name lookup: %s\n", label);
-      
+
       return false;
 }
 
 void compile_vpi_lookup(vpiHandle *handle, char*label)
 {
+      if (strcmp(label, "$time") == 0) {
+	    *handle = vpip_sim_time(vpip_peek_current_scope());
+	    free(label);
+	    return;
+      }
+
+      if (strcmp(label, "$stime") == 0) {
+	    *handle = vpip_sim_time(vpip_peek_current_scope());
+	    free(label);
+	    return;
+      }
+
+      if (strcmp(label, "$realtime") == 0) {
+	    *handle = vpip_sim_realtime(vpip_peek_current_scope());
+	    free(label);
+	    return;
+      }
+
+      if (strcmp(label, "$simtime") == 0) {
+	    *handle = vpip_sim_time(0);
+	    free(label);
+	    return;
+      }
+
       struct vpi_handle_resolv_list_s*res
 	    = new struct vpi_handle_resolv_list_s;
 
@@ -403,9 +498,9 @@ bool code_label_resolv_list_s::resolve(bool mes)
       symbol_value_t val = sym_get_value(sym_codespace, label);
       if (val.num) {
 	    if (code->opcode == of_FORK)
-		  code->cptr2 = val.num;
+		  code->cptr2 = reinterpret_cast<vvp_code_t>(val.ptr);
 	    else
-		  code->cptr = val.num;
+		  code->cptr = reinterpret_cast<vvp_code_t>(val.ptr);
 	    free(label);
 	    return true;
       }
@@ -418,7 +513,6 @@ bool code_label_resolv_list_s::resolve(bool mes)
       return false;
 }
 
-inline static
 void code_label_lookup(struct vvp_code_s *code, char *label)
 {
       struct code_label_resolv_list_s *res
@@ -430,6 +524,40 @@ void code_label_lookup(struct vvp_code_s *code, char *label)
       resolv_submit(res);
 }
 
+/*
+ * Lookup memories.
+ */
+struct memory_resolv_list_s: public resolv_list_s {
+      struct vvp_code_s *code;
+      char *label;
+      virtual bool resolve(bool mes);
+};
+
+bool memory_resolv_list_s::resolve(bool mes)
+{
+      code->mem = memory_find(label);
+      if (code->mem != 0) {
+	    free(label);
+	    return true;
+      }
+
+      if (mes)
+	    fprintf(stderr, "Memory unresolved: %s\n", label);
+
+      return false;
+}
+
+static void compile_mem_lookup(struct vvp_code_s *code, char *label)
+{
+      struct memory_resolv_list_s *res
+	    = new struct memory_resolv_list_s;
+
+      res->code  = code;
+      res->label = label;
+
+      resolv_submit(res);
+}
+
 /*
  * When parsing is otherwise complete, this function is called to do
  * the final stuff. Clean up deferred linking here.
@@ -440,7 +568,12 @@ void compile_cleanup(void)
       int lnerrs = -1;
       int nerrs = 0;
       int last;
-      
+
+      if (verbose_flag) {
+	    fprintf(stderr, " ... Linking\n");
+	    fflush(stderr);
+      }
+
       do {
 	    struct resolv_list_s *res = resolv_list;
 	    resolv_list = 0x0;
@@ -459,12 +592,31 @@ void compile_cleanup(void)
 		  }
 	    }
 	    if (nerrs && last)
-		  fprintf(stderr, 
-			  "compile_cleanup: %d unresolved items\n", 
+		  fprintf(stderr,
+			  "compile_cleanup: %d unresolved items\n",
 			  nerrs);
       } while (nerrs && !last);
 
       compile_errors += nerrs;
+
+      if (verbose_flag) {
+	    fprintf(stderr, " ... Removing symbol tables\n");
+	    fflush(stderr);
+      }
+
+	/* After compile is complete, the vpi symbol table is no
+	   longer needed. VPI objects are located by following
+	   scopes. */
+      delete_symbol_table(sym_vpi);
+      sym_vpi = 0;
+
+	/* Don't need the code labels. The instructions have numeric
+	   pointers in them, the symbol table is no longer needed. */
+      delete_symbol_table(sym_codespace);
+      sym_codespace = 0;
+
+      delete_symbol_table(sym_functors);
+      sym_functors = 0;
 }
 
 void compile_vpi_symbol(const char*label, vpiHandle obj)
@@ -481,8 +633,6 @@ void compile_vpi_symbol(const char*label, vpiHandle obj)
 void compile_init(void)
 {
       sym_vpi = new_symbol_table();
-      compile_vpi_symbol("$time",  vpip_sim_time());
-      compile_vpi_symbol("$stime", vpip_sim_time());
 
       sym_functors = new_symbol_table();
       functor_init();
@@ -502,13 +652,13 @@ void compile_vpi_time_precision(long pre)
       vpip_set_time_precision(pre);
 }
 
-/* 
+/*
  * Run through the arguments looking for the functors that are
  * connected to my input ports. For each source functor that I
  * find, connect the output of that functor to the indexed
  * input by inserting myself (complete with the port number in
  * the vvp_ipoint_t) into the list that the source heads.
- *   
+ *
  * If the source functor is not declared yet, then don't do
  * the link yet. Save the reference to be resolved later.
  *
@@ -527,7 +677,7 @@ void inputs_connect(vvp_ipoint_t fdx, unsigned argc, struct symb_s*argv)
 	    vvp_ipoint_t ifdx = ipoint_input_index(fdx, idx);
 	    functor_t iobj = functor_index(ifdx);
 
-	    if (strcmp(argv[idx].text, "C<0>") == 0) 
+	    if (strcmp(argv[idx].text, "C<0>") == 0)
 		  iobj->set(ifdx, false, 0, St0);
 
 	    else if (strcmp(argv[idx].text, "C<we0>") == 0)
@@ -568,18 +718,22 @@ void inputs_connect(vvp_ipoint_t fdx, unsigned argc, struct symb_s*argv)
 
 
 struct const_functor_s: public functor_s {
-      const_functor_s(unsigned str0, unsigned str1) 
+      const_functor_s(unsigned str0, unsigned str1)
 	    { odrive0 = str0; odrive1 = str1; }
-      virtual void set(vvp_ipoint_t, bool, unsigned, unsigned);   
+      virtual void set(vvp_ipoint_t, bool, unsigned, unsigned);
 };
-void const_functor_s::set(vvp_ipoint_t, bool, unsigned, unsigned) 
+void const_functor_s::set(vvp_ipoint_t p, bool, unsigned val, unsigned)
 {
-      assert(! "const_functor_s::set: don't do that");
+      fprintf(stderr, "internal error: Set value to const_functor 0x%x\n", p);
+      fprintf(stderr, "              : Value is %u, trying to set %u\n",
+	      oval, val);
+      fprintf(stderr, "              : I'm driving functor 0x%x\n", out);
+      assert(0);
 }
 
 
-static vvp_ipoint_t make_const_functor(unsigned val, 
-				       unsigned str0, 
+static vvp_ipoint_t make_const_functor(unsigned val,
+				       unsigned str0,
 				       unsigned str1)
 {
       vvp_ipoint_t fdx = functor_allocate(1);
@@ -604,10 +758,10 @@ static void functor_reference(vvp_ipoint_t *ref, char *lab, unsigned idx)
       else if (strcmp(lab, "C<su0>") == 0)
 	    *ref = make_const_functor(0,7,7);
 
-      else if (strcmp(lab, "C<pu0>") == 0) 
+      else if (strcmp(lab, "C<pu0>") == 0)
 	    *ref = make_const_functor(0,5,5);
 
-      else if (strcmp(lab, "C<we0>") == 0) 
+      else if (strcmp(lab, "C<we0>") == 0)
 	    *ref = make_const_functor(0,3,3);
 
       else if (strcmp(lab, "C<1>") == 0)
@@ -618,16 +772,16 @@ static void functor_reference(vvp_ipoint_t *ref, char *lab, unsigned idx)
 
       else if (strcmp(lab, "C<pu1>") == 0)
 	    *ref = make_const_functor(1,5,5);
- 
+
       else if (strcmp(lab, "C<we1>") == 0)
 	    *ref = make_const_functor(1,3,3);
- 
+
       else if (strcmp(lab, "C<x>") == 0)
 	    *ref = make_const_functor(2,6,6);
-      
+
       else if (strcmp(lab, "C<z>") == 0)
 	    *ref = make_const_functor(3,6,6);
-      
+
       else {
 	    functor_ref_lookup(ref, lab, idx);
 	    return;
@@ -647,7 +801,7 @@ static void make_extra_outputs(vvp_ipoint_t fdx, unsigned wid)
 }
 
 static void make_arith(vvp_arith_ *arith,
-		       char*label, long wid, 
+		       char*label, long wid,
 		       unsigned argc, struct symb_s*argv)
 {
       vvp_ipoint_t fdx = functor_allocate(wid);
@@ -676,7 +830,7 @@ static void make_arith(vvp_arith_ *arith,
       free(argv);
 }
 
-void compile_arith_div(char*label, long wid,
+void compile_arith_div(char*label, long wid, bool signed_flag,
 		       unsigned argc, struct symb_s*argv)
 {
       assert( wid > 0 );
@@ -687,7 +841,7 @@ void compile_arith_div(char*label, long wid,
 	    return;
       }
 
-      vvp_arith_ *arith = new vvp_arith_div(wid);
+      vvp_arith_ *arith = new vvp_arith_div(wid, signed_flag);
 
       make_arith(arith, label, wid, argc, argv);
 }
@@ -768,7 +922,7 @@ void compile_arith_sum(char*label, long wid, unsigned argc, struct symb_s*argv)
       make_arith(arith, label, wid, argc, argv);
 }
 
-void compile_cmp_ge(char*label, long wid, unsigned argc, struct symb_s*argv)
+void compile_cmp_eq(char*label, long wid, unsigned argc, struct symb_s*argv)
 {
       assert( wid > 0 );
 
@@ -778,12 +932,12 @@ void compile_cmp_ge(char*label, long wid, unsigned argc, struct symb_s*argv)
 	    return;
       }
 
-      vvp_arith_ *arith = new vvp_cmp_ge(wid);
+      vvp_arith_ *arith = new vvp_cmp_eq(wid);
 
       make_arith(arith, label, wid, argc, argv);
 }
 
-void compile_cmp_gt(char*label, long wid, unsigned argc, struct symb_s*argv)
+void compile_cmp_ne(char*label, long wid, unsigned argc, struct symb_s*argv)
 {
       assert( wid > 0 );
 
@@ -793,7 +947,39 @@ void compile_cmp_gt(char*label, long wid, unsigned argc, struct symb_s*argv)
 	    return;
       }
 
-      vvp_arith_ *arith = new vvp_cmp_gt(wid);
+      vvp_arith_ *arith = new vvp_cmp_ne(wid);
+
+      make_arith(arith, label, wid, argc, argv);
+}
+
+void compile_cmp_ge(char*label, long wid, bool signed_flag,
+		    unsigned argc, struct symb_s*argv)
+{
+      assert( wid > 0 );
+
+      if ((long)argc != 2*wid) {
+	    fprintf(stderr, "%s; .cmp has wrong number of symbols\n", label);
+	    compile_errors += 1;
+	    return;
+      }
+
+      vvp_arith_ *arith = new vvp_cmp_ge(wid, signed_flag);
+
+      make_arith(arith, label, wid, argc, argv);
+}
+
+void compile_cmp_gt(char*label, long wid, bool signed_flag,
+		    unsigned argc, struct symb_s*argv)
+{
+      assert( wid > 0 );
+
+      if ((long)argc != 2*wid) {
+	    fprintf(stderr, "%s; .cmp has wrong number of symbols\n", label);
+	    compile_errors += 1;
+	    return;
+      }
+
+      vvp_arith_ *arith = new vvp_cmp_gt(wid, signed_flag);
 
       make_arith(arith, label, wid, argc, argv);
 }
@@ -876,7 +1062,7 @@ void compile_shiftr(char*label, long wid, unsigned argc, struct symb_s*argv)
       }
 
       vvp_arith_ *arith = new vvp_shiftr(wid);
-      
+
       make_shift(arith, label, wid, argc, argv);
 }
 
@@ -895,6 +1081,12 @@ void compile_resolver(char*label, char*type, unsigned argc, struct symb_s*argv)
       } else if (strcmp(type,"tri1") == 0) {
 	    obj = new resolv_functor_s(Pu1);
 
+      } else if (strcmp(type,"triand") == 0) {
+	    obj = new table_functor_s(ft_TRIAND, 6, 6);
+
+      } else if (strcmp(type,"trior") == 0) {
+	    obj = new table_functor_s(ft_TRIOR, 6, 6);
+
       } else {
 	    fprintf(stderr, "invalid resolver type: %s\n", type);
 	    compile_errors += 1;
@@ -904,7 +1096,7 @@ void compile_resolver(char*label, char*type, unsigned argc, struct symb_s*argv)
 	    vvp_ipoint_t fdx = functor_allocate(1);
 	    functor_define(fdx, obj);
 	    define_functor_symbol(label, fdx);
-	    
+
 	    inputs_connect(fdx, argc, argv);
       }
 
@@ -969,15 +1161,15 @@ void compile_udp_functor(char*label, char*type,
 {
       struct vvp_udp_s *u = udp_find(type);
       assert (argc == u->nin);
-      
+
       functor_t udp = new udp_functor_s(u);
-      
+
       unsigned nfun = (argc+3)/4;
       vvp_ipoint_t fdx = functor_allocate(nfun);
       functor_define(fdx, udp);
       define_functor_symbol(label, fdx);
-      free(label);  
-      
+      free(label);
+
       if (nfun > 1) {
 	    for (unsigned i=0;  i < nfun-1;  i++) {
 		  functor_t fu = new edge_inputs_functor_s;
@@ -991,7 +1183,7 @@ void compile_udp_functor(char*label, char*type,
 
       inputs_connect(fdx, argc, argv);
       free(argv);
-      
+
       if (u->sequ)
 	    udp->put_oval(u->init, false);
 }
@@ -1005,11 +1197,12 @@ void compile_memory(char *label, char *name, int msb, int lsb,
 
   vpiHandle obj = vpip_make_memory(mem);
   compile_vpi_symbol(label, obj);
+  vpip_attach_to_current_scope(obj);
 
   free(label);
 }
 
-void compile_memory_port(char *label, char *memid, 
+void compile_memory_port(char *label, char *memid,
 			 unsigned msb, unsigned lsb,
 			 unsigned naddr,
 			 unsigned argc, struct symb_s *argv)
@@ -1018,8 +1211,8 @@ void compile_memory_port(char *label, char *memid,
   free(memid);
   assert(mem);
 
-  // This is not a Verilog bit range. 
-  // This is a data port bit range. 
+  // This is not a Verilog bit range.
+  // This is a data port bit range.
   assert (lsb >= 0  &&  lsb<=msb);
   assert (msb < memory_data_width(mem));
   unsigned nbits = msb-lsb+1;
@@ -1053,7 +1246,7 @@ void compile_memory_init(char *memid, unsigned i, unsigned char val)
 /*
  * The parser uses this function to compile and link an executable
  * opcode. I do this by looking up the opcode in the opcode_table. The
- * table gives the operand structure that is acceptible, so I can
+ * table gives the operand structure that is acceptable, so I can
  * process the operands here as well.
  */
 void compile_code(char*label, char*mnem, comp_operands_t opa)
@@ -1061,11 +1254,9 @@ void compile_code(char*label, char*mnem, comp_operands_t opa)
 	/* First, I can give the label a value that is the current
 	   codespace pointer. Don't need the text of the label after
 	   this is done. */
-      if (label) 
+      if (label)
 	    compile_codelabel(label);
 
-      vvp_cpoint_t ptr = codespace_allocate();
-
 	/* Lookup the opcode in the opcode table. */
       struct opcode_table_s*op = (struct opcode_table_s*)
 	    bsearch(mnem, opcode_table, opcode_count,
@@ -1079,7 +1270,7 @@ void compile_code(char*label, char*mnem, comp_operands_t opa)
 
 	/* Build up the code from the information about the opcode and
 	   the information from the compiler. */
-      vvp_code_t code = codespace_index(ptr);
+      vvp_code_t code = codespace_allocate();
       code->opcode = op->opcode;
 
       if (op->argc != (opa? opa->argc : 0)) {
@@ -1101,7 +1292,6 @@ void compile_code(char*label, char*mnem, comp_operands_t opa)
 			yyerror("operand format");
 			break;
 		  }
-
 		  code->bit_idx[0] = opa->argv[idx].numb;
 		  break;
 
@@ -1110,7 +1300,6 @@ void compile_code(char*label, char*mnem, comp_operands_t opa)
 			yyerror("operand format");
 			break;
 		  }
-
 		  code->bit_idx[1] = opa->argv[idx].numb;
 		  break;
 
@@ -1133,7 +1322,7 @@ void compile_code(char*label, char*mnem, comp_operands_t opa)
 			break;
 		  }
 
-		  functor_ref_lookup(&code->iptr, 
+		  functor_ref_lookup(&code->iptr,
 				     opa->argv[idx].symb.text,
 				     opa->argv[idx].symb.idx);
 		  break;
@@ -1164,7 +1353,7 @@ void compile_code(char*label, char*mnem, comp_operands_t opa)
 			free(opa->argv[idx].symb.text);
 
 		  } else {
-			functor_ref_lookup(&code->iptr2, 
+			functor_ref_lookup(&code->iptr2,
 					   opa->argv[idx].symb.text,
 					   opa->argv[idx].symb.idx);
 		  }
@@ -1185,14 +1374,20 @@ void compile_code(char*label, char*mnem, comp_operands_t opa)
 			break;
 		  }
 
-		  code->mem = memory_find(opa->argv[idx].symb.text);
-		  if (code->mem == 0) {
-			yyerror("memory undefined");
-		  }
-
-		  free(opa->argv[idx].symb.text);
+		  compile_mem_lookup(code, opa->argv[idx].symb.text);
 		  break;
 
+		case OA_VPI_PTR:
+		    /* The operand is a functor. Resolve the label to
+		       a functor pointer, or postpone the resolution
+		       if it is not defined yet. */
+		  if (opa->argv[idx].ltype != L_SYMB) {
+			yyerror("operand format");
+			break;
+		  }
+
+		  compile_vpi_lookup(&code->handle, opa->argv[idx].symb.text);
+		  break;
 	    }
       }
 
@@ -1204,9 +1399,9 @@ void compile_code(char*label, char*mnem, comp_operands_t opa)
 void compile_codelabel(char*label)
 {
       symbol_value_t val;
-      vvp_cpoint_t ptr = codespace_next();
+      vvp_code_t ptr = codespace_next();
 
-      val.num = ptr;
+      val.ptr = ptr;
       sym_set_value(sym_codespace, label, val);
 
       free(label);
@@ -1215,13 +1410,11 @@ void compile_codelabel(char*label)
 
 void compile_disable(char*label, struct symb_s symb)
 {
-      if (label) 
+      if (label)
 	    compile_codelabel(label);
 
-      vvp_cpoint_t ptr = codespace_allocate();
-
 	/* Fill in the basics of the %disable in the instruction. */
-      vvp_code_t code = codespace_index(ptr);
+      vvp_code_t code = codespace_allocate();
       code->opcode = of_DISABLE;
 
       compile_vpi_lookup(&code->handle, symb.text);
@@ -1235,13 +1428,12 @@ void compile_disable(char*label, struct symb_s symb)
  */
 void compile_fork(char*label, struct symb_s dest, struct symb_s scope)
 {
-      if (label) 
+      if (label)
 	    compile_codelabel(label);
 
-      vvp_cpoint_t ptr = codespace_allocate();
 
 	/* Fill in the basics of the %fork in the instruction. */
-      vvp_code_t code = codespace_index(ptr);
+      vvp_code_t code = codespace_allocate();
       code->opcode = of_FORK;
 
 	/* Figure out the target PC. */
@@ -1256,10 +1448,8 @@ void compile_vpi_call(char*label, char*name, unsigned argc, vpiHandle*argv)
       if (label)
 	    compile_codelabel(label);
 
-      vvp_cpoint_t ptr = codespace_allocate();
-
 	/* Create an instruction in the code space. */
-      vvp_code_t code = codespace_index(ptr);
+      vvp_code_t code = codespace_allocate();
       code->opcode = &of_VPI_CALL;
 
 	/* Create a vpiHandle that bundles the call information, and
@@ -1273,16 +1463,14 @@ void compile_vpi_call(char*label, char*name, unsigned argc, vpiHandle*argv)
 }
 
 void compile_vpi_func_call(char*label, char*name,
-			   unsigned vbit, unsigned vwid,
+			   unsigned vbit, int vwid,
 			   unsigned argc, vpiHandle*argv)
 {
-      if (label) 
+      if (label)
 	    compile_codelabel(label);
-      
-      vvp_cpoint_t ptr = codespace_allocate();
 
 	/* Create an instruction in the code space. */
-      vvp_code_t code = codespace_index(ptr);
+      vvp_code_t code = codespace_allocate();
       code->opcode = &of_VPI_CALL;
 
 	/* Create a vpiHandle that bundles the call information, and
@@ -1300,18 +1488,26 @@ void compile_vpi_func_call(char*label, char*name,
  * with the start address referenced by the program symbol passed to
  * me.
  */
-void compile_thread(char*start_sym)
+void compile_thread(char*start_sym, char*flag)
 {
+      bool push_flag = false;
+
       symbol_value_t tmp = sym_get_value(sym_codespace, start_sym);
-      vvp_cpoint_t pc = tmp.num;
+      vvp_code_t pc = reinterpret_cast<vvp_code_t>(tmp.ptr);
       if (pc == 0) {
 	    yyerror("unresolved address");
 	    return;
       }
 
+      if (flag && (strcmp(flag,"$push") == 0))
+	    push_flag = true;
+
       vthread_t thr = vthread_new(pc, vpip_peek_current_scope());
-      schedule_vthread(thr, 0);
+      schedule_vthread(thr, 0, push_flag);
+
       free(start_sym);
+      if (flag != 0)
+	    free(flag);
 }
 
 /*
@@ -1319,25 +1515,29 @@ void compile_thread(char*start_sym)
  * write the label into the symbol table.
  */
 void compile_variable(char*label, char*name, int msb, int lsb,
-		      bool signed_flag)
+		      char signed_flag)
 {
       unsigned wid = ((msb > lsb)? msb-lsb : lsb-msb) + 1;
 
       vvp_ipoint_t fdx = functor_allocate(wid);
       define_functor_symbol(label, fdx);
 
+      functor_t fu = new var_functor_s [wid];
       for (unsigned idx = 0 ;  idx < wid ;  idx += 1) {
-	    functor_t fu = new var_functor_s;
-	    functor_define(ipoint_index(fdx, idx), fu);
+	    functor_define(ipoint_index(fdx, idx), fu+idx);
       }
+      count_functors_var += wid;
 
 	/* Make the vpiHandle for the reg. */
       vvp_fvector_t vec = vvp_fvector_continuous_new(wid, fdx);
-      vpiHandle obj = vpip_make_reg(name, msb, lsb, signed_flag, vec);
+      vpiHandle obj = (signed_flag > 1) ?
+			vpip_make_int(name, msb, lsb, vec) :
+			vpip_make_reg(name, msb, lsb, signed_flag!=0, vec);
       compile_vpi_symbol(label, obj);
       vpip_attach_to_current_scope(obj);
 
       free(label);
+      free(name);
 }
 
 void compile_net(char*label, char*name, int msb, int lsb, bool signed_flag,
@@ -1361,64 +1561,70 @@ void compile_net(char*label, char*name, int msb, int lsb, bool signed_flag,
       vpip_attach_to_current_scope(obj);
 
       free(label);
+      free(name);
       free(argv);
 }
 
-/*
- * These functions are in support of the debugger.
- *
- * debug_lookup_functor
- *    Use a name to locate a functor address. This only gets the LSB
- *    of a vector, but it is enough to locate the object, or, is it?
- */
-vvp_ipoint_t debug_lookup_functor(const char*name)
+void compile_param_string(char*label, char*name, char*str, char*value)
 {
-      return lookup_functor_symbol(name);
-}
+      assert(strcmp(str,"string") == 0);
+      free(str);
 
+      vpiHandle obj = vpip_make_string_param(name, value);
+      compile_vpi_symbol(label, obj);
+      vpip_attach_to_current_scope(obj);
+
+      free(label);
+}
 
 /*
  * $Log: compile.cc,v $
- * Revision 1.120  2002/01/11 05:21:47  steve
- *  Magic stime object support.
+ * Revision 1.175  2004/10/04 01:10:58  steve
+ *  Clean up spurious trailing white space.
  *
- * Revision 1.119  2002/01/06 03:15:13  steve
- *  Support weak functor inputs.
+ * Revision 1.174  2004/06/30 02:15:57  steve
+ *  Add signed LPM divide.
  *
- * Revision 1.118  2002/01/03 04:19:02  steve
- *  Add structural modulus support down to vvp.
+ * Revision 1.173  2004/06/19 15:52:53  steve
+ *  Add signed modulus operator.
  *
- * Revision 1.117  2001/12/15 02:11:51  steve
- *  Give tri0 and tri1 their proper strengths.
+ * Revision 1.172  2004/06/16 16:33:26  steve
+ *  Add structural equality compare nodes.
  *
- * Revision 1.116  2001/12/15 01:54:38  steve
- *  Support tri0 and tri1 resolvers.
+ * Revision 1.171  2004/05/19 03:26:24  steve
+ *  Support delayed/non-blocking assignment to reals and others.
  *
- * Revision 1.115  2001/12/06 03:31:24  steve
- *  Support functor delays for gates and UDP devices.
- *  (Stephan Boettcher)
+ * Revision 1.170  2003/09/04 20:26:31  steve
+ *  Add $push flag for threads.
  *
- * Revision 1.114  2001/11/07 03:34:42  steve
- *  Use functor pointers where vvp_ipoint_t is unneeded.
+ * Revision 1.169  2003/07/30 01:13:28  steve
+ *  Add support for triand and trior.
  *
- * Revision 1.113  2001/11/06 03:07:21  steve
- *  Code rearrange. (Stephan Boettcher)
+ * Revision 1.168  2003/07/22 20:30:24  steve
+ *  Forgot to read the /x parameter for %load/x
  *
- * Revision 1.112  2001/11/01 04:42:39  steve
- *  Handle procedural constant functor pointers.
+ * Revision 1.167  2003/07/03 20:03:36  steve
+ *  Remove the vvp_cpoint_t indirect code pointer.
  *
- * Revision 1.111  2001/11/01 03:00:19  steve
- *  Add force/cassign/release/deassign support. (Stephan Boettcher)
+ * Revision 1.166  2003/06/18 03:55:19  steve
+ *  Add arithmetic shift operators.
  *
- * Revision 1.110  2001/10/31 04:27:46  steve
- *  Rewrite the functor type to have fewer functor modes,
- *  and use objects to manage the different types.
- *  (Stephan Boettcher)
+ * Revision 1.165  2003/06/17 19:17:42  steve
+ *  Remove short int restrictions from vvp opcodes.
  *
- * Revision 1.109  2001/10/18 17:30:25  steve
- *  Support rnpmos devices. (Philip Blundell)
+ * Revision 1.164  2003/05/26 04:44:54  steve
+ *  Add the set/x0/x instruction.
  *
- * Revision 1.108  2001/10/16 02:47:37  steve
- *  Add arith/div object.
+ * Revision 1.163  2003/05/25 03:04:55  steve
+ *  Useless cast.
+ *
+ * Revision 1.162  2003/05/24 02:48:37  steve
+ *  More thorough overflow error message.
+ *
+ * Revision 1.161  2003/05/23 03:44:34  steve
+ *  Assert that parameters fix into opcode.
+ *
+ * Revision 1.160  2003/04/23 03:09:25  steve
+ *  VPI Access to named events.
  */
 
